/* tidelib.h - external declarations for tidelib.c */

#include <time.h>
#include <stdlib.h>

typedef float datatype;		/* Float or double - the data arrays */

/*
 * IMPORTANT - FOLLOWING MUST BE SAME AS IN TIDE TOOL SOURCE
 * Note that this defines max database ID (64) and max number database entries (1024)
 */
#define ID_SHIFT 10		/* an index is (index in db) | (db index in array << ID_SHIFT) */
#define ID_MASK ((1 << ID_SHIFT) - 1)

#define DBTYPE_FIRST	19			/* Increase this by 3 every time db file format changes */
#define DBTYPE_ARGS	DBTYPE_FIRST	/* First byte of database header is type */
#define DBTYPE_LOC	(DBTYPE_FIRST+1)	/* Present values are for 2.1b3 and later */
#define DBTYPE_USER	(DBTYPE_FIRST+2)

#define linelen 300
#define MAX_NUM_CSTS 192
#define CST_ARRAY_SIZE 6			/* MAX_NUM_CSTS / 32, rounded up */

/***** This must be the same as in the Pilot program! ****/
#define MAXLOCSTR 60
#define MAXTZSTR 12

/* Width of graph in ASCII graph mode */
#define TEXTWIDTH 73
/* Tstep adjustment for ASCII graph mode */
#define AGTSTEP 20
/* Tstep adjustment for banner mode */
#define BANTSTEP 10

/* These OSs don't have the nifty zoneinfo database that's needed
   to do global time zones, so fake it as best can.
*/
#define BROKEN_ZONEINFO

enum tideerr {OTHER, BADCOLORSPEC, WANTMOREARGS, CONFIGFAIL,
  BADINTEGER, BADFLOAT, BADHHMM, STRETCHTOOSMALL, STRETCHTOOBIG,
  BADTIMESTAMP, MISSINGYEAR, BADGEOMETRY, CANTOPENDISPLAY, CANTOPENFILE,
  OFFSETSTEXTONLY};

extern void sigabort ( int );  
extern long hhmm2seconds (char *hhmm);
extern void free_data (void);
extern void load_data (void);
extern void check_epoch (void);
extern time_t parse_time_string (char *time_string);
extern int yearoftimet (time_t t);
extern void happy_new_year (int new_year);
extern double time2tide (time_t t);
extern void do_timestamp (char buf[13], struct tm *t);
extern struct tm *tmtime (time_t t);
extern int update_high_tide (void);
extern void list_tides (void);
extern void tide2ascii (void);
extern void tide2ppm (char *);
extern time_t prev_hour (time_t);
extern time_t prev_day (time_t);
extern time_t increment_day (time_t);
extern int change_time_zone (char *);
extern void barf (enum tideerr);
extern void do_calendar ( void );
extern void do_banner ( void );
extern char *stradoop (char *);

typedef enum {
	dSun, dMon, dTue, dWed, dThu, dFri, dSat} DayOfWeek;		/* 0 through 6 */

#define M_WEEKDAY 7		/* Mask for day of the week */
#define ON_AFTER 0x08		/* Flags: First designated day >= the given date */
#define ON_DATE 0x10	 	/*        Exactly this date */

struct DSTswitch {
	char month;		/* January = 1; 0 means no DST */
	int hour;		/* Time of switch (in minutes after midnight) */
	char mday;
	char wkday;		/* Day of week plus flags */
	};

struct TimeZone {				/* 50 bytes per entry */
	char longname[20];
	char stdtime[6];			/* Short names for standard time */
	char DSTtime[6];			/*   and daylight time */
	long meridian;
	int DSTswitch;			/* Time added for DST, in seconds */
	struct DSTswitch toDST;
	struct DSTswitch fromDST;
	};

extern struct TimeZone Zones[];

extern char *TZequivs[250][2];

/* CONTROL VARIABLES FOR DUMPING PALM PILOT DATA FILE */
extern char dumpdata;    	/* Flag to dump data for Pilot to compile in */
extern char dumpfactors;    	/* Flag to dump factors data for Pilot to compile in */
extern int dumpyear;		/* Year to start dumping data */
extern int nyears;		/* How many years to dump */
extern double latfrom, latto, longfrom, longto;	/* Bounds of places to dump */
extern int dbID;
extern char dbNAM[];
extern char deftidename[], defcurrname[];
extern char *exclude[10];
extern char *excluden[10];
extern char *include[10];
extern char *offfiles[10];
extern int nexclude, nexcluden, ninclude, nofffiles;
extern int deftideindex, defcurrindex;
extern double mlat, mlon, nlat, nlon;
extern char printplaces;
extern char *fitfilters[];			/* Filters to reduce size of location stg */
extern char *filters[];
extern int filterit(char *[], char *, int );
extern char *listfile, listflag;		/* Flag = 1 to list included files, 2 for all */
#define NREF_STATIONS 100
extern char *ref_stations[NREF_STATIONS];
extern int nref_stations;
extern char IsRefStation(char *,char );

/* DATA VARIABLES FOR DUMPING PALM PILOT DATA FILE */
/* Tide-prediction storage -- cst == constituent */

#define MAXPLACES 1100				/* Any larger and the structure gets over 64K */
							/* If it needs to be bigger, allocate each entry */
#define MAXLOCN 60
#define MAXTZ 10
struct placedata {
	char *location;
	char iscurrent;
	char issilent;				/* These could be bits if we need more room */
	double latitude, longitude;
	long meridian;
	char tzindex;
	char units;
	float datum;
	int ndata;
	long datamask[CST_ARRAY_SIZE];	/* Bit array indicating which data are nonzero */
	datatype *amp;
	datatype *epoch;
	};

struct placeinfodata {				/* Supplementary data for file generation */
	char *origlocation;			/* Original (long) location string */
	long record1, record2;			/* Offset of data records from start of data */
	float absmax, absmin;			/* Tidal extremes at location */
	};

/* Data structures for locations database - from Tide Tool headers */

#define P_UNITS 7			/* Flag byte: bits for units */
#define P_ISCURRENT 0x80	/* If it's a current (otherwise tide) */
#define P_ISOFFSET  0x40	/* If it's an offset of another location */
#define P_HIDDEN    0x20	/* A hidden reference station - never displayed */
typedef unsigned int DBIndex;	/* an index is (index in db) | (db index in array << ID_SHIFT) */

/**** Hack warning: No floats allowed.  Must all be marked as longs!  See swabf. ***/
struct place {			/* Location information for a place (rec. 1) */
	int latitude;		/* In seconds / 10 */
	int longitude;		/* In seconds / 20 */
	int absmax, absmin;	/* Max and min tidal range * 100 */
	unsigned char lloflow;	/* 10 * longitude overflow + latitude overflow */
	char timezone;		/* >= 0: index in time zone array.  < 0: */
					/*	GMT offset in half hours starting with +13, +12:30, ... */
	char flags;
#if 0  /*** Defeat byte alignment on PC by rolling our own here ****/
	struct {			/* How location name is split for display */
	  unsigned char loc:7;	/* Where to split the string */
	  unsigned char font:1;	/* Font to use */
	  } split;
#else
	unsigned char split;
#endif
	char location[1];		/* Actual size may vary; void where prohibited */
	};				/*  (Size in database varies - maximum is MAXLOCSTR+1) */

struct placedat {			/* Tidal computation data for a place (rec. 2) */
	long datum;			/* STET!  Floats are stored as longs in the code. */
	char meridian;		/* Meridian for tide computation - in half hours */
	unsigned char ndata;	/* Number of nonzero data */
	long datamask[CST_ARRAY_SIZE];	/* Bit array indicating which data are nonzero */
	};

/* Place data record continues with 		
	float amp[num_csts];
	float epoch[num_csts];	*/

struct placeoff {			/* Second record in case it's an offset */
	DBIndex primary;		/* Where it's offset from */
	int offsets[4];		/* Time offsets (in min) for h/l or fl/ebb/low sl/hi sl */
	int mulfactors[2];	/* Mult factors or additive factors (* 1/100) for h/l or fl/ebb */
	int addfactors[2];
	};

typedef enum {LENGTH, VELOCITY, BOGUS} unit_type;
typedef struct {
  char *name;
  char *abbrv;
  unit_type type;
  double conv_factor;
} unit;
#define NUMUNITS 9
extern unit known_units[NUMUNITS];
int GetUnits(char *);

extern struct placedata huge *places;
extern struct placeinfodata huge *placeinfo;
extern int nplaces, ncurrent;
extern int num_csts, num_nodes, num_epochs, first_year;
extern char lastline[linelen];
extern char lastloc[linelen];
extern int iscurrent;
extern long meridian;
extern double *cst_speeds, **cst_epochs, **cst_nodes;
extern double *loc_amp, *loc_epoch, *work;

/* Functions in tidelib that help stuff location data */
extern char *dupstr(char *);
extern int CheckCurrent(char *);
extern int InTheSquare(double, double);
extern char *ScreenLocation(char *);
extern int sslackcmp (char *, char *);

#define BADLAT 200.0
#define MISSINGLAT 300.0
extern double longit;
extern double latit;

extern char too_many_places;

extern void RoundLatLon(void);
extern int WriteLocationDatabase(void);		/* Routine to create location database */
extern void WriteLocationRecord(int, struct placeoff * );  /* Write one database record */
extern void WriteFactorsDatabase(void);

extern void ParseOffsetsFile(int);		/* Parse offsets from the indicated file */





