/*  Tidelib  Common functions used by XTide and Tide.
    Last modified 8/11/00

    THIS VERSION MODIFIED TO PRINT C SOURCE FOR HARMONIC DATA FOR
    COMPILATION INTO PALM PILOT VERSION OF XTIDE - Walt Bilofsky

    Further modified 7/00 to generate Palm databases directly. WB

    This program uses the harmonic method to compute tide levels.
    All of the data and constants are read in from the harmonics file.
    Please refer to README for more information.


    Copyright (C) 1995, 1996  David Flater.
    Also starring:  Jef Poskanzer; Rob Miracle; Geoff Kuenning;
    Dale DePriest.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


    The tide prediction algorithm used in this program was developed
    with United States Government funding, so no proprietary rights
    can be attached to it.  For more information, refer to the
    following publications:

    Manual of Harmonic Analysis and Prediction of Tides.  Special
    Publication No. 98, Revised (1940) Edition.  United States
    Government Printing Office, 1941.

    Computer Applications to Tides in the National Ocean Survey.
    Supplement to Manual of Harmonic Analysis and Prediction of Tides
    (Special Publication No. 98).  National Ocean Service, National
    Oceanic and Atmospheric Administration, U.S. Department of
    Commerce, January 1982.

*/

static int MAKEDB = 1;			/* Write .pdb file directly */
						/* Could be define - but this suppresses warnings */
#include "config.h"

#include <stdio.h>
#include <math.h>

#include <stdlib.h>
#include <signal.h>
// Not in Borland c #include <unistd.h>
#include <time.h>
#include <assert.h>

#include <string.h>
// #include <sys/time.h>
#include <ctype.h>
#include <malloc.h>

#ifdef PILOT
#include <Common.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>
#include "sysinclude.h"

#define COMPILEDATA   /* Pilot - always compile the data in */
#endif

#include "tidelib.h"
#include "arglib.h"
//#include "glyphs.h"
#include "timezone.h"

/* Abbreviations and filtering for harmonics file
 *
 * All abbreviations in the first array must be applied
 *
 */
char *filters[] = {
/***** GOT TO CHECK FOR DUPLICATE STRINGS INSTEAD OF REJECTING ON BASIS OF (N) BEC SOME ARE GOOD ****/
	"(expired",			NULL,
	"dubious",		NULL,
	"(unsure)",		NULL,
	" (2)",			"",		/* Mandatory substitutions */
	" (3)",			"",
	" (4)",			"",
	" (5)",			"",
	" (6)",			"",
	" (7)",			"",
	" (8)",			"",
	" (use with caution)",			"",  /* Only applies to Woods Hole any more */
	NULL, NULL  };

#ifdef _WINDOWS
#define M_PI 3.141592654
#define EOL "\r\n"
#define putchar(c)\
  printf("%c", (c))
#define puts(s)\
  printf(s);   \
  printf("\r\n")
int win_fprintf(FILE *stream,char const *fstring,...);
#define fprintf win_fprintf
#else
#define EOL "\n"
#endif

/* Compatibility by Dale DePriest */

#ifdef OS2
#include <float.h>
#endif /* OS2 */

/* Typo prevention */
#define DAYSECONDS 86400
#define HOURSECONDS 3600

/* Exported variables */

char hfile_name[MAXARGLEN+1], location[MAXARGLEN+1], *ppm = NULL,
  tzfile[MAXARGLEN+1];
char next_ht_text[17], next_ht_date[17], next_lt_text[17],
  next_lt_date[17], *geometry = NULL;
char *fgrise_color_arg = NULL, *fgfall_color_arg = NULL,
  *bg_color_arg = NULL, *fgtext_color_arg = NULL, *fgmark_color_arg = NULL,
  *fgmllw_color_arg = NULL, *fgmiddle_color_arg = NULL;
int tadjust = 0, utc = 0, list = 0, checkyear = 0, text = 0, skinny = 0,
  now = 0, graphmode = 0, httimeoff = 0, lttimeoff = 0, tstep = 180,
  middle = 0, mark = 0, mllw = 0, lines = 1, PPMWIDTH = 960,
  PPMHEIGHT = 300, hinc = 0, tinc = 0, loctz = 0, precise = 0,
  curonly = 0, toplines = 0, hincmagic = 0, calendar = 0, banner = 0;
time_t next_ht = 0, prev_ht = 0, faketime = 0, epoch = 0;
double amplitude, htleveloff = 0.0, ltleveloff = 0.0, DATUM, marklev, absmax, absmin;
int deftideindex = 0, defcurrindex = 0;

/* Externals - initialized here */
int nplaces = 0, ncurrent = 0;
int num_csts = 0, num_nodes = 0, num_epochs = 0, first_year = 0;
char lastline[linelen] = {0};
char lastloc[linelen] = {0};
int iscurrent = 0;
long meridian = 0;
double *cst_speeds, **cst_epochs, **cst_nodes;
double *loc_amp, *loc_epoch, *work;
struct placedata huge *places;
struct placeinfodata huge *placeinfo;
char *ref_stations[NREF_STATIONS];
int nref_stations = 0;

unit known_units[NUMUNITS] = {{"feet", "ft", LENGTH, 0.3048},
                              {"meters", "m", LENGTH, 1.0},
                              {"knots", "kt", VELOCITY, 1.0},
                              {"knots^2", "kt^2", BOGUS, 1.0},
	/* These are not in the harmonics file but are in the conversion menu */
					{"inches", "in", LENGTH, 0.3048/12.0 },
					{"mph", "mph", VELOCITY, 1.148},
					{"km/hr", "km/hr", VELOCITY, 1.852 },
					{"m/sec", "m/sec", VELOCITY, 1852.0/3600.0},
					{"ft/sec", "f/sec", VELOCITY, 1852.0/(3600.0*.0305)},
};

/* Local variables */

static int year = 0;
static long lineno = 0L;
static char orig_loc_name[linelen];
static char keep_silent;

char *dupstr(char *s);

/* Exit without core dump on assertion failures. */
void
sigabort ()
{
  exit (-1);
}

/* Avoid portability problems with strdup. */
char *
stradoop (char *src)
{
  char *dest = NULL;
  if (src) {
    assert ((dest = (char *) malloc ((int)(strlen (src)) + 1)) != 0);
    strcpy (dest, src);
  }
  return dest;
}
#ifndef COMPILEDATA   /* Don't allocate data if it's compiled in. */

/* Allocate tide-prediction storage except for node factors and epochs. */
static void
allocate_cst ( void )
{
  assert (num_csts);
  assert ((cst_speeds = (double *) malloc (num_csts * sizeof (double))) != 0);
  assert ((loc_amp = (double *) malloc (num_csts * sizeof (double))) != 0);
  assert ((loc_epoch = (double *) malloc (num_csts * sizeof (double))) != 0);
  assert ((work = (double *) malloc (num_csts * sizeof (double))) != 0);
}

static void
free_cst( void )
{
  assert (cst_speeds);
  assert (loc_amp);
  assert (loc_epoch);
  assert (work);
  free(cst_speeds);
  free(loc_amp);
  free(loc_epoch);
  free(work);
}

/* Allocate cst_nodes. */
static void
allocate_nodes (void)
{
  int a;
  assert (num_csts);
  assert (num_nodes);
  assert ((cst_nodes = (double **) malloc (num_csts * sizeof (double *))) != 0 );
  for (a=0;a<num_csts;a++)
    assert ((cst_nodes[a] = (double *) malloc (num_nodes * sizeof (double))) != 0);
}

static void
free_nodes(void)
{       
  int a;
  assert (num_csts);
  assert (num_nodes);
  assert (cst_nodes);
  for(a=0;a<num_csts;a++)
  {
    free(cst_nodes[a]);
  }                    
  free(cst_nodes);
}

/* Allocate cst_epochs. */
static void
allocate_epochs (void)
{
  int a;
  assert (num_csts);
  assert (num_epochs);
  assert ((cst_epochs = (double **) malloc (num_csts * sizeof (double *))) != 0);
  for (a=0;a<num_csts;a++)
    assert ((cst_epochs[a] = (double *) malloc (num_epochs * sizeof (double))) != 0);
}

static void
free_epochs(void)
{
  int a;
  assert(num_csts);
  assert(num_epochs);
  for(a=0;a<num_csts;a++)
  {
    free(cst_epochs[a]);
  }                    
  free(cst_epochs);
}

static int max_year;
/* Figure out max amplitude over all the years in the node factors table. */
/* This function by Geoffrey T. Dairiki */
static void
figure_amplitude ()
{
  int		i, a;
  double fakedatum, fakeamplitude;
max_year = 0;
	amplitude = 0;
      for (i = 0; i < num_nodes; i++)	
	{
	  double year_amp = 0.0;

	  for (a=0; a < num_csts; a++)
	      year_amp += loc_amp[a] * cst_nodes[a][i];
	  if (year_amp > amplitude) {
		max_year = i;
	      amplitude = year_amp;
	}	}
      assert(amplitude > 0.0);

      /* Figure other, related global values (DWF) */
      absmax = DATUM + amplitude;
      absmin = DATUM - amplitude;
      assert (absmax > absmin);

      if (absmax > 126.0 || absmin < -125.0)
        fprintf (stderr, "ERROR: ******* Tidal range out of bounds at %s\n",location);
}

/* Turn a time displacement of the form [-]HH:MM into the number of seconds. */
long
hhmm2seconds (char *hhmm)
{
  long h, m;
  char s;
  if (sscanf (hhmm, "%ld:%ld", &h, &m) != 2)
    barf (BADHHMM);
  if (sscanf (hhmm, "%c", &s) != 1)
    barf (BADHHMM);
  if (h < 0 || s == '-')
    m = -m;
  return h*HOURSECONDS + m*60;
}

/* Slackful strcmp; 0 = match.  It's case-insensitive and accepts a
   prefix instead of the entire string.  The second argument is the
   one that can be shorter. */
int
sslackcmp (char *a, char *b)
{
  int c, cmp, n;
  n = strlen (b);
  if ((int)(strlen (a)) < n)
    return 1;
  for (c=0;c<n;c++)
  {
    cmp = ((a[c] >= 'A' && a[c] <= 'Z') ? a[c] - 'A' + 'a' : a[c])
            -
          ((b[c] >= 'A' && b[c] <= 'Z') ? b[c] - 'A' + 'a' : b[c]);
    if (cmp)
      return cmp;
  }
  return 0;
}

/* Read a line from the harmonics file, skipping comment lines */
static int
next_line (FILE *fp, char linrec[linelen], int end_ok)
{
  char *p;
  static int iscomment = 0;
  again: do {
    ++lineno;
    if (!fgets (linrec, linelen, fp)) {
      if (end_ok)
        return 0;
      else {
        fprintf (stderr, "Unexpected end of harmonics file '%s' near line %ld\n",
        hfile_name,lineno);
        exit (-1);
      }
    }
    if (*linrec == '\n') {
	goto again;
	}
    assert (strlen(linrec) < linelen);
    strcpy(lastline, linrec);
    if (dumpdata) {
	/* Dumping data - look for lat/long as the last 2 lines in the comments only */
	if ((p = strstr(linrec,"!longitude:")) != NULL) 
		sscanf(p+strlen("!longitude:"),"%lf",&longit);
	else if ((p = strstr(linrec,"!latitude:")) != NULL) 
		sscanf(p+strlen("!latitude:"),"%lf",&latit);	
	else if (linrec[0] == '#')longit = latit = BADLAT; 
	}
    if (dumpdata || dumpfactors) {
	/* Keep track of and output comments */
	if (!strncmp(linrec,"#/*",3)) iscomment = 1;
	if (iscomment && !MAKEDB) printf("%s",linrec+1);
	if (!strncmp(linrec,"#*/",3)) iscomment = 0;
	}
  } while (linrec[0] == '#');
  return 1;
}
#endif  /* COMPILEDATA - End of routines not needed if compiling in data

/* Get rid of trailing garbage in linrec */
static char *
nojunk (char *linrec)
{
  char *a;
  a = &(linrec[strlen(linrec)]);
  while (a > linrec)
    if (*(a-1) == '\n' || *(a-1) == '\r' || *(a-1) == ' ')
      *(--a) = '\0';
    else
      break;
  return linrec;
}

/* Convert a struct tm in GMT back to a time_t. */
time_t
tm2gmt (struct tm *ht)
{
  time_t temp;
#if defined(sun) && !defined(SVR4) && !defined(__svr4__)
  /* Old SunOS */
  temp = timegm (ht);
  /* {
    struct tm *sunos;
    temp = timelocal (ht);
    sunos = localtime (&temp);
    temp += sunos->tm_gmtoff;
  } */
#else
  assert ((temp = mktime (ht)) != -1);
  /* Determine the local time zone offset and correct for it */
  /* Taking timezone from gettimeofday didn't work (always zero) */
#ifdef ultrix
 {
  /* Ultrix compatibility courtesy of Rob Miracle */
  struct tm *tzinfo;
  tzinfo = localtime (&temp);
  temp += tzinfo->tm_gmtoff;
 }
#else
  localtime(&temp);
  temp -= timezone;
#endif
#endif
  return temp;
}

/* Calculate time_t of the epoch. */
static void
set_epoch (void)
{
  struct tm ht;
  if (year < first_year || year >= first_year + num_epochs) {
    fprintf (stderr, "Tidelib:  Don't have equilibrium arguments for %d\n",
      year);
    barf (MISSINGYEAR);
  }
  ht.tm_year = year - 1900;
  ht.tm_sec = ht.tm_min = ht.tm_hour = ht.tm_mon = 0;
  ht.tm_mday = 1;
  ht.tm_isdst = 0;
  epoch = tm2gmt (&ht);
}

/* Attempt to load up the local time zone of the location.  Moof! */
/* Munged by Walt Bilofsky to generate a simple zone string for the Pilot */
int
change_time_zone (char *tz)
{
	int i;
	while (*tz == ':') ++tz;
/* See if it's mapped into one the Pilot database will know about */
	for (i = 0; TZequivs[i][0][0]; ++i) {
		if (!strcmp(tz, TZequivs[i][0])) tz = TZequivs[i][1];
		}
	for (i = 0; Zones[i].longname[0]; ++i) 
		if (!strcmp(tz,Zones[i].longname)) return i;
//			else {fprintf(stderr,"%s %s %d\n",tz,Zones[i].longname,strcmp(tz,Zones[i].longname)); exit(1); }
	for (i = 0; TZequivs[i][0][0]; ++i);
	TZequivs[i][0] = dupstr(tz);			/* Add to table */
	TZequivs[i][1] = TZequivs[0][1];
	TZequivs[i+1][0]= "";
	fprintf(stderr,"ERROR: Time zone not in table: <%s> at %s\n",tz,location);
	printf("ERROR: Time zone not in table: <%s> at %s\n",tz,location);
	return 0;
#if 0
  /* If our level of support is less than NEWZONEINFO, find the
     translation for the timezone string */
  if (tzlevel != NEWZONEINFO) {
    int index = 0;
    while (1) {
      if (subs[index][0] == NULL) {
        /* Not found.  Probably best not to issue a warning, since somebody
           might use something that will work. */
        break;
      }
      if (!strcmp (subs[index][0], tz)) {
        strcpy (tz, subs[index][tzlevel]);
        break;
      }
      index++;
    }
  }
#endif
}

int filterit(char *filters[], char *s, int oneflag) {
	/* Rewrite location string or reject certain locations */
	/* Return 1 for filtered, -1 for reject else 0 */
	/* Oneflag means (a) just perform one substitution, and (b) don't substitute
	   before a comma, since it's probably a state name */
	int a,b; char *p, *q;
	for (a = b = 0; filters[2*a]; ++a) {
	    p = s;
	    while ((p = strstr(p, filters[2*a])) != 0) {
		if (!filters[2*a+1]) { b = -1; break; }	/* NULL means reject */
		if (!oneflag && (q = strchr(s,',')) != 0 && q > p) {
			++p; continue; }			/* Don't sub state before , */
		strcpy(p, filters[2*a+1]);			/* else substitute */
		strcpy(p + strlen(filters[2*a+1]), p + strlen(filters[2*a]));
		b = 1;
		if (oneflag) return 1;
	    }	}
	return b;
}

/* free harmonics data */
void
free_data (void)
{
#ifndef COMPILEDATA
    free_nodes();
    free_epochs();
    free_cst();
#endif
}

char *dupstr(char *s) {
	char *p;
  	assert ((p = (char *) malloc ((1+strlen(s)) * sizeof (char))) != 0 );
	strcpy(p,s);
	return p;
	}

/* Check if this location is a current, and strip off "Current" if it is. */
int CheckCurrent(char *location) {
    char *p;
    if ((p = strstr (location, " Current")) != 0) {
	*p = 0;			/* Strip "Current" in location name */
	return 1;
	}
    return 0;
    }
/* Get code for the units */
int GetUnits(char *junk) {
	int kunits;
	for (kunits = 0; kunits < NUMUNITS;  ++kunits)
		if (!strcmp(junk,known_units[kunits].name)) break;
	return kunits;
	}

/* Round the lat and long, and compute the max and min enclosing */
double longit = BADLAT;
double latit = BADLAT;

void RoundLatLon(void) {
   if (!keep_silent) {
	if (latit > mlat) mlat = latit;
	if (latit < nlat) nlat = latit;
	if (longfrom >= -100 || longto <= 100) {	/* Regular rules */
		if (longit > mlon) mlon = longit;
		if (longit < nlon) nlon = longit; }
	 else {						/* Pacific/dateline rules */
		if (longit < 0 && longit > mlon) mlon = longit;
		if (longit >= 0 && longit < nlon) nlon = longit; }
	}
#define RDOFF (0.5/3600.0)
   latit += latit >= 0 ? RDOFF : -RDOFF;				/* Round off */
   longit += longit >= 0 ? RDOFF : -RDOFF;
   }

/* Tell if it's in the area we're looking for */
int InTheSquare(double latit, double longit) {
	/* Toss out if it's not in the area we're looking for */
	if (latit <= latfrom || latit > latto) return 0;
	if (longfrom > -100 || longto < 100) {	/* Normal rules */
		if (longit <= longfrom || longit > longto) return 0; }
	 else							/* Pacific rules - includes date line */
		if (longit > longfrom && longit <= longto) return 0;
	return 1;
}
/* Screen the location name string and lat/long to see if it will be included.
 * Apply mandatory filtering, and check if it is a current.
 * Return pointer to current-filtered name, or NULL if rejected.
 */

char *ScreenLocation(char *linrec) {
	static char curstrip[linelen];
//static int qq=0;
	char flag = 0;
	int i, j;
	keep_silent = 0;
	if (latit == BADLAT)				/* If we don't have valid data (i.e., */
		return NULL;				/*  skipping through a location) just return; */
	for (i = 0; i < nexclude; ++i)
		if (strstr(linrec,exclude[i]) != NULL) { flag = 1; goto keepit; }
	for (i = 0; i < nexcluden; ++i)
		if (strstr(linrec,excluden[i]) == NULL) { flag = 1; goto keepit; }
	for (i = 0; i < ninclude; ++i)	/* If included, skip the position screening */
		if (strstr(linrec,include[i]) != NULL) goto keepit;
	/* Toss out if it's not in the area we're looking for */
//printf("%d: Checking %s %lg %lg\n",++qq,linrec,latit,longit);
	if (!InTheSquare(latit, longit)) flag = 1;

	keepit:
	if (flag) {							/* Failed a test. */
		if (!IsRefStation(linrec, 0)) 		/* If not a required ref station */
			return NULL;				/*   kill it. */
		keep_silent = 1;					/* Otherwise just silence it. */
fprintf(stderr,"Note: Reference Station %s included silently\n",linrec);
printf("Note: Reference Station %s included silently\n",linrec);
		}
	strcpy(orig_loc_name, linrec);			/* Save original location name */
	/* Apply mandatory substitutions and look for strings that rule it out */
	while ((i = filterit(filters, linrec, 0)) > 0);
      if (i == -1) return NULL;
	 /* Is it a current? */
	strcpy (curstrip, linrec);
	iscurrent = CheckCurrent(curstrip);
	/* Check for duplicate location */
	for (j = 0; j < nplaces; ++j)
	   if (places[j].iscurrent == iscurrent && !stricmp(curstrip,places[j].location)) {
//		printf("Eliminated duplicate location: %s\n", curstrip);
		return NULL;
		}
	return curstrip;
	}

char IsRefStation(char *p,char add) {
	int i;
	for (i = 0; i < nref_stations; i++)
		if (!stricmp(p, ref_stations[i])) return 1;
	if (add) {
		assert (nref_stations < NREF_STATIONS);
		ref_stations[nref_stations++] = dupstr(p);
//printf("ADDED reference station %s\n",p);
		}
	return 0;
	}

/***
 *** 
 *** Load harmonics data 
 ***
 ***
*/
double mlat = -200, mlon = -200, nlat = 200, nlon = 200;

void
load_data (void)
{
#ifndef COMPILEDATA
  FILE *fp;
  char linrec[linelen], junk[80], *p;
  int a, b, kunits, maxloc, maxtz, tzindex, i;
  int j,k,m,realdat;
  double t;
  long l1, l2;

  if (latfrom > latto) { t = latto; latto = latfrom; latfrom = t; }
  if (longfrom > longto) { t = longto; longto = longfrom; longfrom = t; }

  if (MAKEDB) ParseOffsetsFile(0);				/* Get the list of offset locations */

  if (!(fp = fopen (hfile_name, "r"))) {
    fprintf (stderr, "Could not open harmonics file '%s'\n", hfile_name);
    barf (CANTOPENFILE);
  }
  fprintf(stderr, "Loading %s\n",hfile_name);
  next_line (fp, linrec, 0);
  assert (sscanf (linrec, "%d", &num_csts) == 1);
  if (dumpfactors & !MAKEDB) printf("static int num_csts = %d;\nstatic float cst_speeds[%d] = {\n",
	num_csts,num_csts);
  if (dumpfactors & !MAKEDB) fprintf(stderr,"\nOpened %s = %d factors\n",hfile_name,num_csts);
  allocate_cst ();
  /* Load constituent speeds */
  for (a=0;a<num_csts;a++) {
    next_line (fp, linrec, 0);
    assert(sscanf (linrec, "%s %lf", junk, &(cst_speeds[a])) == 2);
    cst_speeds[a] /= 3600.0;  /* Convert to degrees per second */
    if (dumpfactors & !MAKEDB) printf("%lg, ",cst_speeds[a]);
  }
  /* Get first year for nodes and epochs */
  next_line (fp, linrec, 0);
  assert (sscanf (linrec, "%d", &first_year) == 1);
  if (dumpfactors & !MAKEDB) {
	if (dumpyear == 0) dumpyear = first_year;
	printf("};\nstatic int first_year = %d;\n",dumpyear);
	}

  /* Load epoch table */
  next_line (fp, linrec, 0);
  assert (sscanf (linrec, "%d", &num_epochs) == 1);
  if (dumpfactors & !MAKEDB) {
	if (nyears == 0) nyears = num_epochs;
	printf("static int num_epochs = %d;\nstatic float cst_epochs[%d][%d] = { ",
		nyears,num_csts,nyears);
	}
  allocate_epochs ();
  for (a=0;a<num_csts;a++) {
    assert (fscanf (fp, "%s", linrec) == 1);
    ++lineno;
    for (b=0;b<num_epochs;b++) {
      assert (fscanf (fp, "%lf", &(cst_epochs[a][b])) == 1);
      if (dumpfactors & !MAKEDB  && b >= dumpyear - first_year && b < nyears + dumpyear - first_year) 
//		printf("%lg, ",cst_epochs[a][b]);
		/*** Now scale for cosine function here to save time ***/
		printf("%lg, ",cst_epochs[a][b] * (1024.0 / 90.0));
     }
     if (dumpfactors & !MAKEDB) printf("\n   ");
  }

  /* Sanity check */
  assert (fscanf (fp, "%s", linrec) == 1); ++lineno;
  assert (!strcmp (linrec, "*END*"));

  /* Load node factor table */
//next_line (fp, linrec, 0);  /* Skip lingering CR from *END* */
  next_line (fp, linrec, 0);
  assert (sscanf (linrec, "%d", &num_nodes) == 1);
  if (dumpfactors & !MAKEDB) 
    printf(" };\nstatic int num_nodes = %d;\nstatic float cst_nodes[%d][%d] = { ",
		nyears,num_csts,nyears);
  allocate_nodes ();
  for (a=0;a<num_csts;a++) {
    assert (fscanf (fp, "%s", linrec) == 1);  ++lineno;
    for (b=0;b<num_nodes;b++) {
      assert (fscanf (fp, "%lf", &(cst_nodes[a][b])) == 1);  
      if (dumpfactors & !MAKEDB && b >= dumpyear - first_year && b < nyears + dumpyear - first_year) 
		printf("%lg, ",cst_nodes[a][b]);
      }
	if (dumpfactors & !MAKEDB) printf("\n   ");
  }
  /* Sanity check */
  assert (fscanf (fp, "%s", linrec) == 1);  ++lineno;
  assert (!strcmp (linrec, "*END*"));
  if (dumpfactors & !MAKEDB) {
	printf("};\n");			/* Done dumping factors */
	exit(0);
	}
  if (dumpfactors & MAKEDB) {		/* Dump factors directly to file */
	WriteFactorsDatabase();
	exit(0);	
	}
  /* List locations and exit? */
  if (list) {
//next_line (fp, linrec, 0);  /* Skip lingering CR from *END* */
    while (next_line (fp, linrec, 1)) {
      nojunk (linrec);
      if (curonly)
        if (!strstr (linrec, "Current"))
          continue;
      printf ("%s", linrec);
      next_line (fp, linrec, 0);
      if (list == 2)
        printf ("\t%s\n", nojunk(linrec));
      else
        printf (EOL);
      for (a=0;a<num_csts+1;a++)
        next_line (fp, linrec, 0);
    }
    fclose (fp);
 #ifdef _WINDOWS
    return;
 #else
    exit (0);
 #endif
  }

/************************
 *
 * Load (and dump) harmonic constants for desired location
 *
 ************************/
if (dumpdata && !MAKEDB) {
	printf("/* Locations data begins here */\n\n#define NUM_CSTS %d\n",num_csts);
	printf("#define FIRST_YEAR %d\n#define NUM_YEARS %d\n",dumpyear,nyears);
	}
if (dumpdata && (!(places = (placedata *) halloc(MAXPLACES ,sizeof(struct placedata))) 
		|| !(placeinfo = (placeinfodata *) halloc(MAXPLACES ,sizeof(struct placeinfodata)))))
	barf(OTHER);
  while (next_line (fp, linrec, 1)) {
    nojunk (linrec);
    if (!dumpdata) {
	    if (curonly)
 	     if (!strstr (linrec, "Current"))
 	       continue;
	    if (sslackcmp (linrec, location))
		continyew:
 	     continue;
	}
    strcpy (lastloc, linrec);
    if (dumpdata) {
//printf("Checking %s\n",linrec);

	if (! (p = ScreenLocation(linrec))) latit = BADLAT;
	  else strcpy(location,p);			/* Save name with "current" stripped */ 
	}
    if (latit == BADLAT) continue;			/* Skipping all the lines in a location */
    /* Names, names:	orig_location_name is the original string from the file.
     *			location is the name with mandatory substitutions and "Current" stripped
     */

    /* Get meridian */
    next_line (fp, linrec, 0);
    meridian = hhmm2seconds (linrec);
    /* Get tzfile, if present */
    if (sscanf (nojunk(linrec), "%s %s", junk, tzfile) < 2)
      	strcpy (tzfile, ":UTC");
    tzindex = change_time_zone(tzfile);
    if (dumpdata) {
  	  meridian = (meridian / 60);
//  	  meridian = (meridian / 60) - Zones[tzindex].meridian;  WAS THIS WAY - DB CHANGE IN 2.0B4
 	  if (meridian % 30) {
		   /* No need for error message normally - meridian corrns are OK per Dave Flater */
		   fprintf(stderr,"Meridian corr'n. %ld min from %s at %s\n",meridian,Zones[tzindex].longname,location);
		   fprintf(stderr,"*** MERIDIAN OFFSET NOT A MULTIPLE OF 30 ***\n");
		   }
   	  meridian /= 30;
	  }
    if (!strncmp(Zones[tzindex].longname,"GMT",3)) /* Convert GMT offsets to a neg number */
		tzindex = Zones[tzindex].meridian / 30L - 27;
    /* Get DATUM */
    next_line (fp, linrec, 0);
    assert (sscanf (linrec, "%lf %s", &DATUM, &junk) ==2);
    if (dumpdata) 
		kunits = GetUnits(junk);
    assert (kunits < NUMUNITS);

    /* Get constituents */
    for (a=0;a<num_csts;a++) {
      char junk[80];
      next_line (fp, linrec, 0);
      assert (sscanf (linrec, "%s %lf %lf", junk, &(loc_amp[a]),
        &(loc_epoch[a])) == 3);
      assert (loc_amp[a] >= 0.0);
      }

    ncurrent += iscurrent;
    if (dumpdata) { /* If dumping, record location and keep on reading file */
	RoundLatLon();
	places[nplaces].location = dupstr(location);
	placeinfo[nplaces].origlocation = dupstr(orig_loc_name);
//printf("At %d: %s - %s\n",nplaces,placeinfo[nplaces].origlocation,placeinfo[1].origlocation );
	places[nplaces].iscurrent = iscurrent;
	places[nplaces].issilent = keep_silent;
	places[nplaces].latitude = latit;
	places[nplaces].longitude = longit;
	places[nplaces].meridian = meridian;
	places[nplaces].tzindex = tzindex;
	places[nplaces].datum = DATUM;
	places[nplaces].units = kunits;
	figure_amplitude ();
	placeinfo[nplaces].absmax = absmax;
	placeinfo[nplaces].absmin = absmin;
//printf("Tides from %lg to %lg in year %d at %s\n",absmin, absmax, max_year, places[nplaces].location);

	/* Pack the actual data, eliminating the zero entries */
	for (j = realdat = 0; j < num_csts; ++j) 				/* First count nonzero entries */
		if (loc_amp[j] != 0 || loc_epoch[j] != 0) ++realdat;
	for (j = 0; j < CST_ARRAY_SIZE; ) places[nplaces].datamask[j++] = 0;	/* Allocate memory for them */
	if (! (places[nplaces].amp = (datatype *) malloc (realdat * sizeof (datatype))) 
         || ! (places[nplaces].epoch = (datatype *) malloc (realdat * sizeof (datatype)))) {
		printf("Out of Memory - %d locations is too many!\n",nplaces); barf(OTHER); }
	places[nplaces].ndata = realdat;
	a = k = 0;
	m = 31;
	for (j = 0; j < num_csts; ++j) {
		if (loc_amp[j] != 0 || loc_epoch[j] != 0) {
			places[nplaces].amp[a] = loc_amp[j];
			places[nplaces].epoch[a] = loc_epoch[j];
			++ a;
			places[nplaces].datamask[k] |= 1L << m;		/* Set a bit to say it's real data */
			}
		if (--m < 0) {
			++k;
			m = 31;
			}
		}
      assert (a == realdat);
	/*
	**  Now write the current record on the temp data file
	*/
	WriteLocationRecord(nplaces, NULL);
	if (!iscurrent && *deftidename && strstr (location,deftidename)) {
		if (!MAKEDB) printf("#define DEF_TIDEPLACE %d\n",nplaces);
		deftideindex = nplaces;
		*deftidename = 0; }
	if (iscurrent && *defcurrname && strstr (location,defcurrname)) {
		if (!MAKEDB) printf("#define DEF_CURRPLACE %d\n",nplaces);
		defcurrindex = nplaces;
		*defcurrname = 0; }
	if (++nplaces >= MAXPLACES){				/* Don't abort yet - we may just be generating a list */
		if (!too_many_places) {
			fprintf(stderr,"***** FATAL ERROR:  MORE THAN %d PLACES*****\n",MAXPLACES);
			printf("***** FATAL ERROR:  MORE THAN %d PLACES*****\n",MAXPLACES);
			}
		too_many_places = 1;
		--nplaces;
		}
	continue;
	}
   fclose (fp);
   if (!dumpdata) return;
  }
 if (dumpdata) {
   if (!MAKEDB) {
	fprintf(stderr,"\n%d places, %d currents output\n",nplaces,ncurrent);
	printf("#define DB_NAM \"%s\"\n#define DB_ID %u\n",dbNAM,dbID);
	printf("#define LATFROM %ld\n#define LATTO %ld\n",
				(long)(mlat * 3600),(long)(nlat * 3600));
	printf("#define LONGFROM %ld\n#define LONGTO %ld\n",
				(long)(mlon * 3600),(long)(nlon * 3600));
	printf("#define NPLACES %d\n\n#ifdef MAIN\nchar *locations[%d];\nlong latitudes[%d] = {",nplaces,nplaces,nplaces,nplaces);
	for (a = 0; a < nplaces; ++a) printf(" %ld,",(long)(3600 * places[a].latitude));
	printf("}; \nlong longitudes[%d] = {",nplaces);
	for (a = 0; a < nplaces; ++a) printf(" %ld,",(long)(3600 * places[a].longitude));
	printf("};\nchar flags[%d] = {",nplaces);
	for (a = 0; a < nplaces; ++a) 
		printf(" %d,",places[a].units + (places[a].iscurrent ? 0x80 : 0));
	printf("}; \nfloat data[%d] = {",nplaces);
	for (a = 0; a < nplaces; ++a) printf(" %g,",places[a].datum);
	printf("}; \nchar tzones[%d] = {",nplaces);
	for (a = 0; a < nplaces; ++a) printf(" %d,",places[a].tzindex);
	printf("}; \nchar meridoff[%d] = {",nplaces);
	for (a = 0; a < nplaces; ++a) printf(" %ld,",places[a].meridian);
	printf("}; \n#endif\n\n#ifdef DAT1\nfloat amps[%d][%d] = { ",nplaces,num_csts);

/*** WARNING - THIS STUFF IS NOW PACKED SO THE PRINTING NO LONGER WORKS ****/

	for (b = 0; b < nplaces; ++b) {
		for (a = 0; a < num_csts; a++) printf(" %g,",places[b].amp[a]);
		printf("\n   ");
		}
	printf("}; \n#endif\n\n#ifdef DAT2\nfloat epochs[%d][%d] = { ",nplaces,num_csts);
	for (b = 0; b < nplaces; ++b) {
		for (a = 0; a < num_csts; a++) 
//			printf(" %g,",places[b].epoch[a]);
		/*** Now scale for cosine function here to save time ***/
			printf(" %g,",places[b].epoch[a] * (1024.0 / 90.0));
		printf("\n   ");
		}
	printf("};\n#endif\n\n#ifdef MAIN\nstatic void init_data() {");
	}
   for (a = 0; a < nplaces; ++a) {
		/* If too long, try to filter it */
		while (strlen(places[a].location) >= MAXLOCSTR && 
						filterit(fitfilters, places[a].location, 1) == 1);
		if (strlen(places[a].location) >= MAXLOCSTR) {
			fprintf(stderr,"Location string %d too long: %s\n",
				strlen(places[a].location) - MAXLOCSTR + 1,places[a].location);
				places[a].location [MAXLOCSTR-1] = 0; }
		if (!MAKEDB) {
			printf("locations[%d] = \"%s\";", a, places[a].location);
			if (places[a].iscurrent) printf("\t/* current */");
			printf("\n");
			}
		}
   if (!MAKEDB) {
	printf("}\n\nextern float epochs[%d][%d];\nextern float amps[%d][%d];\n",
		nplaces,num_csts,nplaces,num_csts);
	printf("\n#endif\n\n#define MAXLOCSTR %d\n#define MAXTZSTR %d\n#define NCURRENTS %d\n", 
			MAXLOCSTR, MAXTZSTR, ncurrent);

	fprintf(stderr,"Latitude range %g - %g\n",nlat,mlat);
	fprintf(stderr,"Longitude range %g - %g\n",nlon,mlon);
	}
   if (*deftidename) {
		fprintf(stderr,"ERROR: NO TIDE FOUND FOR %s\n",deftidename );
		printf("ERROR: NO TIDE FOUND FOR %s\n",deftidename ); }
   if (*defcurrname) {
		fprintf(stderr,"ERROR: NO CURRENT FOUND FOR %s\n",defcurrname );
		printf("ERROR: NO CURRENT FOUND FOR %s\n",defcurrname );
		}
   /**** If generating database on Windows, now go write the whole thing. *****/

   if (MAKEDB) {
	ParseOffsetsFile(1);
	WriteLocationDatabase();
	}
   exit(0);
   }
 if (curonly)
    fprintf (stderr, "Could not find current set '%s' in harmonics file\n",
      location);
  else
    fprintf (stderr, "Could not find location '%s' in harmonics file\n",
      location);
  exit (-1);
#endif    /* End conditional for loading data vs compiling it */
}

/* Check for errors in equilibrium arguments by inferring values from
   neighboring years and comparing with the stored values.  This
   function assumes that the values are normalized [0..360). */
void
check_epoch (void)
{
  double inferred, stored, error;
  int a, final_year;
  time_t prev_epoch, next_epoch;
  final_year = first_year + num_epochs - 1;
  if (checkyear < first_year || checkyear > final_year) {
    fprintf (stderr, "Tidelib:  Don't have equilibrium arguments for %d\n",
      checkyear);
    barf (MISSINGYEAR);
  }
  if (checkyear != first_year) {
    year = checkyear - 1;
    set_epoch ();
    prev_epoch = epoch;
  }
  if (checkyear != final_year) {
    year = checkyear + 1;
    set_epoch ();
    next_epoch = epoch;
  }
  year = checkyear;
  set_epoch ();
  puts ("#  Stored  Inferred  Error");
  for (a=0;a<num_csts;a++) {
    if (checkyear == first_year)
      inferred = fmod (cst_speeds[a] * (-(long)(next_epoch - epoch)) +
      cst_epochs[a][1], 360.0);
    else if (checkyear == final_year)
      inferred = fmod (cst_speeds[a] * (epoch - prev_epoch) +
      cst_epochs[a][final_year-1-first_year], 360.0);
    else {
      double t1, t2;
      t1 = fmod (cst_speeds[a] * (-(long)(next_epoch - epoch)) +
        cst_epochs[a][checkyear+1-first_year], 360.0);
      t2 = fmod (cst_speeds[a] * (epoch - prev_epoch) +
        cst_epochs[a][checkyear-1-first_year], 360.0);
      if (t1 < 0.0)
        t1 += 360.0;
      if (t2 < 0.0)
        t2 += 360.0;
      inferred = (t1 + t2) / 2.0;
      if (fabs (inferred - t1) > 90.0 || fabs (inferred - t2) > 90.0)
        inferred = fmod (inferred + 180.0, 360.0);
    }
    stored = cst_epochs[a][checkyear-first_year];
    if (inferred < 0.0)
      inferred += 360.0;
    error = fabs (stored - inferred);
    if (fabs (stored + 360.0 - inferred) < error)
      error = fabs (stored + 360.0 - inferred);
    if (fabs (stored - inferred - 360.0) < error)
      error = fabs (stored - inferred - 360.0);
    printf ("%d  %010.6f  %010.6f  %f\n", a, stored, inferred, error);
  }
}

/* Convert YYYY:MM:DD:HH:MM to time_t.  Don't know what happens to DST. */
time_t
parse_time_string (char *time_string)
{
  struct tm ht;
  time_t temp;
  if (sscanf (time_string, "%d:%d:%d:%d:%d",
    &(ht.tm_year), &(ht.tm_mon), &(ht.tm_mday), &(ht.tm_hour),
    &(ht.tm_min)) != 5)
    barf (BADTIMESTAMP);
  ht.tm_sec = 0;
  (ht.tm_mon)--;
  ht.tm_year -= 1900;
  ht.tm_isdst = -1;
  if ((temp = mktime (&ht)) == -1)
    barf (BADTIMESTAMP);
  return temp;
}

/* Return the year for a time_t.  This function gets called a lot and might
   be a performance bottleneck if gmtime is slow. */
int
yearoftimet (time_t t)
{
  return ((gmtime (&t))->tm_year) + 1900;
}

/* Make a reasonable choice for hinc for a given amplitude. */
void
pick_hinc (void)
{
  if (amplitude > 30.0)
    hinc = 10;
  else if (amplitude > 15.0)
    hinc = 5;
  else if (amplitude > 8.0)
    hinc = 2;
  else
    hinc = 1;
}

/* Figure out normalized multipliers for constituents for a particular
   year.  Save amplitude for drawing unit lines. */
static void
figure_multipliers ( void )
{
  int a;
  amplitude = 0.0;
  if (year < first_year || year >= first_year + num_nodes) {
    fprintf (stderr, "Tidelib:  Don't have node factors for %d\n", year);
    barf (MISSINGYEAR);
  }
  for (a=0;a<num_csts;a++)
    amplitude += (work[a] = loc_amp[a] * cst_nodes[a][year-first_year]);
  assert (amplitude != 0.0);
  for (a=0;a<num_csts;a++)
    work[a] /= amplitude;
  if (hincmagic)
    pick_hinc ();
}

/* Doh...  Forgot to convert to radians. */
static double
cos_degrees (double x)
{
  return cos (x * M_PI / 180.0);
}

/* Re-initialize for a different year */
void
happy_new_year (int new_year)
{
  year = new_year;
  figure_multipliers ();
  set_epoch ();
}

/* Calculate the normalized tide (-1.0 .. 1.0) for a given time. */
double
time2tide (time_t t)
{
  double tide = 0.0;
  int a, new_year = yearoftimet (t);
  if (new_year != year)
    happy_new_year (new_year);
  for (a=0;a<num_csts;a++)
    tide += work[a] *
      cos_degrees (cst_speeds[a] * ((long)(t - epoch) + meridian) +
      cst_epochs[a][year-first_year] - loc_epoch[a]);
  return tide;
}

/* Calculate the denormalized tide. */
double
time2atide (time_t t)
{
  return time2tide (t) * amplitude + DATUM;
}

/* Next high tide, low tide, transition of the mark level, or some
   combination.
       Bit      Meaning
        0       low tide
        1       high tide
        2       falling transition
        3       rising transition
*/
int
next_big_event (time_t *tm)
{
  double p, q;
  int delta = precise ? 1 : 60;
  int flags = 0, slope = 0;
  p = time2atide (*tm);
  *tm += delta;
  q = time2atide (*tm);
  *tm += delta;
  if (p < q)
    slope = 1;
  while (1) {
    if ((slope == 1 && q < p) || (slope == 0 && p < q)) {
      /* Tide event */
      flags |= (1 << slope);
    }
    /* Modes in which to return mark transitions: */
    /*    -text (no -graph)   */
    /*    -graph (no -text)   */
    /*    -ppm                */
    if (mark && ((text && !graphmode) || (!text && graphmode)
    || ppm))
      if ((p > marklev && q <= marklev) || (p < marklev && q >= marklev)) {
        /* Transition event */
        if (p < q)
          flags |= 8;
        else
          flags |= 4;
        if (!(flags & 3)) {
          /* If we're incredibly unlucky, we could miss a tide event if we
             don't check for it here:

                             . <----   Value that would be returned
                        -----------    Mark level
                      .           .   
          */
          p = q;
          q = time2atide (*tm);
          if ((slope == 1 && q < p) || (slope == 0 && p < q)) {
            /* Tide event */
            flags |= (1 << slope);
          }
        }
      }
    if (flags) {
      *tm -= delta;
      /* Don't back up over a transition event, but do back up to where the
         tide changed if possible.  If they happen at the same time, then
         we're off by a minute on the tide, but if we back it up it will
         get snagged on the transition event over and over. */
      if (flags < 4)
        *tm -= delta;
      return flags;
    }
    p = q;
    q = time2atide (*tm);
    *tm += delta;
  }
}

/* Used by do_timestamp. */
static int
is_ascii (char a)
{
  if (a >= ' ' && a <= '~')
    return 1;
  return 0;
}

/* Generate a timestamp in the appropriate style. */
void
do_timestamp (char buf[13], struct tm *t)
{
  int a; char b[5];
  if (skinny)
    strftime (buf, 13, "%H%M", t);
  else if (tadjust)
    strftime (buf, 13, "%I:%M %p", t);
  else if (utc)
    strftime (buf, 13, "%H:%M UTC", t);
  else
    strftime (buf, 13, "%I:%M %p %Z", t);
  if (!skinny && buf[0] == '0') {
    if (text && !ppm && !banner)
      buf[0] = ' ';
    else
#if defined(sun) && !defined(SVR4) && !defined(__svr4__)
    /* Old SunOS */
    {
      for (a=0;a<12;a++)
        buf[a] = buf[a+1];
    }
#else
      memmove (buf, buf+1, 12);
#endif
	}
  /* Some time zone files don't fill in the time zone properly.  We
     can get blanks, or worse yet, unterminated binary garbage. */
  buf[12] = '\0';
  for (a=11;a>=8;a--)
    if (!is_ascii (buf[a]))
      buf[a] = '\0';
  nojunk (buf);
  if (precise) { strcat(buf,":"); strcat(buf,itoa(t->tm_sec,b,10)); }

}

/* Convert a genuine time_t to a specially mangled struct tm. */
struct tm *
tmtime (time_t t)
{
  if (tadjust) {
    t += tadjust;
    return gmtime (&t);
  } else if (utc)
      return gmtime (&t);
    else
      return localtime (&t);
}

/* Find the previous hour-transition from a specified time. */
time_t
prev_hour (time_t t)
{
  struct tm ttm;
  time_t temp;
  if (utc || tadjust) {
    ttm = *(gmtime (&t));
    ttm.tm_sec = ttm.tm_min = 0;
    temp = tm2gmt (&ttm) - (tadjust % HOURSECONDS);
  } else {
    ttm = *(localtime (&t));
    ttm.tm_sec = ttm.tm_min = 0;
    temp = mktime (&ttm);
  }
  if (temp <= t - HOURSECONDS)
    temp += HOURSECONDS;
  if (temp > t)
    temp -= HOURSECONDS;
  assert (temp > t - HOURSECONDS && temp <= t);
  return temp;
}

/* Find the previous day-transition from a specified time. */
time_t
prev_day (time_t t)
{
  struct tm ttm;
  time_t temp;
  if (utc || tadjust) {
    ttm = *(gmtime (&t));
    ttm.tm_sec = ttm.tm_min = ttm.tm_hour = 0;
    temp = tm2gmt (&ttm) - tadjust;
  } else {
    ttm = *(localtime (&t));
    ttm.tm_sec = ttm.tm_min = ttm.tm_hour = 0;
    temp = mktime (&ttm);
  }
  if (temp <= t - DAYSECONDS)
    temp += DAYSECONDS;
  if (temp > t)
    temp -= DAYSECONDS;
  assert (temp > t - DAYSECONDS && temp <= t);

  /* If there has been a DST change, the time that we have calculated
     can be wrong.  Test for the error, and correct it when found.  It
     is assumed that the DST change is by an integer number of hours
     and by no more than 2 hours. */
  if (!(utc || tadjust)) {
    if (localtime(&temp)->tm_hour) {
      int day;
      temp += 3 * HOURSECONDS;
      day = localtime(&temp)->tm_yday;
      while (localtime(&temp)->tm_yday == day)
        temp -= HOURSECONDS;
      temp += HOURSECONDS;
    }
  }

  assert (temp > t - HOURSECONDS * 26L && temp <= t);
  return temp;
}

/* Find the Sunday on which to start a calendar. */
time_t
sunday_month (time_t t)
{
  struct tm ttm;
  int almost_done = 0;
  ttm = *(tmtime(t));
  if (ttm.tm_mday == 1)
    almost_done = 1;
  while (ttm.tm_wday || !almost_done) {
    t = prev_day (t-1);
    ttm = *(tmtime(t));
    if (ttm.tm_mday == 1)
      almost_done = 1;
  }
  return t;
}

/* Find the next day-transition from a specified time, ASSUMING that the
   parameter passed in is ALREADY on a DAY BOUNDARY. */
time_t
increment_day (time_t t)
{
  time_t iday;
  long day;
  struct tm *ttm;

  /* These are easy because there's no DST to worry about */
  if (utc || tadjust)
    return t + DAYSECONDS;

  /* Local time is the troublesome case!  The following code will work
     provided that the DST change is by an integer number of hours
     and by no more than 2 hours. */

  day = localtime(&t)->tm_yday;
  iday = t + HOURSECONDS * 22L;
  ttm = localtime(&iday);
  while (ttm->tm_yday == day) {
    iday += HOURSECONDS;
    ttm = localtime(&iday);
  }

  /* Insure that we were not caught by a non-integer DST change. */
  assert (!(ttm->tm_min));

  return iday;
}

/* Determine next big event and set up text string for write_high_tide.
   Returns are same as next_big_event. */
int
update_high_tide ( void )
{
  struct tm *t;
  char *tidedate, *tidetime;
  int event_type;
  prev_ht = next_ht;
  event_type = next_big_event (&next_ht);
  if ((event_type & 2) || graphmode || text || ppm || banner) {
    tidedate = next_ht_date;
    tidetime = next_ht_text;
  } else {
    tidedate = next_lt_date;
    tidetime = next_lt_text;
  }
  if (event_type & 2)
    t = tmtime (next_ht + httimeoff);
  else
    t = tmtime (next_ht + lttimeoff);
  do_timestamp (tidetime, t);
  if (skinny)
    strftime (tidedate, 11, "%m-%d", t);
  else
    strftime (tidedate, 11, "%Y-%m-%d", t);
  return event_type;
}


/**********************************************************************/

/* Calendar and plain text mode */

#define MAXTIDESPERDAY 25
#define DAYWIDTH 7
/* static char week[DAYWIDTH*7][MAXTIDESPERDAY*2]; */
static char **week = NULL;

/* Set one tide in the calendar week being constructed */
static void
do_one_calendar_tide (int dayofweek, int tidenumber, int event_type)
{
  int x;
  char desc, temp[DAYWIDTH];
  double adjust = 0.0;
  assert (tidenumber < MAXTIDESPERDAY - 1);  /* Leave room for dates */
  assert (dayofweek < 7);
  if (iscurrent) {
    switch (event_type) {
    case 1:
      desc = 'E';
      break;
    case 2:
      desc = 'F';
      break;
    case 4:
    case 8:
      desc = 'S';
      break;
    default:
      assert (0);
    }
  } else {
    switch (event_type) {
    case 1:
      desc = 'L';
      adjust = ltleveloff;
      break;
    case 2:
      desc = 'H';
      adjust = htleveloff;
      break;
    case 4:
      desc = 'F';
      break;
    case 8:
      desc = 'R';
      break;
    default:
      assert (0);
    }
  }
  week[dayofweek*DAYWIDTH][2+tidenumber*2] = desc;
  for (x=0;x<(int)(strlen(next_ht_text));x++)
    week[dayofweek*DAYWIDTH+x+1][2+tidenumber*2] = next_ht_text[x];
  /* This will die horribly if the amplitude is too big. */
  sprintf (temp, "%0.2f", adjust + time2atide (next_ht));
  for (x=0;x<(int)(strlen(temp));x++)
    week[dayofweek*DAYWIDTH+x][3+tidenumber*2] = temp[x];
}

/* Do one day for a calendar. */
static void
do_calendar_day (time_t start, time_t stop, int dayofweek, int *maxline)
{
  int tidenumber = 0, event_type, x;
  next_ht = start - 180; /* Don't miss tides that happen at midnight */
  while (1) {
    event_type = update_high_tide ();
    if (next_ht < start)
      continue;
    if (next_ht >= stop)
      break;

    /* Add date stamp from first tide (hope we get one) */
    if (tidenumber == 0)
      for (x=0;x<(int)(strlen(next_ht_date));x++)
        week[dayofweek*DAYWIDTH+x][0] = next_ht_date[x];

    /* Refer to list_tides to understand why this is like this. */
    if ((event_type & 4) && (event_type & 1))
      do_one_calendar_tide (dayofweek, tidenumber++, 4);
    if ((event_type & 8) && (event_type & 2))
      do_one_calendar_tide (dayofweek, tidenumber++, 8);
    if (event_type & 1)
      do_one_calendar_tide (dayofweek, tidenumber++, 1);
    if (event_type & 2)
      do_one_calendar_tide (dayofweek, tidenumber++, 2);
    if ((event_type & 4) && !(event_type & 1))
      do_one_calendar_tide (dayofweek, tidenumber++, 4);
    if ((event_type & 8) && !(event_type & 2))
      do_one_calendar_tide (dayofweek, tidenumber++, 8);

    if (3+(tidenumber-1)*2 > *maxline)
      *maxline = 3+(tidenumber-1)*2;
  }
}

/* Do one week for a calendar. */
static time_t
do_calendar_week (time_t start)
{
  int x, y, maxline = 0, dayofweek;
  time_t stop;
  for (x=0;x<DAYWIDTH*7;x++)
    for (y=0;y<MAXTIDESPERDAY*2;y++)
      week[x][y] = ' ';
  start = prev_day (start);
  stop = increment_day (start);
  for (dayofweek=0;dayofweek<7;dayofweek++) {
    do_calendar_day (start, stop, dayofweek, &maxline);
    start = stop;
    stop = increment_day (stop);
  }
  for (y=0;y<=maxline;y++) {
    for (x=0;x<DAYWIDTH*7;x++)
      putchar (week[x][y]);
    printf (EOL);
  }
  return start;
}

/* Do a calendar. */
void
do_calendar (void)
{
  time_t start = sunday_month (faketime);
  int month = tmtime(faketime)->tm_mon, looper;

  if (!week) {
    /* Allocate storage */
    assert ((week = (char **) malloc (DAYWIDTH*7 * sizeof (char *))) != 0);
    for (looper=0;looper<DAYWIDTH*7;looper++) {
     assert ((week[looper] = (char *) malloc (MAXTIDESPERDAY*2*sizeof(char))) != 0);
    }
  }

  puts (location);
  if (mark && !iscurrent)
    printf ("Mark level:  %f\n", marklev);
  printf (EOL);
  printf ("Sunday Monday Tuesda Wednes Thursd Friday Saturday\n");
  do {
    printf (EOL);
    start = do_calendar_week (start);
  } while (tmtime(start)->tm_mon != (month+1)%12);
}

/* List several days of high and low tides to standard output.  This
got ugly because of the need to make things come out in the right order
with the enhanced mark code. */
void
list_tides (void)
{
  int a, event_type,p;
  char *high, *low, *rise, *fall;
  long l;
  next_ht = faketime;
  if (iscurrent) {
    high =        "  Max Flood";
    low =         "    Max Ebb";
    rise = fall = "Slack Water";
  } else {
    high =        "  High Tide";
    low =         "   Low Tide";
    rise =        "     Rising";
    fall =        "    Falling";
  }
  puts (location);
  if (mark && !iscurrent)
    printf ("Mark level:  %f\n", marklev);
  if (httimeoff || lttimeoff || htleveloff != 0.0 || ltleveloff != 0.0)
    printf ("Offsets:  hloff=%f  htoff=%dmin  lloff=%f  ltoff=%dmin\n",
    htleveloff, httimeoff/60, ltleveloff, lttimeoff/60);
  if (text < 1)
    text = 21;
  p = precise ? 5 : 2;
  for (a=0;a<text;a++) {
    event_type = update_high_tide ();
    if ((event_type & 4) && (event_type & 1)) {
      printf ("%s:  %s %s  %0.*f\n", fall, next_ht_date, next_ht_text,
      p,time2atide (next_ht));
    }
    if ((event_type & 8) && (event_type & 2)) {
      printf ("%s:  %s %s  %0.*f\n", rise, next_ht_date, next_ht_text,
      p,time2atide (next_ht));
    }
    if (event_type & 1) {
      printf ("%s:  %s %s  %0.*f\n", low, next_ht_date, next_ht_text,
      p,ltleveloff + time2atide (next_ht));
    }
    if (event_type & 2) {
      printf ("%s:  %s %s  %0.*f\n", high, next_ht_date, next_ht_text,
      p,htleveloff + time2atide (next_ht));
    }
    if ((event_type & 4) && !(event_type & 1)) {
      printf ("%s:  %s %s  %0.*f\n", fall, next_ht_date, next_ht_text,
      p,time2atide (next_ht));
    }
    if ((event_type & 8) && !(event_type & 2)) {
      printf ("%s:  %s %s  %0.*f\n", rise, next_ht_date, next_ht_text,
      p,time2atide (next_ht));
    }
//    if (precise) for (l = - 10; l <= 10; ++l)
//	printf("%ld: %0.8f\n",l,time2atide(next_ht+l));
  }
}


/**********************************************************************/


/* Banner mode */
#define BANNERWIDTH 68
#define BANNERSTART 11
void
do_banner (void)
{
  int a, b, c, markchar, middlechar, mllwchar;
  struct tm *t;
  time_t curtime;
  char bdate[6], btime[5], buf[80], tbuf[80];
  if (text < 1)
    text = 21;
  curtime = next_ht = faketime;
  puts (location);
  /* Initialize the amplitude. */
  happy_new_year (yearoftimet (faketime));
  if (mark) {
    if (!iscurrent)
      printf ("Mark level:  %f\n", marklev);
    markchar = (int)((double)BANNERWIDTH * (((marklev - DATUM) / amplitude)
           * 0.5 + 0.5)) + BANNERSTART;
    if (markchar < BANNERSTART || markchar > 78)
      mark = 0;
  }
  if (middle)
    middlechar = BANNERWIDTH / 2 + BANNERSTART;
  if (mllw) {
    mllwchar =  (int)((double)BANNERWIDTH * (((-DATUM) / amplitude)
           * 0.5 + 0.5)) + BANNERSTART;
    if (mllwchar < BANNERSTART || mllwchar > 78)
      mllw = 0;
  }
  buf[79] = '\0';
  update_high_tide ();
  for (a=0;a<text;a++) {
    t = tmtime (curtime);
    strftime (btime, 5, "%H%M", t);
    strftime (bdate, 6, "%m-%d", t);
    b = (int)((double)BANNERWIDTH * (time2tide(curtime) * 0.5 + 0.5))
      + BANNERSTART;
    sprintf (buf, "%s %s ", bdate, btime);
    if (iscurrent) {
      for (c=BANNERSTART;c<79;c++)
        buf[c] = ' ';
      if (b < markchar)
        for (c=b;c<markchar;c++)
          buf[c] = '*';
      else
        for (c=markchar+1;c<=b;c++)
          buf[c] = '*';
    } else {
      for (c=BANNERSTART;c<=b;c++)
        buf[c] = '*';
      for (c=b+1;c<79;c++)
        buf[c] = ' ';
    }
    if (mark)
      buf[markchar] = '|';
    if (middle)
      buf[middlechar] = '|';
    if (mllw)
      buf[mllwchar] = '|';
    if (curtime+(tstep>>1) >= next_ht) {
      sprintf (tbuf, "%s %0.2f", next_ht_text, time2atide (next_ht));
      strncpy (buf+BANNERSTART+1, tbuf, strlen (tbuf));
      while (curtime+(tstep>>1) >= next_ht)
        update_high_tide ();
    }
    puts (buf);
    curtime += tstep;
  }
}


/**********************************************************************/


/* ASCII graph mode was done by Jef Poskanzer. */
/* Generate an ASCII graph of the tides. */
#define TEXTHEIGHT 14
#define TTEXTHEIGHT 10
static char textmap[TEXTWIDTH*TEXTHEIGHT];
static void t_setchr (int x, int y, char chr) {
  if (x < 0 || y < 0 || x >= TEXTWIDTH || y >= TEXTHEIGHT)
    return;
  textmap[y*TEXTWIDTH+x] = chr;
}
static char t_getchr (int x, int y) {
  if (x < 0 || y < 0 || x >= TEXTWIDTH || y >= TEXTHEIGHT)
    return '\0';
  return textmap[y*TEXTWIDTH+x];
}
static double t_tide2wl (double tide) {
  return ((double)TTEXTHEIGHT) * (-tide) * 0.5 + (double)TTEXTHEIGHT * 0.5;
}
static double t_wl2tide (int y) {
  return -2.0*(y - TTEXTHEIGHT * 0.5) / ((double)TTEXTHEIGHT);
}
static void t_putstr (int x, int y, char *s) {
  int a;
  for (a=0;a<(int)(strlen(s));a++)
    t_setchr (x+a, y, s[a]);
}
static void t_center_text (int x, int y, char *text)
{
  int l = strlen (text);
  t_putstr (x - l / 2, y, text);
}
static void t_write_tide (int slope, int x)
{
  if (slope) {
    t_center_text (x, TTEXTHEIGHT, next_ht_date);
    t_center_text (x, TTEXTHEIGHT+1, next_ht_text);
  } else {
    t_center_text (x, TTEXTHEIGHT+2, next_ht_date);
    t_center_text (x, TTEXTHEIGHT+3, next_ht_text);
  }
}
void
tide2ascii (void)
{
  int x, y, event_type, midwl;
  time_t start = faketime, this_time = time(NULL);

  puts (location);
  if (mark && !iscurrent)
    printf ("Mark level:  %f\n", marklev);

  /* Set the background. */
  for (x=0;x<TEXTWIDTH;x++)
    for (y=0;y<TEXTHEIGHT;y++)
      t_setchr (x, y, ' ');

  /* Initialize the amplitude. */
  happy_new_year (yearoftimet (faketime));

  /* Draw day separators. */
  if (lines) {
    time_t day;
    for (day=prev_day(start);day<start+TEXTWIDTH*tstep;
                             day=increment_day(day)) {
      x = (int)((day - start) / tstep + 0.5);
      for (y=0;y<TEXTHEIGHT;y++)
        t_setchr (x, y, '|');
    }
  }

  /* Draw the tides. */
  midwl = (int)(t_tide2wl (-DATUM / amplitude) + 0.5);
  for (x=0;x<TEXTWIDTH;x++) {
    int yy = (int)(t_tide2wl (time2tide (faketime)) + 0.5);
    if (iscurrent) {
      if (yy < midwl)
        for (y=midwl-1;y>=yy;y--)
          t_setchr (x, y, '*');
      else
        for (y=midwl+1;y<=yy;y++)
          t_setchr (x, y, '*');
    } else {
      for (y=TTEXTHEIGHT-1;y>=yy;y--)
        t_setchr (x, y, '*');
    }
    faketime += tstep;
  }

  /* Extra lines */
  if (mark) {
    y = (int) (t_tide2wl ((marklev - DATUM) / amplitude) + 0.5);
    for (x=0;x<TEXTWIDTH;x++)
      if (t_getchr (x, y) == '*' || iscurrent)
        t_setchr (x, y, '-');
  }
  if (middle) {
    y = (int) (t_tide2wl (0.0) + 0.5);
    for (x=0;x<TEXTWIDTH;x++)
      if (t_getchr (x, y) == '*')
        t_setchr (x, y, '-');
  }
  if (mllw) {
    y = (int) (t_tide2wl (-DATUM / amplitude) + 0.5);
    for (x=0;x<TEXTWIDTH;x++)
      if (t_getchr (x, y) == '*')
        t_setchr (x, y, '-');
  }

  if (now) {
    /* X marks the spot */
    x = (int)((this_time - start) / tstep);
    y = (int) (t_tide2wl (time2tide (this_time)) + 0.5);
    t_setchr (x, y, '+');
  }

  /* Scrawl timestamps */
  next_ht = start;
  event_type = update_high_tide ();
  while (next_ht < start + TEXTWIDTH*tstep) {
    t_write_tide ((event_type & 2), (int)((next_ht - start) / tstep + 0.5));
    event_type = update_high_tide ();
  }

  /* Write output. */
  for (y=TTEXTHEIGHT;y<TTEXTHEIGHT+2;y++) {
    printf("[");
    for (x=0;x<TEXTWIDTH;x++)
      putchar(t_getchr(x,y));
    printf("]\n");
  }
  for (y=0;y<TTEXTHEIGHT;y++) {
    double temp;
    printf("[");
    for (x=0;x<TEXTWIDTH;x++)
      putchar(t_getchr(x,y));
    temp = t_wl2tide(y) * amplitude + DATUM;
    /* Chop off minus sign for super-low tides to save space */
    if (temp <= -10.0)
      temp = -temp;
    printf("]%0.1f\n", temp);
  }
  for (y=TTEXTHEIGHT+2;y<TTEXTHEIGHT+4;y++) {
    printf("[");
    for (x=0;x<TEXTWIDTH;x++)
      putchar(t_getchr(x,y));
    printf("]\n");
  }
}



/**********************************************************************/


/* Generate a PPM graph of the tides.  While we're at it, re-invent the
   wheel.
*/
static int BGR = 135;
static int BGG = 206;
static int BGB = 235;
static int FGR = 0;
static int FGG = 0;
static int FGB = 255;
static int EBR = 46;
static int EBG = 139;
static int EBB = 87;
static int TXR = 0;
static int TXG = 0;
static int TXB = 0;
static int MKR = 255;
static int MKG = 0;
static int MKB = 0;
static int MDR = 255;
static int MDG = 255;
static int MDB = 0;
static int MLR = 255;
static int MLG = 255;
static int MLB = 255;
#define margin 0.05
unsigned char *pixmap;
/* Colors must be specified in the form rgb:hh/hh/hh, where hh is a 2-byte
   hexidecimal number. */
static void p_parsecolor (int *r, int *g, int *b, char *col) {
  if (col) {
    if (sscanf (col, "rgb:%x/%x/%x", r, g, b) != 3)
      barf (BADCOLORSPEC);
  }
  if (!((*r >= 0 && *r <= 255)&&(*g >= 0 && *g <= 255)&&
  (*b >= 0 && *b <= 255)))
    barf (BADCOLORSPEC);
}
static void p_parseallcolors (void)
{
  p_parsecolor (&FGR, &FGG, &FGB, fgrise_color_arg);
  p_parsecolor (&EBR, &EBG, &EBB, fgfall_color_arg);
  p_parsecolor (&TXR, &TXG, &TXB, fgtext_color_arg);
  p_parsecolor (&MKR, &MKG, &MKB, fgmark_color_arg);
  p_parsecolor (&MLR, &MLG, &MLB, fgmllw_color_arg);
  p_parsecolor (&MDR, &MDG, &MDB, fgmiddle_color_arg);
  p_parsecolor (&BGR, &BGG, &BGB, bg_color_arg);
}
static void p_setval (int x, int y, int c, int val) {
  if (x < 0 || y < -12 || x >= PPMWIDTH || y >= PPMHEIGHT)
    return;
  pixmap[((y+12)*PPMWIDTH+x)*3+c] = (unsigned char)val;
}
static int p_getval (int x, int y, int c) {
  if (x < 0 || y < -12 || x >= PPMWIDTH || y >= PPMHEIGHT)
    return 0;
  return (int) (pixmap[((y+12)*PPMWIDTH+x)*3+c]);
}
static void p_setpix (int x, int y, int r, int g, int b) {
  p_setval (x, y, 0, r);
  p_setval (x, y, 1, g);
  p_setval (x, y, 2, b);
}
static int p_interp (int a, int b, double x) {
  return (int)((double)a + x * (double)(b-a) + 0.5);
}
static double p_tide2wl (double tide) {
  return ((double)PPMHEIGHT * (1.0 - 2.0 * margin)) * (-tide) * 0.5
  + (double)PPMHEIGHT * 0.5;
}
static double p_wl2tide (int y) {
  return -2.0*(y - PPMHEIGHT * 0.5) / ((double)PPMHEIGHT *
  (1.0 - 2.0 * margin));
}
static void p_putglyph (int x, int y, char glyph) {
  int base = (int)(glyph - ' '), xl, yl;
  assert (base >= 0 && base < 95);
  base *= 7;
  for (yl=0;yl<11;yl++)
    for (xl=0;xl<7;xl++)
#if 0
      if (glyphs[yl][base+xl])
     p_setpix (x+xl, y+yl, TXR, TXG, TXB);
#endif
	x; y;
}

static void p_putstr (int x, int y, char *s) {
  int a;
  for (a=0;a<(int)(strlen(s));a++)
    p_putglyph (x+a*7, y, s[a]);
}
static void p_center_text (int x, int y, char *text)
{
  int l = strlen (text);
  p_putstr (x - 7 * l / 2, y, text);
}
static void p_write_high_tide (int x)
{
  p_center_text (x, 1, next_ht_date);
  p_center_text (x, 13, next_ht_text);
}
static void p_write_trans_time (int x)
{
  int bottom = PPMHEIGHT - 19;
  if (tinc)
    bottom -= 12;
  p_center_text (x, bottom-12, next_ht_date);
  p_center_text (x, bottom, next_ht_text);
}
static void p_draw_depth_lines (int *tmax, int *tmin)
{
  int a, x, y;
  if (lines) {
    *tmin = 30;
    *tmax = PPMHEIGHT - 14;
    if (mark)
      *tmax -= 24;
    if (tinc)
      *tmax -= 12;
    *tmin = (int) (p_wl2tide (*tmin) * amplitude + DATUM);
    *tmax = (int) (ceil (p_wl2tide (*tmax) * amplitude + DATUM));
    for (a=*tmax;a<=*tmin;a++) {
      if (amplitude > 30.0 && (a % 10))
        continue;
      y = (int) p_tide2wl (((double)a - DATUM) / amplitude);
      if (hinc) {
        if (!(a % hinc)) {
          char temp[5];
          int l;
          sprintf (temp, "%d", abs(a));
          l = strlen (temp);
          for (x=0;x<PPMWIDTH-7*l-2;x++)
            p_setpix (x, y, TXR, TXG, TXB);
          continue;
        }
      }
      for (x=0;x<PPMWIDTH;x++)
        p_setpix (x, y, TXR, TXG, TXB);
    }
  }
}
void
tide2ppm (char *outfile)
{
  FILE *fp;
  int x, y, event_type, a, tmax, tmin, midwl;
  time_t start = faketime, this_time = time(NULL);

  assert (sizeof (unsigned char) == 1);
#if 0
  assert (pixmap = (unsigned char *) malloc (3 * PPMWIDTH *
    (PPMHEIGHT + 12)));
#endif

  p_parseallcolors ();

  /* Set the background. */
  for (x=0;x<PPMWIDTH;x++)
    for (y=-12;y<PPMHEIGHT;y++)
      p_setpix (x, y, BGR, BGG, BGB);

  /* Initialize the amplitude before drawing the depth lines. */
  happy_new_year (yearoftimet (faketime));

  if (!toplines)
    p_draw_depth_lines (&tmax, &tmin);

  /* Draw the tides. */
  midwl = (int)(p_tide2wl (-DATUM / amplitude));
  for (x=0;x<PPMWIDTH;x++) {
    double i, yy = p_tide2wl (time2tide (faketime));
    if (iscurrent) {
      if ((int)yy < midwl) {
        for (y=midwl-1;y>(int)yy;y--)
       p_setpix (x, y, FGR, FGG, FGB);
        /* Anti-aliasing */
        i = yy - floor (yy);
        p_setval (x, (int)yy, 0, p_interp (FGR, p_getval (x, (int)yy, 0), i));
        p_setval (x, (int)yy, 1, p_interp (FGG, p_getval (x, (int)yy, 1), i));
        p_setval (x, (int)yy, 2, p_interp (FGB, p_getval (x, (int)yy, 2), i));
      } else {
        for (y=midwl+1;y<=(int)yy;y++)
       p_setpix (x, y, EBR, EBG, EBB);
        /* Anti-aliasing */
        i = yy - floor (yy);
        p_setval (x, (int)yy+1, 0, p_interp (p_getval (x, (int)yy+1, 0), EBR,
        i));
        p_setval (x, (int)yy+1, 1, p_interp (p_getval (x, (int)yy+1, 1), EBG,
        i));
        p_setval (x, (int)yy+1, 2, p_interp (p_getval (x, (int)yy+1, 2), EBB,
        i));
      }
    } else {
      for (y=PPMHEIGHT-1;y>(int)yy;y--)
     p_setpix (x, y, FGR, FGG, FGB);
      /* Anti-aliasing */
      i = yy - floor (yy);
      p_setval (x, (int)yy, 0, p_interp (FGR, p_getval (x, (int)yy, 0), i));
      p_setval (x, (int)yy, 1, p_interp (FGG, p_getval (x, (int)yy, 1), i));
      p_setval (x, (int)yy, 2, p_interp (FGB, p_getval (x, (int)yy, 2), i));
    }
    faketime += tstep;
  }

  if (toplines)
    p_draw_depth_lines (&tmax, &tmin);

  /* Label unit lines */
  if (hinc) {
    for (a=tmax;a<=tmin;a++) {
      if (!(a % hinc)) {
        char temp[5];
        int l;
        y = (int) p_tide2wl (((double)a - DATUM) / amplitude);
        sprintf (temp, "%d", abs(a));
        l = strlen (temp);
        p_putstr (PPMWIDTH - 7 * l, y-4, temp);
      }
    }
  }

  /* Extra lines */
  if (mark) {
    y = (int) p_tide2wl ((marklev - DATUM) / amplitude);
    for (x=0;x<PPMWIDTH;x++)
      p_setpix (x, y, MKR, MKG, MKB);
  }
  if (middle) {
    y = (int) p_tide2wl (0.0);
    for (x=0;x<PPMWIDTH;x++)
      p_setpix (x, y, MDR, MDG, MDB);
  }
  if (mllw) {
    y = (int) p_tide2wl (-DATUM / amplitude);
    for (x=0;x<PPMWIDTH;x++)
      p_setpix (x, y, MLR, MLG, MLB);
  }

  /* Draw tick marks */
  if (lines) {
    time_t hour, prevday;
    prevday = prev_day (start);
    for (hour=prev_hour(start);hour<=start+PPMWIDTH*tstep+HOURSECONDS;
    hour+=HOURSECONDS) {
      x = (hour - start) / tstep;
      for (y=PPMHEIGHT-1;y>PPMHEIGHT-8;y--)
        p_setpix (x, y, TXR, TXG, TXB);
      if (tinc) {
        if (!(((hour - prevday) / 3600) % tinc)) {
          char buf[17];
          do_timestamp (buf, tmtime (hour));
          if (buf[1] == ':')
            buf[1] = '\0';
          else
            buf[2] = '\0';
          p_center_text (x, PPMHEIGHT-19, buf);
        }
      }
    }
    for (hour=prevday;hour<=start+PPMWIDTH*tstep+HOURSECONDS;
                              hour=increment_day(hour)) {
      x = (hour - start) / tstep;
      for (y=PPMHEIGHT-1;y>PPMHEIGHT-8;y--) {
        p_setpix (x-1, y, TXR, TXG, TXB);
        p_setpix (x+1, y, TXR, TXG, TXB);
      }
    }
  }

  if (now) {
    int looper;
    /* X marks the spot */
    x = (int)((this_time - start) / tstep);
    y = (int) p_tide2wl (time2tide (this_time));
    for (looper=-4;looper<=4;looper++) {
      p_setpix (x+looper, y, TXR, TXG, TXB);
      p_setpix (x, y+looper, TXR, TXG, TXB);
    }
  }

  /* Scrawl timestamps */
  next_ht = start;
  event_type = update_high_tide ();
  while (next_ht < start + PPMWIDTH*tstep) {
    x = (int)((next_ht - start) / tstep);
    if (event_type & 3)
      p_write_high_tide (x);
    if (event_type & 12) {
      p_write_trans_time (x);
      for (y=PPMHEIGHT-1;y>=PPMHEIGHT-7;y--)
        p_setpix (x, y, MKR, MKG, MKB);
    }
    event_type = update_high_tide ();
  }

  /* Identify the pixmap */
  p_center_text (PPMWIDTH>>1, -11, location);

  /* Write output. */
  if (strcmp (outfile, "-")) {
#ifdef OS2
    if (!(fp = fopen (outfile, "wb")))
#else
    if (!(fp = fopen (outfile, "w")))
#endif
      barf (CANTOPENFILE);
  }
  else
    fp = stdout;
  fprintf (fp, "P6\n%d %d\n255\n", PPMWIDTH, PPMHEIGHT+12);
  fwrite (pixmap, 3, PPMWIDTH*(PPMHEIGHT+12), fp);
  fclose (fp);
  free (pixmap);
}
