/* tidedb.c:  Pilot version of xtide tide table program
 *	Database routines.

Tide Tool - A tide and current prediction program for the USR PalmPilot
Copyright (c) 1997-2003 Walt Bilofsky

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <PalmOS.h>
#include <PalmCompatibility.h> 
#include <string.h>

#include "sysinclude.h"
#include "tide.h"
#include "tidelib.h"

struct DBdescriptor TideDB = { { NULL } };
UInt32 TideAppID = 'Ttbl';
#define TideDBType   'Ttbl'					/* Was Data until 2.1b2 */
#define TideOldDBType   'Data'				/* Old db type - for error detection */
UInt32 dbcreator;

struct Prefstruct Preferences = { 
	DBTYPE_USER,		/* Database type */
	PREFVERSTG,
	1, 1,				/* Default to first time through */
	3, 0, 			/* default units */
#if LANG == LANG_FRENCH
	t_17h01,				/* French: default time display to 17h01 */
#else
	t5_01pm,				/*   otherwise 5:01 pm */
#endif
	0,				/* Don't ignore DST */
	0L, 0L,			/* initial today is not a day we will encounter */
	true, false, false, true,/* Initially display today, and tide, and list format */
	true,				/* Display coefficients */
	-1, -1, -1,			/* and we don't know any default places */
	false, 			/*  or any recent places */
	unknownColor, 		/* Color preference has not yet been selected */
	false, 			/* User has not indicated an alert tide level */
	LANG==LANG_FRENCH?1:0,	/* Default units to ft for English, m. for French */
	0.0,				/* and no level */
	0, {0},			/* No recent places */
	{ F_NONE },			/* "Find" selection method not chosen yet */
	{ 0 },			/* All databases enabled */
	{ 0 },			/* No notices given */
	};

/* Handles for record that holds *cst_speeds, *cst_nodes, *cst_epochs */
static struct DBHandle CstSpeedsHandle = { { NULL } };
static struct DBHandle CstEpochsHandle = { { NULL } };
static struct DBHandle CstNodesHandle = { { NULL } };
static struct DBHandle PlaceHandle = { { NULL } };
static struct DBHandle PlaceDatHandle = { { NULL } };
static struct DBHandle PlaceOffHandle = { { NULL } };
static void UnlockHand(struct DBHandle *);
static void DBNotify(struct LocDBHdr *);

int nPlaceDBs = 0;
int nEnabledDBs = 0;
int num_places = 0;
int num_currents = 0;
int num_offsets = 0;
int num_offcurrents = 0;
struct DBdescriptor PlaceDBs[NPLACEDBS];
struct DBHandle PlaceHdrHandles[NPLACEDBS];
struct LocDBHdr *PlaceHdrs[NPLACEDBS];
static LocalID PlaceIDs[NPLACEDBS];
static int this_placedb;

static LocalID TideArgsID;
static long args_created;
static UInt CardNo;

/*
 *	Package to provide database functions independent of whether database is
 *	in RAM or on a card (VFS);
 */

static Boolean VFS_present = false;			/* If expansion card present */
#define VFS_PATH "/PALM/Launcher/"			/* Where files are kept on expansion card */
static char *VFSPathname(int , char *);
static DmSearchStateType state;

void HandleUnlock(struct DBHandle *h) {		/* Unlock or release a database record */
	if (h->in_ram) MemHandleUnlock(h->p.handle);
	    else MemHandleFree(h->p.handle);
	}

static void CloseDatabase(struct DBdescriptor *h) {		/* Close a database */
	if (h->volume == IN_RAM) DmCloseDatabase(h->ref.ram);
	    else VFSFileClose(h->ref.card);
	}

/*
 * 	Open the next database.
 *	If it exists, set dbd to the database descriptor and hand to the handle descriptor of record 0.
 *	Returns database id if in RAM, ??? if in VFS, else 0.
 *	If newsearch = 1, start a new search.  If -1, abandon search.
 *	Searchstate:  1: New search; 2: searching RAM dbs; 3: Need another VFS volume;
 *		4: Good volume, need another file.
 */
static UInt32 volIterator;
static UInt16 volRefNum;
static FileRef dirRef;
static UInt32 fileIterator;
static enum DB_searchstate { stop_search = -1, continue_search, 
		  start_search, searching_ram, 
		  start_card_search, search_next_card, search_next_file } srchstate;
#define PATH_LEN 100
static char pathname[PATH_LEN];			/* Name of last DB opened on card */

static LocalID OpenNextDatabase(UInt16 newsearch, struct DBdescriptor *dbd, char *filename) {
    	LocalID id;
	FileInfoType fileInfo;
	Char name[PATH_LEN];
	int i;
	UInt32 creator;

      fileInfo.nameP = name;
	fileInfo.nameBufLen = PATH_LEN - 1;
	switch (newsearch) {
	    case stop_search:				/* Abort search, close current database */
		if (dirRef) VFSFileClose (dirRef);
		CloseDatabase(dbd);
		return 0;
	    case start_search:				/* Start a new search (full, or on card only) */
	    case start_card_search:
		srchstate = newsearch;
		dirRef = 0;
		}

	if (srchstate <= searching_ram) {
	    if (!DmGetNextDatabaseByTypeCreator(srchstate == start_search, &state, TideDBType, 
				TideAppID, false, &CardNo, &id)) {	/* Got a database in RAM */
		srchstate = searching_ram;				/* No longer the first search */
		DmDatabaseInfo(CardNo, id, filename, NULL, NULL, NULL, NULL, NULL,
			NULL, NULL, NULL, NULL, &dbcreator);	/* Get the filename */
		dbd->volume = IN_RAM;
		dbd->ref.ram = DmOpenDatabase ( CardNo, id, dmModeReadOnly);
 		return id;
		}
	    srchstate = start_card_search;
	    }
	if (srchstate == start_card_search) {
		volIterator = vfsIteratorStart;
		volRefNum = vfsInvalidVolRef;
		srchstate = search_next_card;
	   	}
	/* Now see if the VFS has any databases */
	if (!VFS_present) return 0;
	while (1) {
		if (srchstate == search_next_card) {		/* Need another volume */
 			if (volIterator == vfsIteratorStop ||
				VFSVolumeEnumerate(&volRefNum, &volIterator)) return 0;
			StrCopy ( name, VFS_PATH );	/* Open the directory */
			if (VFSFileOpen( volRefNum, name, vfsModeRead, &dirRef )) 
				continue;			/* If not there, abandon the volume */
			srchstate = search_next_file;
			fileIterator = vfsIteratorStart;
			}
									/* Next file on same volume */
		if (fileIterator == vfsIteratorStop ||
				VFSDirEntryEnumerate( dirRef, &fileIterator, &fileInfo )) {
			VFSFileClose (dirRef);			/* If no more files */
			srchstate = search_next_card;		/*  close directory and go on */
			continue;
			}
		i = strlen(fileInfo.nameP);
		if (i < 5 || StrCaselessCompare(".pdb",&fileInfo.nameP[i-4])) 
			continue;				/* If not a pdb file, continue */
		StrCopy(pathname, VFS_PATH);
		strcat(pathname, name);			/* Construct path name */
//printf("%s: ",name);
		if ((i = VFSFileOpen (volRefNum, pathname, vfsModeRead, &(dbd->ref.card)))) {
//printf("Can't open: error %d\n",i); SysTaskDelay(200);
			continue;				/* Open the file if possible */
			}
		if (VFSFileDBInfo (dbd->ref.card, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
				NULL, NULL, NULL, &creator, NULL) || creator != TideAppID) {
//printf("Bad creator: %lx\n",creator); SysTaskDelay(200);
			VFSFileClose (dbd->ref.card);	/* If not a Tide Tool DB */
			continue;				/*  close it and get next file */
			}
//printf("OK\n",i); SysTaskDelay(50);
		dbd->volume = volRefNum;		/* If Tide Tool database */
		return 1;					/*  return success */
	}	}

void SavePreferences() {
	Preferences.Ptoday = LocalDateNow;
	PrefSetAppPreferences(TideAppID,0,0,&Preferences,sizeof(Preferences),true);
    	}

/* Returns whether something has changed.  
 * Calling routine is responsible for reloading databases.
 */
Boolean DBEnable(int index, Boolean enable) {
	UInt32 mask;
	int n;
	Boolean old;
	old = DBEnabled(index);
	n = PlaceHdrs[index]->db_ID;			/* Convert from index to ID */
	mask = 1L << (n & 31);
	n = n/32;
	Preferences.PDBDisabled[n] &= ~mask;
	if (!enable) Preferences.PDBDisabled[n] |= mask;
	return (old != enable);
	}

/* 
 * If there is a notice associated with the database, post it up for the user.
 */
static void DBNotify(struct LocDBHdr *h) {
	UInt32 mask;
	int n, notice;
	if (!(notice = h->db_notice)) return;
	n = h->db_ID;					/* Convert from index to ID */
//printf("Notification %d\n",n); SysTaskDelay(50);
	mask = 1L << (n & 31);
	n = n/32;
	if (Preferences.PDBNotified[n] & mask) 	/* If previously notified, */
		return;					/*  do nothing. */
	Preferences.PDBNotified[n] |= mask;		/* Otherwise remember we did this, */
	FrmHelp(notice);					/*  and post the notice. */
	}

/* 
 Tide Factors database:
   Record 0 = six ints: first_year, num_epochs, num_csts, num_nodes, then the
			time created (as a long).  (AFAIK, num_epochs == num_nodes)
   Record 1 = cst_speeds[num_csts]
   Then record 2*n+2 = cst_epochs[n][0 ... num_csts] for epoch n
        record 2*n+3 = cst_nodes[n][0 ... num_csts] for epoch n
 Locations database:
   Record 0 = header (see tidelib.h)
   Remainder of records are:
	Record n+1: struct place for nth place (lat/long, current flag and name)
	Record n+2: data for the place
 User database:
   Record 0 = preferences and data from last run of program (see tidelib.h)
   Records 1-NEVENTS = DisEvents array of displayed tide/current tables (see tidedis.c)
*/

/* Helper function:
   While opening database files, check that the number of factors is the same in all.
*/
static Boolean ck_num_csts(int num) {
	if (num_csts) {
		if (num != num_csts) {
			barf(INCOMPATBASE);
			return 1;
		}	}
	  else num_csts = num;
	assert (num_csts < MAX_CSTS);
	return 0;
	}

Char dbfilename[33] = {0};							/* Info for any error message */

Boolean OpenDatabase(void)				/* Initialize databases */
{							/* Called once in program startup */
  int *ip, i, def_current = -1;
  UInt16 u;
  LocalID id;
  struct DBHandle handle;
  enum DB_searchstate newsrch;
  union {
	struct LocDBHdr *rec0;
	struct Prefstruct *pp;
	struct ArgsDBHeader *hp;
	} hdr;
  struct DBdescriptor dbref;
  struct Prefstruct prefs;
  UInt32 version_dummy;
  UInt8 dbnrs[NPLACEDBS];
  DBIndex coeff, plp;
  char goodname[PATH_LEN];
  FileRef fileref;

startallover:
  nPlaceDBs = nEnabledDBs = num_places = num_currents = num_offsets = 
		num_offcurrents = num_csts = 0;
  TideArgsID = 0;
  coeff = coeff_PlaceIndex = -1;
  VFS_present = (FtrGet(sysFileCExpansionMgr, expFtrIDVersion, &version_dummy) == errNone &&
		 FtrGet(sysFileCVFSMgr, vfsFtrIDVersion, &version_dummy) == errNone);
//if (!VFS_present) { printf("No VFS! %x %x\n",(int)FtrGet(sysFileCExpansionMgr, expFtrIDVersion, &version_dummy) == errNone,
//(int)FtrGet(sysFileCVFSMgr, vfsFtrIDVersion, &version_dummy) ); SysTaskDelay(100); }
//VFS_present = true;
  if (!DmGetNextDatabaseByTypeCreator(true, &state, TideOldDBType, TideAppID,
		false, &CardNo, &id)) {				/* Check for old database files */
	DmDatabaseInfo(CardNo, id, dbfilename, NULL, NULL, NULL, NULL, NULL,
		NULL, NULL, NULL, NULL, NULL);		/* Get info for error message */
	barf(OLDDBASE);						/*   of type data */
	return 1;						/*   and tell user to delete them */
	}
/*
 * Open all database files with the Tide Tool app ID.
*/
   newsrch = start_search;
   while ((id = OpenNextDatabase(newsrch, &dbref, dbfilename))) {
	newsrch = continue_search;
	hdr.rec0 = QueryRecord(&handle, &dbref, 0);			/* Get record 0 */
  	ErrFatalDisplayIf(!hdr.rec0,T_NO_ZERO_RECORD);
	switch (i = hdr.rec0->db_type) {
		default: 
			barf(i <= 0 ? BADDBASE : i > DBTYPE_FIRST ? BADDBASE : OLDDBASE); 
		badreturn:
			HandleUnlock(&handle);
			OpenNextDatabase(stop_search, &dbref, NULL);
			return 1;
		case DBTYPE_ARGS:				/* Argument database - set global params */
			if (TideArgsID) { barf(TOOMANYARGDBS); 
					HandleUnlock(&handle);
					CloseDatabase(&dbref);
					break; }
 			TideArgsID = id;
			TideDB = dbref;
			ip = hdr.hp->params;
			first_year = *ip++;
			num_epochs = *ip++;
			if (ck_num_csts(*ip++)) return 1;
			num_nodes = *ip++;
			args_created = hdr.hp->created;
//printf("Type %d ID %ld ref %d - %d yrs %d csts\n",i,id,dbref.volume,num_epochs,num_csts); SysTaskDelay(300);
			HandleUnlock(&handle);
			break;
		case DBTYPE_LOC:				/* One of location databases - add to array */
			for (i = 0; i < nPlaceDBs; ++i)
			    if (dbnrs[i] == hdr.rec0->db_ID) 
				barf(DUPLOCDB);			/* Warn of duplicate dbs */
			if (nPlaceDBs == NPLACEDBS) { 
				barf(TOOMANYLOCS); goto badreturn; }
			PlaceHdrs[nPlaceDBs] = hdr.rec0;
			PlaceHdrHandles[nPlaceDBs] = handle;
			PlaceIDs[nPlaceDBs] = id;
			dbnrs[nPlaceDBs] = hdr.rec0->db_ID;
			PlaceDBs[nPlaceDBs] = dbref;
			nPlaceDBs++;
			if (ck_num_csts(hdr.rec0->db_num_csts))	/* Verify databases match */	
					return 1;
			if (coeff_PlaceIndex == -1 && 		/* See if coefficients needed */
					hdr.rec0->db_coeff != -1) {	/*  and are in this database */
				coeff = hdr.rec0->db_coeff;
				}
			/*  Because filenames on card can vary depending on how they got there,
			 * 	we can't easily delete or beam them.  So check if a database on
			 *	card has the canonical name, and if not, try to rename it that way.
			 */
			if (PlaceDBs[nPlaceDBs-1].volume != IN_RAM 	/* If name of card file isn't canonical: */
					&& strcmp(VFSPathname(nPlaceDBs-1, goodname),pathname)) {
			   if (!VFSFileOpen (volRefNum, goodname, vfsModeRead, &fileref)) {
				VFSFileClose(fileref);				/* If good name exists, slink off. */
//printf("Could not rename %s - %s exists.\n",pathname,goodname); SysTaskDelay(400);
				}
			    else {							/* Otherwise, */
				CloseDatabase(&dbref);				/*  close the current database/ */
				VFSPathname(-nPlaceDBs, goodname);
				if (VFSFileRename (volRefNum,		/*  and rename it. */
						pathname, goodname))		/* If rename doesn't work, */
					barf(CANTRENAMEDB);			/*  got to abort, or this */
				OpenNextDatabase(stop_search, &dbref, NULL); /*  would loop forever. */
//printf("Renamed %s to %s\n",pathname,goodname); SysTaskDelay(200);
				CloseDatabases();				/* If it succeeds, close up */
				goto startallover;				/*  and start all over. */
			}	}
			break;
		}
#if 0		/* This should be unnecessary now that we block the conduit backup */
		SetBackup(0, CardNo, id);			/* Clear backup bit.  BEAR TRAP: If we ever want to */
#endif
	}							/*  back up location databases, an attribute must be */
  if (!TideArgsID) { barf(NODATABASE); return 1; }		/*  changed in generating it.  See build.cpp. */
   u = sizeof(Preferences);
   i = PrefGetAppPreferences(TideAppID,0,&prefs,&u,true);
   if (i != noPreferenceFound)  {			/* If there was a saved preference */
	if (u != sizeof(Preferences) || strcmp(prefs.ver_stg,PREFVERSTG)) 
		barf(LOSTPREFS);				/* If old version, ignore it */
	   else MemMove(&Preferences, &prefs, sizeof(Preferences));  /* else load prefs */
   	}
  if (nPlaceDBs == 0) { 
	barf(NOLOCNBASE);	
	return 1; }

  if (Preferences.Pfirstime) {			/* If first time in program */
	FrmHelp(TTWarning);				/* warn the user */
	Preferences.Pfirstime = 0;
	}

  for (i = 0; i < nPlaceDBs; ++i)
	if (DBEnabled(i)) {
		num_places += PlaceHdrs[i]->db_num_places;
		num_currents += PlaceHdrs[i]->db_num_currents;
		num_offsets += PlaceHdrs[i]->db_num_offsets;
		num_offcurrents += PlaceHdrs[i]->db_num_curroffsets;
		if (num_currents && def_current == -1) def_current = 
				PlaceHdrs[i]->db_default_current_place | 
					(PlaceHdrs[i]->db_ID << ID_SHIFT);
		DBNotify(PlaceHdrs[i]);				/* Post a notice if the DB wants it */
		nEnabledDBs++;
		}

  /* If not the same database, or first time through, set preferences from database */
//printf("Last place %x default place %x coeff %x\n",Preferences.Plastplace,PlaceHdrs[0]->db_default_tide_place,coeff); SysTaskDelay(300);
  if (Preferences.Plastplace == -1) {
	i = PlaceHdrs[0]->db_ID << ID_SHIFT;
	Preferences.Plasttide = Preferences.Plastplace = 
						i | PlaceHdrs[0]->db_default_tide_place;
	Preferences.Piscurrent = 0;				/* Start with tide */
	Preferences.Pdistoday = 1;				/* Display today */
	Preferences.Plastcurrent = -1;
	}
  if (Preferences.Plastcurrent == -1) 			/* If there are currents, this guar-*/
	Preferences.Plastcurrent = def_current;		/*  antees a valid default */
  plp = Preferences.Plastplace;
  if (coeff != -1) {						/* If we need coefficient data */
	GetPlaceData(coeff);
	SwapCoefficientData();					/*   load it up before loading a loc'n */
	coeff_swapped = false;					/*   and mark this as the "normal" state. */
	}
  GetPlaceData(plp);						/* LOad the current location. */
  return 0;
  }

/* Swap the current place data with the data for the French tide coefficient location */

#define swapf(f,g) { float t; t = f; f = g; g = t; }
#define swapl(f,g) { long t; t = f; f = g; g = t; }
void SwapCoefficientData() {
	int i; double d;
	for (i = 0; i < num_csts; i++) {
		swapf(loc_amp[i],coeff_amp[i]);
		swapf(loc_epoch[i],coeff_epoch[i]);
		}
	swapl(tmeridian,coeff_tmeridian);
	i = coeff_PlaceIndex; coeff_PlaceIndex = PlaceIndex; PlaceIndex = i;
	d = coeff_datum; coeff_datum = DATUM; DATUM = d;
	coeff_swapped = !coeff_swapped;
	figure_multipliers();				/* Re-initialize tide computation */
	}

int GetDBEntry(int dbid) {
	int db;
	for (db = 0; db < nPlaceDBs; ++db) 			/* Find dbase with that id */
		if (PlaceHdrs[db]->db_ID == dbid) return db;
	return -1;
	}

int GetPlaceData(DBIndex new_place) {
	int i, j, k;
	long l = 0;
	float *amp, *epoch;
	if (new_place == PlaceIndex) return 0;
//printf("GetPlaceData(%x) - last was %x\n",new_place,PlaceIndex); SysTaskDelay(50);
	UnlockHand(&PlaceHandle);	/* Unlock previous records */
	UnlockHand(&PlaceDatHandle);
	UnlockHand(&PlaceOffHandle);
	PlaceOff = NULL;

	i = 2 * (new_place & ID_MASK) + 1;
	this_placedb = GetDBEntry(new_place >> ID_SHIFT);
	if (this_placedb == -1 || !(Place = QueryRecord(&PlaceHandle, &PlaceDBs[this_placedb],i))) {
		barf(BADLOCNBASE);
		return 1; }

	getdat:
	if (!(PlaceDat = QueryRecord(&PlaceDatHandle , &PlaceDBs[this_placedb],i + 1))) {
		barf(BADLOCNBASE);
		return 1; }

	/* Check if it's an offset */
	if (!(PlaceOffHandle.p.handle) && (Place->flags & P_ISOFFSET)) {	/* If an offset */
		PlaceOff = (void *)PlaceDat;				/* Save offset info */
		PlaceOffHandle = PlaceDatHandle;
		i = 2 * (PlaceOff->primary & ID_MASK) + 1;
		goto getdat;
		}

	latitude = UnpackLat(Place);					/* Unpack latitude and longitude */
	longitude = UnpackLon(Place);
	unit = P_UNITS & Place->flags;
//if (Place->flags & P_COEFF) { printf("Loc'n needs coefficient!\n"); SysTaskDelay(50); }
	iscurrent = (P_ISCURRENT & Place->flags) != 0;
	DATUM = PlaceDat->datum;
	absmax = 0.01 * Place->absmax;
	absmin = 0.01 * Place->absmin;
	/* Unpack floating data to restore the zero entries */
	amp = PlaceDat->amp;
	epoch = amp + PlaceDat->ndata;
	memset(loc_amp, 0, sizeof(loc_amp));
	memset(loc_epoch, 0, sizeof(loc_epoch));
	for (i = k = 0; i < num_csts; ++i) {
		if (! (i & 31)) l = PlaceDat->datamask[k++];
		if (l < 0) {						/* If the mask bit is on */
			loc_amp[i] = *amp++;				/* this data item is nonzero */
			loc_epoch[i] = *epoch++;
			}
		l = l << 1;							/* Shift the mask */
		}
	assert(amp - PlaceDat->amp == PlaceDat->ndata);		/* Better have used all the data */	
	if (PlaceOff) {							/* Compute max and min time offsets */
		minoff = 30000; maxoff = -30000;
		for (i = iscurrent ? 3 : 1; i >=0; --i) {
			if (PlaceOff->offsets[i] < minoff) minoff = PlaceOff->offsets[i];
			if (PlaceOff->offsets[i] > maxoff) maxoff = PlaceOff->offsets[i];
		}	}
	if (Place->timezone >= 0) {
		timezone = &Zones[(unsigned char)(Place->timezone)];
		meridian = 60L * timezone->meridian;
		}
	 else {					/* Zone is an offset from GMT */
		timezone = NULL;
		meridian = (60L * 30L) * (Place->timezone + 29);	/* v 2.3: -1 = GMT+14, etc. */
		}
	/* Meridian for computation may be offset from time zone, but offsets don't have that field. */
	tmeridian = PlaceDat->meridian * (30L * 60L);
	PlaceIndex = new_place;
	figure_multipliers();				/* Re-initialize tide computation */
	if (Place->flags & P_HIDDEN) 			/* If it's hidden (i.e., the coefficient locn) */
		return 0;					/*  then we're done. */

	Preferences.Plastplace = PlaceIndex;	/* Set some preferences */
  	if (iscurrent) Preferences.Plastcurrent = PlaceIndex;
		else Preferences.Plasttide = PlaceIndex;
	if (FrmGetActiveFormID() == MainForm) 
	CtlSetLabel(GetObjectPtr(OTHERTYPE), disgraph ? T_TABLE : T_GRAPH);
	j = PlaceIndex;					/* Insert index in recent places list */
//printf("GetPlaceData args %s, yr %d.\n",CstSpeedsHandle.p.handle ? "open" : "closed", year);
//SysTaskDelay(100);
	for (i = 0; i < Preferences.PnRecentPlaces; ++i) {
		k = Preferences.RecentPlaces[i];	/* Go down from top of list */
		Preferences.RecentPlaces[i] = j;
		if (k == PlaceIndex) return 0;	/*  until finding the current place */
		j = k;
		}						/*  or hitting the end of the list */
	if (Preferences.PnRecentPlaces < NRECENT)	
		Preferences.RecentPlaces[Preferences.PnRecentPlaces++] = j;
	return 0;	
	}

/* Load *cst_nodes, *cst_speeds, *cst_epochs and lock them */

int GetYearData(int new_year) {
	int i;
//printf("GetYearData(%d) was %d\n",new_year,year); SysTaskDelay(50);
	if (year == new_year) return 0;
	year = new_year;	
	if (year < first_year || year >= first_year + num_epochs) {
		barf (MISSINGYEAR);
		return 1;
  		}
	UnlockHand(&CstEpochsHandle); 	/* Free previous records */
	UnlockHand(&CstNodesHandle);

	if (!CstSpeedsHandle.p.handle) {				/* Only load this one once */
//printf("Opening Tideargs\n"); SysTaskDelay(50);
		cst_speeds = QueryRecord(&CstSpeedsHandle, &TideDB, 1);
		}

	i = 2 * (year - first_year) + 2;
	if (!(cst_epochs = QueryRecord(&CstEpochsHandle,&TideDB,i))) {
		barf(BADDATABASE);
		return 1; }

	if (!(cst_nodes = QueryRecord(&CstNodesHandle, &TideDB, i+1))) {
		barf(BADDATABASE);
		return 1; }
#if 0
printf("Year %d = %d\n",year,year-first_year);
printf("%d years, %d consts\n",num_epochs,num_csts);
for (i = 0; i < 5; ++i)
printf("epochs[%d] = %e\nnodes[%d] = %e\n",i,cst_epochs[i],i,cst_nodes[i]);
SysTaskDelay(500);
#endif
	return 0;	
	}

static void UnlockHand(struct DBHandle *pHandle) {
	if (pHandle->p.handle) HandleUnlock(pHandle);
	pHandle->p.handle = NULL;
	}

void CloseDatabases(void) {
  int i;
//printf("Closing databases\n"); SysTaskDelay(50);
  UnlockHand(&CstEpochsHandle);
  UnlockHand(&CstNodesHandle);
  UnlockHand(&CstSpeedsHandle);
  cst_speeds = cst_epochs = cst_nodes = NULL;		/* Note TideArgs is closed too */
  InitializeEpoch();						/* Tell tidelib that */
  UnlockHand(&PlaceHandle);
  UnlockHand(&PlaceDatHandle);
  UnlockHand(&PlaceOffHandle);
  for (i = 0; i < nPlaceDBs; ) {
	HandleUnlock(&PlaceHdrHandles[i]);
	CloseDatabase(&PlaceDBs[i++]);
	}
  if (TideArgsID) CloseDatabase(&TideDB);
  TideArgsID = 0;
  nPlaceDBs = 0;
  PlaceIndex = coeff_PlaceIndex = -1;
}

Boolean DeleteDatabase(int j) {				/* Delete database j */
	int k, l, ll;
	if (PlaceDBs[j].volume == IN_RAM &&		/* Is it in internal flash? */
		MemLocalIDKind(PlaceIDs[j]) == memIDPtr) FrmAlert(DBINROM);
	  else if (FrmCustomAlert (DBDELETE, PlaceHdrs[j]->db_name, 
			nPlaceDBs == 1 ? T_ONLYDB : "", 
			NULL) == 0) {
		k = PlaceHdrs[j]->db_ID;
		ll = 0;
		for (l = 0; l < nPlaceDBs; ++l)		/* Check if removing duplicate database */
			if (k == PlaceHdrs[j]->db_ID && k != l) ll = 1;
		if (ll) {
			/* Make sure no preference points to deleted database */
			if (k == Preferences.Plasttide >> ID_SHIFT
				|| k == Preferences.Plastcurrent >> ID_SHIFT
				|| nPlaceDBs == 1)
					Preferences.Plastplace = -1;
			/* Delete references to database in recent places list */
			for (l = ll = 0; l < Preferences.PnRecentPlaces; ++l) 
				if (k != (Preferences.RecentPlaces[l] >> ID_SHIFT))
					Preferences.RecentPlaces[ll++] = Preferences.RecentPlaces[l];
			Preferences.PnRecentPlaces = ll;
			DBEnable(j, true);		/* Don't leave it disabled */
			}
		DeleteOrReloadDatabases(j);
		return false;
		}
	return true;					/* User cancelled */
  	}

void DeleteOrReloadDatabases(int j) {
	char pathname[256];
	/* Get out of loc form - it must reference the database */
	switch (FrmGetFormId(FrmGetActiveForm())) {
		case FindForm: 
			FrmReturnToForm (LocForm); 
		case LocForm: 
			far_CloseLocList();
			FrmReturnToForm (MainForm); 
		}
	SavePreferences();
	CloseDatabases();
	if (j >= 0) {
		if (PlaceDBs[j].volume == IN_RAM) {
			assert (DmDeleteDatabase(CardNo,PlaceIDs[j]) == 0);
			}
		else {
			CloseDatabase(&PlaceDBs[j]);
			VFSPathname(j,pathname);
			if (VFSFileDelete(PlaceDBs[j].volume, pathname))
				barf(CANTDELETEDBONCARD);
		}	}
	OpenDatabase();
	far_InitializeTime();
	}

/* 
 * Put VFS pathname of database j into string pathname
 */
static char *VFSPathname(int j, char *pathname) {
	if (j < 0) { j = -j-1; *pathname = 0; }	/* j < 0: just generate name, not path */
		else strcpy(pathname, VFS_PATH);
	strcat(pathname,PlaceHdrs[j]->db_name);
	strcat(pathname,".pdb");
	return pathname;
	}

/* Beaming functions courtesy of Palm - http://oasis.palm.com/dev/kb/samples/1411.cfm */

/***********************************************************************
 *
 * FUNCTION:   WriteDBData
 *
 * DESCRIPTION: Callback for ExgDBWrite to send data with exchange manager
 *
 * PARAMETERS:  dataP : buffer containing data to send
 *                  sizeP : number of bytes to send
 *                  userDataP: app defined buffer for context
 *                       (holds exgSocket when using ExgManager)
 *
 * RETURNED:    error if non-zero
 *
 ***********************************************************************/
static Err WriteDBData(const void* dataP, ULong* sizeP, void* userDataP)
{
     Err            err;

     // Try to send as many bytes as were requested by the caller
     *sizeP = ExgSend((ExgSocketPtr)userDataP, (void*)dataP, *sizeP, &err);

     return err;
}

/***********************************************************************
 *
 * FUNCTION:        SendDatabase
 *
 * DESCRIPTION:     Sends data in the input field using the Exg API
 *
 * PARAMETERS: cardNo: card number of db to send (usually 0)
 *                  dbID:     databaseId of database to send
 *                  nameP:  public filename for this database
 *                            This is the name as it appears on a PC file listing
 *                            It should end with a .prc or .pdb extension
 *                  description: Optional text description of db to show to user
 *                            who receives the database.
 *
 * RETURNED:        error code or zero for no error.
 *

***************************************************************************/

static Err SendDatabase (Word cardNo, LocalID dbID, CharPtr nameP, CharPtr
descriptionP)
{
  ExgSocketType       exgSocket;
  Err                      err;

  do {
     // Create exgSocket structure
     MemSet(&exgSocket, sizeof(exgSocket), 0);
     exgSocket.description = descriptionP;
     exgSocket.name = nameP;

     // Start and exchange put operation
     err = ExgPut(&exgSocket); 
     if (!err) {
          // This function converts a palm database into its external (public)
          // format. The first parameter is a callback that will be passed parts of
          // the database to send or write.
          err = ExgDBWrite(WriteDBData, &exgSocket, NULL, dbID, cardNo);
          // Disconnect Exg and pass error
          err = ExgDisconnect(&exgSocket, err);
          }
     } while (err == exgErrDeviceBusy);		/* If busy, retry */

  return err;
  }

/***********************************************************************
 *
 * FUNCTION:        SendMe
 *
 * DESCRIPTION:     Sends this application
 *
 * PARAMETERS: none
 *
 * RETURNED:   error code or zero for no error.
 *
**************************************************************************/
void BeamTideTool(void)
{
     Err err = 0;
	LocalID dbID;
	UInt16 i, j;
	char buf1[64];
	char buf2[200];
//	if (!sys_30) { barf(NOBEAMING); return; }	/* Don't need because menu item is not there any more */
	if ((i = FrmAlert(BEAMWHAT)) == 3) return;		/* Cancel */
	if (i != 1) {							/* If not beaming just data */	
		/* Find our app using its internal name.
		 * 	Don't worry about it being on an expansion card, because
		 *	system had to copy it into RAM prior to launch. */
		dbID = DmFindDatabase(0, "Tide Tool");
		if (dbID)    // send it giving external name and description
			err = SendDatabase(0, dbID, "Tide Tool.prc", T_TTAPP);
		else
			err = DmGetLastErr();
		}
//printf("Mark 1 err %x\n",err); SysTaskDelay(200);
	if (!err && i != 2) {						/* If not beaming just app */
	   dbID = TideArgsID;
	   if (TideDB.volume != IN_RAM) {
			strcpy(buf1, VFS_PATH);				/* Build path name */
			strcat(buf1,"TideArgs.pdb");			/*  and temporarily copy to RAM */
			err = VFSImportDatabaseFromFile(TideDB.volume, buf1, &i, &dbID);
//printf("Mark 2 err %x %s\n",err,buf1); SysTaskDelay(300);
			}
	   if (!err) {
		SendDatabase(0, dbID, "TideArgs.pdb", T_TTDBASE);
	   	if (TideDB.volume != IN_RAM) 				/* If DB was in expansion memory */
			DmDeleteDatabase(0, dbID);			/*  delete temporary database */
			}
  	   for (j = 0; !err && j < nPlaceDBs; ++j) if (DBEnabled(j)) {
		dbID = PlaceIDs[j];
		if (PlaceDBs[j].volume != IN_RAM) {			/* If DB in expansion memory */
			VFSPathname(j,buf1);
			err = VFSImportDatabaseFromFile(PlaceDBs[j].volume, buf1, &i, &dbID);
//printf("Mark 3 err %x %s\n",err,buf1); SysTaskDelay(200);
			}
		if (!err) {							/* If no error */
			strcpy(buf1,PlaceHdrs[j]->db_name);
			strcat(buf1,".pdb");
			strcpy(buf2,PlaceHdrs[j]->db_name);
			strcat(buf2,T_REGIONAL_DB);
			err = SendDatabase(0, dbID, buf1, buf2);
			if (PlaceDBs[j].volume != IN_RAM) 		/* If DB was in expansion memory */
				DmDeleteDatabase(0, dbID);		/*  delete temporary database */
			}
		}
	   }
	if (err) FrmCustomAlert(ERRSTRING,T_BEAMING_INCOMPLETE ,"","");
//if (err) { printf("Error %d\n",err); SysTaskDelay(200); }
}

/***************************************************
 *
 *	Stuff down here calls faraway functions so wants to be near the end.
 *
 ***************************************************/

long UnpackLat(struct place *p){
	long latitude; int i;
	latitude = p->latitude * 10L;
	i = p->lloflow % 10;
	if (latitude < 0) i = -i;
	return latitude + i;
	}

long UnpackLon(struct place *p){
	long longitude; int i;
	longitude = p->longitude * 20L;
	i = p->lloflow / 10;
	if (longitude < 0) i = -i;
	return longitude + i;
	}

/*
 * 	Look up a record in the database, and lock the handle.
 */
MemPtr QueryRecord(struct DBHandle *ans, struct DBdescriptor *d, UInt16 recno) {
	if ((ans->in_ram = (d->volume == IN_RAM))) {			/* If db is in ram */
		if ((ans->p.handle = DmQueryRecord(d->ref.ram, recno)) == NULL) return NULL;
		}
	 else {
		if (VFSFileDBGetRecord (d->ref.card, recno, &ans->p.handle, NULL, NULL)) return NULL;
		}
	return MemHandleLock(ans->p.handle);;
	}

Boolean DBEnabled(int n) {
	n = PlaceHdrs[n]->db_ID;			/* Convert from index to ID */
	return  !(Preferences.PDBDisabled[n/32] & (1L << (n & 31)));
	}

/*
 *	Display database info.  j = regional database number, or
 *					    -1 for factors database
 *					    -2 for location info
 */
void DisplayDBInfo(int j) {
      char *p, buf[400], dbuf[LATLONLEN > longDateStrLength ? LATLONLEN : longDateStrLength];
      int k, l, notice = 0;
      long lm;
      time_t t;
      struct place *refsta;
      struct DBHandle refhandle;
	UInt16 inram;

  	p = buf;
      switch (j) {
	 case -2:						/* Display place data */
		if (!Place) return;
		p += sprintf(buf,"%s\n",Place->location);
		far_LatLonToA(dbuf, latitude, T_NORTH_SOUTH);
		p += sprintf(p,"    %s\n",dbuf);
		far_LatLonToA(dbuf, longitude, T_EAST_WEST);
		p += sprintf(p,"    %s\n",dbuf);
		if (PlaceOff) {
			if (!(refsta = QueryRecord(&refhandle, &PlaceDBs[this_placedb],
							2 * (PlaceOff->primary & ID_MASK) + 1))) {
				barf(BADLOCNBASE);
				return; }
			p += sprintf(p,T_REFERENCE_STA,refsta->location);
			HandleUnlock (&refhandle);
			}
		/* This counts the bits in the datamask[] array, believe it or not. */
		for (k = l = 0; k < CST_ARRAY_SIZE; ++k) 
			for (lm = PlaceDat->datamask[k]; lm; lm &= ~ (lm ^ (lm - 1))) ++l;
		p += sprintf(p,T_HAS_N_HARMONIC_FACTORS,l);
		p += sprintf(p,T_REGION,PlaceHdrs[this_placedb]->db_name);
		break;
       default:
		p += sprintf(buf,T_LOC_DB_NR,
			PlaceHdrs[j]->db_ID,PlaceHdrs[j]->db_name);
		far_LatLonToA(dbuf, PlaceHdrs[j]->latfrom, T_NORTH_SOUTH);
		p += sprintf(p,T_TIDES_CURRENTS_AREA,
			PlaceHdrs[j]->db_num_places -	PlaceHdrs[j]->db_num_currents,
			PlaceHdrs[j]->db_num_currents,dbuf);
		far_LatLonToA(dbuf, PlaceHdrs[j]->latto, T_NORTH_SOUTH);
		p += sprintf(p,"    %s\n",dbuf);
		far_LatLonToA(dbuf, PlaceHdrs[j]->longfrom, T_EAST_WEST);
		p += sprintf(p,T__TO,dbuf);
		far_LatLonToA(dbuf, PlaceHdrs[j]->longto, T_EAST_WEST);
		p += sprintf(p,"    %s\n",dbuf);
		t = PlaceHdrs[j]->db_created_date;
		notice = PlaceHdrs[j]->db_notice;
		inram = PlaceDBs[j].volume;
		goto created;
	 case -1:
		p += sprintf(p,T_ALGORITHM_DATABASE_COVERS,
					first_year, first_year + num_epochs - 1, num_csts);
		t = args_created;
		inram = TideDB.volume;
	  created:
		p += sprintf(p,T_STORED,(inram == IN_RAM) ? T_IN_RAM : T_ON_CARD);
		far_do_timestamp(NULL, dbuf, far_LocalTimeToDST(&t));
		p += sprintf(p,T_CREATED,dbuf);
		}
	if (FrmCustomAlert (notice ? DBINFOPLUS : DBINFO, buf, NULL, NULL))
		FrmHelp(notice);			/* If he taps MORE, give it him. */
	}
