/* 
 *	tidedis.c - Palm Pilot display and event caching routines

Tide Tool - A tide and current prediction program for the USR PalmPilot
Copyright (c) 1997-2003 Walt Bilofsky

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Main entry routines for this module are:

ShowResult(GMTime_t time, enum TideEvent eventno, float param) - enters
	the next event of the day into DisEvents, and displays it on the
	next line if the display is active.

CompleteDisplay() - marks the current display complete.

DisplayDate() - Displays the current date on line 3.

RedoDisplay() - Puts all the other info up on the screen and reinitializes
	the computation so the tide will be redisplayed.
*/

///#define HANDERA		/***** NOT WORKING YET (and probably never will) *****/

/* Stuff for formatting the tide/current display screen */

#include <PalmOS.h>
#include <PalmCompatibility.h> 

#include "sysinclude.h"
#include "tide.h"
#include "tidelib.h"

#define lTideStgs 13
char TideStgs[7][lTideStgs] = {
	T_HIGHTIDE ,T_LOWTIDE,T_SLACKWATER ,T_SLACKWATER ,
	T_MAX_FLOOD, T_MAX_EBB, "", };

#define MAXLINELEN 40
#define NLINES 12

#ifdef PRECISE
#define FORMLEFT 10
#define FORMRIGHT 60
#else
#define FORMLEFT 0
#define FORMRIGHT 78
#endif
#define COEFSTART 110					/* Where the coefficient label starts */

static char dis_lines[NLINES][MAXLINELEN];
static enum TideEvent dis_labels[NLINES] = {	/* If NoLabel, line's either empty */
	NoLabel,NoLabel,NoLabel,NoLabel,		/* or contains pure text.  If */
	NoLabel,NoLabel,NoLabel,NoLabel,		/* it's another event, contains */
	NoLabel,NoLabel,NoLabel};			/* the event data string */
static unsigned char EightEvents;			/* Whether displaying 4 or 8 events */
unsigned char dis_y[2][NLINES] = {			/* Heights for 4 or 8 events */
	{18,29,50,63,78,89,100,111,122,133,133,133},	/* For four (most tides) */
	{18,18,31,44,57,68,79,90,101,112,123,134}	/* For eight */
	};

#define cBLACK 0
#define cRED 1
#define cDKGRAY 2
#define cLTBLUE 3
#define cMDBLUE 4
#define cWHITE 5
#define cLTGRAY 6
#define cBRTRED 7
#define cOLIVE 8
#define cBLUE 9
#define ncolor_types 10
static const RGBColorType color_types[ncolor_types] = { /* Colors */
	{ 0, 0, 0, 0 },						/* black */
	{ 0, 0x80, 0, 0 },					/* dark red */
	{ 0, 0x44, 0x44, 0x44 },				/* dark gray */
	{ 0, 0xcc, 0xff, 0xff },				/* light blue */
	{ 0, 0x66, 0xff, 0xff },				/* medium blue */
	{ 0, 0xff, 0xff, 0xff },				/* white */
	{ 0, 0xbb, 0xbb, 0xbb },				/* light gray */
	{ 0, 0xff, 0x00, 0x00 },				/* bright red */
	{ 0, 0x00, 0x66, 0x00 },				/* olive */
	{ 0, 0x00, 0x00, 0xcc },				/* blue */
	};
static IndexedColorType color_indexes[ncolor_types];


#define UPGRAPH 0
#define DOWNGRAPH 1
#define NEGGRAPH 2
#define DAYBACK 3
#define NIGHTBACK 4
#define ARROWCOLOR 5
#define THERMFILL 6
#define LEVLINE 7
#define BASECOLOR 8
static const unsigned char colors[][2] = {
/*      Grayscale	 Color   */
	{ cBLACK, 	cBLUE },		/* Upward graph color */
	{ cDKGRAY,	cOLIVE },		/* Downward graph color */
	{ cBLACK, 	cRED },		/* Negative graph color */
	{ cLTGRAY,	cLTBLUE },		/* Day background color */
	{ cWHITE,	cMDBLUE },		/* Night background color */
	{ cBLACK,	cBRTRED },		/* Arrow color */
	{ cBRTRED,	cBRTRED },		/* Thermometer interior fill */
	{ cDKGRAY,	cBRTRED },		/* Thermometer interior fill */
	{ cBLACK, 	cBLACK },		/* Base color (set before drawing icons) */
	};

Boolean sys_40 = false;					/* Screen parameters */
Boolean sys_35 = false;					/* Screen parameters */
Boolean sys_30 = false;
UInt32 scr_depth, scr_depths;
Boolean scr_color = false;
UInt32 max_depth = 1;

char NumspaceStr[2] = "\200";

static float scale;					/* Variables for graphing mode */
static int zeroline;
static int graph_top, graph_bot;
static Boolean bogograph;				/* If converting bogoknots for graph */
#ifdef LABELGRAPH
static unsigned char graph_y[160];			/* If graphing, the y values drawn */
#endif

unsigned char max_events[2] = { 6, 8 };
static RectangleType linesrect = { {0, 18 },		/* Rectangle covering all lines */
					     {160, 134+11-18 } };
static RectangleType oneline = { {0, 0 },		/* Rectangle covering all lines */
					   {160, 0 } };	
#define DATAFONT 1			/* Font for display of event data */
#define LBLFONT 0
#define HDRFONT 1

#define ARROWWID 7					/* Stuff for drawing the arrow - moof! */
#define ARROWHT 7
static char arrow_lin[] = {0, 1, 5, 6, 5, 1, 0, -1 };
static char darrow_st[] = {3, 3, 0, 0, 0, 3, 3 };
static char uarrow_st[] = {3, 2, 1, 0, 1, 2, 3 };
static RectangleType behind_arrow = {{0,0},{ARROWWID, ARROWHT}};
#define arrow_x behind_arrow.topLeft.x
#define arrow_y behind_arrow.topLeft.y
static char *arrow_to_draw = NULL;
static void draw_arrow(char *);
static void draw_axes(void);
Boolean graph_complete = false;
static void setcolor(int);
static void savecolor(void);
static void restorecolor(void);
static void setbackground(int);
#ifdef HANDERA
static void drawvline(int, int, int );
#endif

static char font4line[NLINES] = {HDRFONT,HDRFONT,HDRFONT,0,LBLFONT,LBLFONT,
				LBLFONT,LBLFONT,LBLFONT,LBLFONT,LBLFONT,LBLFONT};

static int EventNr;			/* Event number on screen of current line */

static enum Units units_to;		/* Units the user wants to see on the screen */

/* Order must correspond to tidelib.h, and to tidelib.c in the database generating program. */
units known_units[NUMUNITS] = {{/*"feet",*/ T_FEET_ABBR, LENGTH, 0.3048,2},
                              {/*"meters",*/ T_METERS_ABBR, LENGTH, 1.0,2},
                              {/*"knots",*/ T_KNOT_ABBR, VELOCITY, 1.0,2},
                              {/*"knots^2",*/ T_KNOTSQ_ABBR, BOGUS, 1.0,2},
	/* These are not in the harmonics file but are in the conversion menu */
					{/*"inches",*/ T_INCH_ABBR, LENGTH, 0.3048/12.0 ,1},
					{/*"mph",*/ T_MPH_ABBR, VELOCITY, 0.871,2},
					{/*"km/hr",*/ T_KMPH_ABBR, VELOCITY, 0.540 ,2},
					{/*"m/sec",*/ T_MTRS_SEC_ABBR, VELOCITY, 0.540*3600.0/1000.0,2},
					{/*"ft/sec",*/ T_FT_SEC_ABBR, VELOCITY, 0.3048*0.540*3600.0/1000.0,2},
};

static enum Units hUnits[] = { feet, meters, inches, Default};	/* If changed, also change default in tidedb.c */
static enum Units vUnits[] = { knots, mph, kph, msec, fsec, Default};

static void DisplayLine(enum TideEvent eventno, char * line, int lineno, int offset);
static void DisplaySunMoon(void);
static void ClearSunMoon(void);
static int rtoa(Double, char *, int);
#define PAD 7
#define SMALLPAD 2
static float screen_units(Double);

#if 1
#define checkwindows(s)
#else
static void checkwindows(char *p) {
	LocalID formid;
	WinHandle active, mainwin;
	mainwin = FrmGetWindowHandle(FrmGetFormPtr(MainForm));
	active = WinGetDrawWindow ();
	if (active == mainwin) return;
	formid = FrmGetActiveFormID();
	printf("Check fails: %s\n",p);
	printf("Main: %d hand %lx\n",MainForm,mainwin);
	printf("Draw: %d hand %lx\n",formid,active);
	SysTaskDelay(100);
	}
#endif

void InitializeDisplay() {			/* Initialize display on entering program */
	int i;
	char *p;
      ChrNumericSpace(&Numspace);		/* Get correct numeric space for this ROM version */
	for (i = 0; i < NLINES; ) dis_lines[i++][0] = '\0';
	sys_40 = (RomVersion >= 0x04000000);	/* Get system version */
	sys_35 = (RomVersion >= 0x03500000);
  	sys_30 = (RomVersion >= 0x03000000);

	/* If OS 3.0 or later, set screen to maximum bit depth */
	/* Note - WinScreenMode is backwards compatible to sys_30 using ScrDisplayMode */
	if (sys_30) {
		WinScreenMode(winScreenModeGet, NULL, NULL, &scr_depth, &scr_color);
		WinScreenMode(winScreenModeGetSupportedDepths, 0, 0, &scr_depths, 0);
  		for (i = 0; i < 24; ++i) if (scr_depths & (1L << i)) max_depth = i + 1;
		WinScreenMode(winScreenModeSet, NULL, NULL,&max_depth, NULL);
		}
	if (UseColor == unknownColor) 		/* If color preference hasn't ever been set */
		UseColor = (max_depth == 1) ? noColor :	/*  init to whether color/grayscale is available */
			scr_color ? fullColor : grayColor;
	if (sys_35) for (i = 0; i < ncolor_types; ++i)
		color_indexes[i] = WinRGBToIndex(&color_types[i]);
	savecolor();					/* Save default drawing color */

	p = TideStgs[0];
	for (i = 7*lTideStgs; --i; ++p) if (*p == '\200') *p = Numspace;
	}

void RestoreDisplay() {				/* Restore display on program exit */
	if (sys_30) WinScreenMode(winScreenModeSetToDefaults, NULL, NULL, NULL, NULL);
	}

static RGBColorType defaultcolor, defaultback, dummy;
static IndexedColorType defaultindex;
static Boolean gotdefault = false;
static void setcolor(int index) {				/* Set drawing color */
	if (!UseColor || !sys_30) return;			/* OS < 3.0 - ignore */
	index = colors[index][scr_color];			/* Look up proper color for this screen */
	if (sys_35) WinSetForeColor(color_indexes[index]);/* OS >= 3.5 - set color index */
		else WinSetColors(&color_types[index], &dummy, 0, 0);
	gotdefault = true;
	}

static void setbackground(int index) {			/* Set background - 3.0 <= OS < 3.5 */
	if (!UseColor || !sys_30 || sys_35) return;
	index = colors[index][scr_color];			/* Look up proper color for this screen */
	WinSetColors(0, 0, &color_types[index], &dummy);
	}

static void savecolor(void) {					/* Save current color */
	if (!UseColor || !sys_30) return;			/* OS < 3.0 - ignore */
	if (sys_35) defaultindex = WinSetForeColor(0);	/*  by setting to 0, then restoring */
	  else if (sys_30) WinSetColors(&color_types[0], &defaultcolor, 
			&color_types[1], &defaultback);
	restorecolor();
	}

static void restorecolor(void) {
	if (!UseColor || !sys_30) return;			/* OS < 3.0 - ignore */
	if (sys_35) WinSetForeColor(defaultindex);
	  else if (sys_30) WinSetColors(&defaultcolor, &dummy, &defaultback, &dummy);
	}

static int fntwid(char *s) {
	return FntCharsWidth (s, strlen(s));
	}

static float screen_units(Double val) {
	Boolean neg = 0;
	if (val != (Double)SLACKER) {
	  if (unit == bogoknots && units_to != bogoknots) {
		if ((val -= DATUM) < 0) { neg = 1; val = -val; }
		val = mysqrt(val);
		if (neg) val = -val;
		val += DATUM; 
		}
	  val *= known_units[unit].conv_factor / known_units[units_to].conv_factor;
	  }
	return val;
	}

int rtoa(Double val, char *p, int pad) {
	val = screen_units(val);
	if (val != (Double)SLACKER) {
	  strcpy(p,sfpr(val,pad,known_units[units_to].places,1));
	  strcat(p,known_units[units_to].abbrv);
	  }
	FntSetFont(DATAFONT);
	return fntwid (known_units[units_to].abbrv);
	}

Boolean ShowResult(GMTime_t time, enum TideEvent eventno, float param) {
	char buf[50],*p;
	Boolean i;
	long tdif,maxtdif;
	int k, j = -1;
#ifdef LABELGRAPH
	long l;
	int wid1, wid2, halfwid, horiz;
#endif
	struct tm *t;
	
	if (EventNr >= max_events[EightEvents]) return 0;
	do_timestamp(buf, NULL, t = localtime(&time));
	if (!disgraph) {
		i = rtoa(param, buf + strlen(buf), PAD);
//#if LANG == LANG_FRENCH				/* Yuk.  Ok, well, someday ... */
//		i -= 5;
//#endif
		if (coeff_displaying) {	
		    if (EventNr == 0 && !EightEvents) {	/* If first event and coeffs, display heading */
				DisplayLine(NoLabel,T_COEFF,4, -1);
				++EventNr;
				}
		    i += 18;			/* Fixed space for coeff display - ugh! */
		    if (eventno == HighTide) { /* Display of French tidal coefficients */
			maxtdif = 24L*hoursInSeconds;	/* Get closest coefficient */
			for (k = 0; k < ncoeffs; ++k) {
				tdif = time - coeff_values[k].time;
				if (tdif < 0) tdif = -tdif;
				if (tdif < maxtdif) { maxtdif = tdif; j = k; }
				}
			if (j >= 0) {
//printf("Event at %ld Off %d dt %ld\n",time/3600,j,maxtdif/60);
				p = buf+strlen(buf);
				*p++ = ' ';
				if (coeff_values[j].val < 100) *p++ = Numspace;
				StrIToA(p,coeff_values[j].val);
		    }	}	}
//if (eventno == HighTide) { printf("Event %d j = %d cd %d\n",EventNr,j,coeff_displaying); SysTaskDelay(100); }
		DisplayLine(eventno, buf, 4 + EventNr, i);
		}
#ifdef LABELGRAPH				/* This feature needs more work */
	  else {
		rtoa(param, buf+25, 0);
		FntSetFont(LBLFONT);
		wid1 = fntwid(buf);
		wid2 = fntwid(buf+25);
		halfwid = (1 + max(wid1, wid2)) / 2;
		horiz = (t->tm_min + 60 * t-> tm_hour) / 9;
		if (horiz + halfwid > 160) horiz = 160 - halfwid;
		  else if (horiz - halfwid < 0) horiz = halfwid;
		WinDrawChars(buf, (int)strlen(buf), horiz - wid1/2, dis_y[EightEvents][4]);
		WinDrawChars(buf+25, (int)strlen(buf+25), horiz - wid2/2, dis_y[EightEvents][5]);
		}
#endif
	return (++EventNr < max_events[EightEvents]);
	}

/* Draw (or erase if text==NULL) a line of tide data on the screen. 
 * Save the text of the line in dis_lines.
 * Calling program must set dis_labels first if it's an event.
 */
static void drawline(int i, char *text, int offset) {
	char *p;
	int left = 0;			/* Label lines start at 0 */
	int font = font4line[i];
checkwindows("drawline");
//	assert (FrmGetWindowHandle(FrmGetFormPtr(MainForm)) == WinGetActiveWindow ());
	if (text == NULL) {
		oneline.topLeft.y = dis_y[EightEvents][i];
		oneline.extent.y = FntCharHeight();
		WinEraseRectangle(&oneline, 0);
		dis_labels[i] = NoLabel;	/* erase the saved data */
		dis_lines[i][0] = 0;
		return;
		}
	if (dis_labels[i] != NoLabel) {	/* If it's an event line */
		FntSetFont(LBLFONT);		/* do the label first */
		p = TideStgs[dis_labels[i]];
		WinDrawChars(p, (int)strlen(p), FORMLEFT, dis_y[EightEvents][i]);
		left = FORMRIGHT - offset;	/* and set up for the text */
		font = DATAFONT;
		}
	FntSetFont(font);
	strcpy(dis_lines[i],text);
	if (dis_labels[i] == NoLabel) {	/* If no label, */
		left = 160- fntwid (text);	/* center the line */
		if (offset != -1) left /= 2;	/* or, if offset -1, right justify it */
		}
	if (left < 0) {				/* Too wide - use font 0 */
		FntSetFont(0);
		left = (160 - fntwid (text)) / 2;
		}
	if (left < 0) left = 0;			/* Shouldn't happen, but ... */
	WinDrawChars(dis_lines[i], (int)strlen(dis_lines[i]), left, 
									dis_y[EightEvents][i]);
	}

#if 0		/* This doesn't happen any more */
void DisplayDeferred(void) {
	int i;
	if (FrmGetWindowHandle(FrmGetFormPtr(MainForm)) != WinGetActiveWindow ())
		return;
	if (DisplayDef == 0) return;
	DisplayDef = 0;
	for (i = 0; i < NLINES; i++) if (deferred[i]) {
		deferred[i] = 0;
		drawline(i,dis_lines[i],def_offset[i]);
	}	}
#endif /* 0 */

static time_t seconds_in_today;			/* Usually 24 * 60 * 60 except on DST switchover days */

#define GRIDMARGIN 3					/* Because the graph doesn't get close to the limits */
static void gridline(LocalTime_t hour) {
	int j;
	LocalTime_t linetime;
	linetime = DateBeingDisplayed + hour;
	linetime = local2GMT(linetime,false) - local2GMT(DateBeingDisplayed,false);
	j = (linetime * 160L) / seconds_in_today;
	if (hour  == 12 * 3600L) WinDrawLine(j,graph_top+GRIDMARGIN ,j,graph_bot-GRIDMARGIN ); 
		else WinDrawGrayLine(j,graph_top+GRIDMARGIN ,j,graph_bot-GRIDMARGIN );
	}

static RectangleType backrect = { {0, 0 },		/* Rectangle for background fill */
					    {160, 160 } };
static void fillback(x0, xf) {
	backrect.topLeft.x = x0;
	backrect.extent.x = xf;
	WinDrawRectangle(&backrect, 0);
	}

static unsigned char intervals[] = {1, 2, 5, 10, 20, 30, 50, 80, 0};
#define TICKLEN 3
void InitializeGraph(LocalTime_t lday) {
	LocalTime_t gline;
	float gmin, gmax;
	int sunup, sundown;
	struct tm *t;
	int j;
	Boolean noevent = false;
	int color = NIGHTBACK;

checkwindows("InitGraph");
//	assert (FrmGetWindowHandle(FrmGetFormPtr(MainForm)) == WinGetActiveWindow ());
	seconds_in_today = lday;
	graph_top = dis_y[EightEvents][3] + FntCharHeight();
	graph_bot = dis_y[EightEvents][NLINES-1] + FntCharHeight() - 1;
	gmin = absmin;
	gmax = absmax;
	bogograph = unit == bogoknots && units_to != bogoknots;
	if (bogograph) {
		gmax = screen_units(gmax);
		gmin = screen_units(gmin);
		}
	if (gmin > 0) gmin = 0;
	scale = (graph_bot - graph_top + 1) / (gmax - gmin);
	zeroline = graph_top + gmax * scale;
	if (UseColor) {					/* Paint background to show sun up and down */
		backrect.topLeft.y = graph_top;
		backrect.extent.y = graph_bot - graph_top;
		Fiddler(DateBeingDisplayed);
		if ((sunrise & SUN_MASK) && (sunset & SUN_MASK)) {
			t = localtime(&sunrise);
			sunup = (t->tm_hour * 60 + t->tm_min) / 9;
			t = localtime(&sunset);
			sundown = (t->tm_hour * 60 + t->tm_min) / 9;
		 	if (sunup > sundown) { 	/* Sun up at midnight */
				j = sunup; sunup = sundown; sundown = j; color = DAYBACK;}
			}					/* Cosmetic - WB 12/03 */
		  else {					/* Don't display sun up unless we have both */
			noevent = true;
			sunup = 0;				/*  a sunrise and a sunset */
		 	sundown = (24 * 60)/9;
			if (sunrise == ALWAYS_LIGHT || sunset == ALWAYS_LIGHT) color = DAYBACK;
			}
//if (sunup == 0 || sundown == (24 * 60)/9) {
//printf("%lx %lx\n",sunrise, sunset); SysTaskDelay(200); }
		setcolor(color);
		fillback(0, 160);
		if (!noevent) {
			setcolor(color ^ (NIGHTBACK ^ DAYBACK));	/* Flip to opposite color */
			fillback(sunup, sundown - sunup);
			}
		restorecolor();
		}
	for (gline = 3 * 3600L; gline < 24 * 3600L; gline += 3 * 3600L) gridline(gline); 
	graph_complete = false;
	draw_axes();
	}

/* Draw and label the axes.  Okay, this is kinda' cutesey: 
 * Routine is called twice, once before the graph is drawn and once after.
 * In OS before 3.5, or if color/grayscale is not in use, this just draws grid 
 *	and labels before graph is drawn.
 * Beginning OS 3.5, it draws the grid before graph is drawn, then paints the
 *	labels afterwards using xor.  But needs to do all the work before too, in
 *	order to know where to start the grid lines.
 */
static void draw_axes(void) {
	int k, j, y, w, l, delta, ht, ht2;
	unsigned char *ip;
	char buf[10],*p;
	float factor;
	Boolean drawletters = 				/* Draw letters first unless OS 3.5 or later */
		(graph_complete ^ !sys_35);

	FntSetFont(LBLFONT);
	ht = FntCharHeight();
	delta = 11 * ht;
	factor = 10.0 / scale;
	if (!bogograph) {
		delta = screen_units(delta);
		factor = screen_units(factor);
		}
	delta = 0.9 + delta / scale;	/* Difference between Y labels, in tenths */
	for (ip = intervals; *ip; ++ip) 			/* Try to make the interval nice */
		if (delta <= *ip) { delta = *ip; break; }	
	if (!*ip) delta = 100 * ((delta + 99) / 100);
	ht2 = ht / 2;
	for (j = 0; ; j -= delta) {
		y = zeroline - j / factor;
		if (y > graph_bot - ht2) break;
		} 
	j += delta;
	if (sys_35) WinSetDrawMode(scr_color ? winOverlay : winInvert);
	setbackground(NIGHTBACK);				/* Only for drawing chars on color */
	if (!iscurrent && Preferences.Puselevel) {	/* Draw line for alert level */
		y = zeroline - (int)(0.5 + (scale * Preferences.Plevel 
				* known_units[Preferences.Plevelunits].conv_factor
			 	/ known_units[unit].conv_factor));
//if (graph_complete) printf("lev %d => %d * %d = %d\n",(int)(100*Preferences.Plevel),
//(int)(100 * Preferences.Plevel * known_units[Preferences.Plevelunits].conv_factor/ known_units[unit].conv_factor),
//(int) (100*scale), (y));
		setcolor(LEVLINE );
		if (y <= graph_bot && y >= graph_top) WinDrawGrayLine(0,y,160,y);
		restorecolor();
		}
	while ((y = zeroline - j / factor) >= graph_top + ht2) {
		p = buf;
		k = j;
//printf("At %d show %d\n",y-ht2,j);
		if (k < 0) { *p++ = '-'; k = -k; }
		p += sprintf(p,"%d",k/10);
		if (delta < 10) { *p++ = '.'; *p++ = k % 10 + '0'; }
		strcpy(p,known_units[units_to].abbrv);
		l = strlen(buf);
		if (drawletters) {
			if (sys_35) WinPaintChars(buf,l,0,y - ht2);
			   else WinDrawChars(buf,l,0,y - ht2);
			}
		w = FntCharsWidth (buf,l);
		if (!graph_complete)				/* Always draw grid before graph */
			WinDrawGrayLine(w+1,y,160,y);
		j += delta;
		}
	restorecolor();
	if (sys_35) WinSetDrawMode(winPaint);
	}

void DisplayGraphLine(int i, float value) {
	int j;
	static float lastvalue;
	if (bogograph) value = screen_units(value);
	if (i < 0) {					/* Just setting the slope for graph color */
		lastvalue = value;
		return; }
	if (iscurrent) setcolor(value < 0 ? DOWNGRAPH : UPGRAPH);
		else setcolor(value < 0 ? NEGGRAPH : value >= lastvalue ? UPGRAPH : DOWNGRAPH);
	lastvalue = value;
	j = zeroline - (int)(value * scale);
#ifndef HANDERA
	WinDrawLine(i,zeroline,i,j);
#else
	drawvline(i,zeroline,j);
#endif
	restorecolor();						/* Set to default color */
#ifdef LABELGRAPH
	graph_y[i] = j;
#endif
	if (i == 159) {
		graph_complete = true;
		draw_axes();
		draw_arrow(arrow_to_draw);
		}
	}

/* Erase the lines area on the screen */
static void ClearDisplay(void) {
	int i;
	ReleaseDisplayHandle();
	WinEraseRectangle(&linesrect, 0);
	for (i = 0; i < NLINES; ++i) {
		dis_lines[i][0] = 0; 
		dis_labels[i] = NoLabel;
		}
	units_to = iscurrent ? vUnits[Preferences.Pvunits] : 
				hUnits[Preferences.Phunits];
	if (units_to == Default) units_to = unit;				/* No conversion - default */
	}

/* Display the present location on one line.  Break tide into two if necessary */
static void DisplayLocation(void) {
	int j = 1;
	char *p, c[MAXLOCSTR+2];
checkwindows("DisLocn");
	strcpy(c,Place->location);
	FntSetFont(Place->split.font);
	if (!EightEvents && Place->split.loc) {
		p = &c[Place->split.loc];
		*p = 0;
		DisplayLine(NoLabel, p+1, 1, 0);
		j = 0;
		}
	DisplayLine(NoLabel, c, j, 0);
}

/* 
 *	Clears the screen, puts up all the other info, and calls BackToMidnight
 *	to cause the tide to be recomputed and displayed.
 */

void RedoDisplay(Boolean justvals) {
  
  if (iscurrent) {
    dis_high = MaxFlood;
    dis_low = MaxEbb;
    dis_rise = LowSlack;
    dis_fall = HighSlack;
  } else {
    dis_high = HighTide;
    dis_low = LowTide;
    dis_rise = dis_fall = Slack;		/* Only used for tide level alerts */
  }
  EightEvents = iscurrent || Preferences.Puselevel;
  if (!justvals) {
  	ClearDisplay();
  	DisplayLocation();			/* Display location */
  	DisplayDate();				/* and date */
  	EventNr = 0;
  	TimeToDisplay = 0;			/* Force redisplay of sun/moon and "situation now" line */

/* Set some preferences.  Today may not be the date now, either because 
 *	this is the first run of the program today, or because it's past midnight.
 *    In either event, invalidate the displays for today.
 */
  	Preferences.Ptoday = LocalDateNow;
  	BackToMidnight();
	}
}

static void DisplayLine(enum TideEvent eventno, char * line, int lineno, int offset) {
	dis_labels[lineno] = eventno;
	drawline(lineno, line, offset);
	}

void DisplayDate(void) {
  char buf[50];
  struct tm *date;
  Boolean d;
  date = LocalTimeToDST(&DateBeingDisplayed);
  DateToAscii(date->tm_mon,date->tm_mday,date->tm_year,LongDateFormat,buf);
  strcat(buf,"  -  ");
  strcat(buf,ZoneName(d = IsDST(DateBeingDisplayed)));
  if (IsDST(DateBeingDisplayed + 24L*60L*60L) != d) {
	strcat(buf,"/");
	strcat(buf,ZoneName(!d));
	}
  drawline(2, NULL, 0);					/* Erase line first */
  ClearSunMoon();						/* and pretty pix */
  DisplayLine(NoLabel,buf, 2, 0);
  }

#define SUNWID 13
#define SUNHT  13
#define SUNX 0
#define MOONX 146
static unsigned char sun_y[2] = { 49, 30 };
static RectangleType sunrect = { {SUNX, 0 },
					   {SUNWID, SUNHT } };

static void ClearSunMoon(void) {
	sunrect.topLeft.x = SUNX;
	sunrect.topLeft.y = sun_y[EightEvents];
	WinEraseRectangle(&sunrect,0);
	sunrect.topLeft.x = MOONX;
	WinEraseRectangle(&sunrect,0);
	}

int HitSunMoon(int x, int y) {
	int i;
	if (y >= sun_y[EightEvents] && y < sun_y[EightEvents]+SUNHT) {
		if (x >= MOONX && x < MOONX+SUNWID) return 1;
		if (x >= SUNX && x < SUNX+SUNWID) return 2;
		}
	if (disgraph) {
		if (y >= graph_top && y <= graph_bot) {
			/* Get the time of the pen hit and display the conditions then */
			DisplayCurrentCondx(faketime + (x * seconds_in_today)/160L);
			}
		  else if (DisplayingToday) {			/* If touched time, return to now */
			i = dis_y[EightEvents][3];
			if (y >= i && y < i + FntCharHeight()) TimeToDisplay = TimGetSeconds();
		}	}
	return 0;
	}

void MoveArrow(int i) {				/* Fiveway controller wants to shift arrow */
	if (disgraph) {
		if ((i += arrow_x + ARROWWID / 2) >= 160) i = 159;
		if (i < 0) i = 0;
		DisplayCurrentCondx(faketime + (i * seconds_in_today)/160L);
	}	}

static void OneOrb(int which, int x) {
	VoidHand p;
	BitmapPtr bp;
	setcolor(BASECOLOR);
	p = DmGet1Resource('Tbmp', MOON_BM0 + which);
	bp = MemHandleLock(p);
	WinDrawBitmap(bp, x, sun_y[EightEvents]);
	MemHandleUnlock(p);
	DmReleaseResource(p);
	}

static void DisplaySunMoon(void) {
	GMTime_t GMTnow;
      enum {set = 12, horizon = 13, up = 14} sunstate = set;		/* Sun picture indices */
	OneOrb(MoonPhase(DateBeingDisplayed), MOONX);
	Fiddler(LocalDateNow);				/* Do "Sunrise, Sunset" :-) */
	GMTnow = local2GMT(TimGetSeconds(),0);
	/* Make simplifying assumption that twilight = up 20 min to down 40 min */
	if (sunrise == ALWAYS_LIGHT ) sunstate = up;
	  else if (sunrise ==  ALWAYS_DARK || GMTnow < sunrise - 40*60);
	  else if (GMTnow < sunrise + 20*60) sunstate = horizon;
	  else sunstate = up;
	if (!(sunset & SUN_MASK) || GMTnow < sunset - 20*60);
	  else if (GMTnow < sunset + 40*60) sunstate = horizon;
	  else sunstate = set;
	OneOrb(sunstate,SUNX);	
	}

/* Display the current situation on line 3 */
static char timestrs[4][14] = T_TIDE_CURRENT_DESCR;
static WinHandle bits_behind_arrow = 0;

static void draw_arrow(char *arrow) {		/* Draw an arrow that's waiting to happen */
	char *p;
	int i;
	Word dummy;
	if (!graph_complete || !arrow) {
		arrow_to_draw = arrow;
		return;
		}
	ReleaseDisplayHandle();
	bits_behind_arrow = WinSaveBits(&behind_arrow, &dummy);
	i = arrow_x;
	setcolor(ARROWCOLOR);
	for (p = arrow_lin; *p >= 0; ++i, arrow++) 	/* Draw the arrow */
#ifndef HANDERA
		WinDrawLine(i, arrow_y + *arrow, i, arrow_y + *arrow + *p++);
#else
		drawvline(i, arrow_y + *arrow, arrow_y + *arrow + *p++);
#endif
	restorecolor();
	arrow_to_draw = NULL;
	}

void DisplayCurrentCondx(GMTime_t when) {
   char buf2[80], *arrow;
   GMTime_t gmt, nowgmt, adjgmt;
   LocalTime_t ttd;
   struct tm *tm;
   float t, tf;
   int slope,adj;
   char *p, *pp;
#ifndef GTD_COSINE   
   if (InitCosine(false)) return;		/* If no cosine yet, return to event loop */
#endif
   DisplaySunMoon();				/* Update sun/moon display if necessary */
   ttd = ((TimGetSeconds()+5) / 60L) * 60L;	/* Get time rounded to last whole minute */
   nowgmt = gmt = local2GMT(ttd, 0);
   if (when) {
	gmt = when;
	}
//printf("dt%d At %lu displaying %lu GMT\n",DisplayingToday,ttd-3051900000,gmt-3051900000);
   if (when || DisplayingToday) {
	if ((t = TideAnyplace(gmt, &slope)) == SLACKER) {
		if (!when)
			TimeToDisplay = TimGetSeconds() + 
				SysTicksPerSecond(); 	/* If can't interpolate the value yet */
		return;					/*  try again in a second. */
		}
	tm = when ? localtime(&when) : LocalTimeToDST(&ttd);
	strcpy(buf2,when ? T_AT : T_NOW);
	p = buf2 + strlen(buf2);
	p = TimeToString(p, 60*tm->tm_hour + tm->tm_min,true,false,0,Preferences.Ptunits);
	p += sprintf(p,"%s %s ", 
			timestrs[(iscurrent != 0) + 2 * (when != 0)], 
			nowgmt == gmt ? T_IS : nowgmt < gmt ? T_WILLBE_WAS);
	rtoa(t,p,0);
	p = buf2 + strlen(buf2);
	if (iscurrent) strcpy(p, ".");
	   else {
		pp = slope ? T_RISING_FALLING ;
		strcpy(p, pp);
		FntSetFont(LBLFONT);			/* Use widest "rising falling" that fits */
		if (fntwid(buf2) > 160) *p = 0;
		}
	if (disgraph) {					/* Draw an arrow on the graph */
		if (bogograph) t = screen_units(t);
		ReleaseDisplayHandle();			/* Blank the old one if necessary */
		adj = ARROWWID / 2;			/* Compute position adjust for left edge */
		arrow_x = (160L * (gmt - localtm2GMT(tm, true))) / seconds_in_today;
		arrow_x -= adj;
		adjgmt = gmt - 9L * 60L * adj;	/* 9 minutes per graph line - close enough for slope */
		arrow_y = zeroline - (int)(t * scale) + 2;
		arrow = uarrow_st;
		if (arrow_y > zeroline && 
			arrow_y + ARROWHT >= graph_bot) /* If arrow would be too low */
				arrow_y = zeroline;
		if (arrow_y <= zeroline) {		/* If arrow needs to point down */
			arrow_y -= ARROWHT + 3;		/* If too high, it'll get blatzed */
			arrow = darrow_st;		/*  by text line, so don't worry. */
			slope = !slope;
			}
		tf = TideAnyplace(slope ? adjgmt : adjgmt + 9L * 60L * (ARROWWID - 1), &slope);
		if (bogograph) tf  = screen_units(tf);
		adj = (int)((tf - t) * scale); /* Adjust height of arrow for slope */
		if (adj > 0) adj = -adj;
		if (arrow == uarrow_st) adj = -adj;
		arrow_y += adj;
		draw_arrow(arrow);
		}
	drawline(3,NULL,0);				/* Erase previous display if any */
	DisplayLine(NoLabel,buf2,3,0);
	}
   TimeToDisplay = ttd + 60;				/* Do this again in a minute */
   }								/* If penned a time, this puts off redisplay for 60 sec */

void ReleaseDisplayHandle(void) {			/* Do this separately so we can release the memory on exit (?) */
	if (bits_behind_arrow) 
		WinRestoreBits(bits_behind_arrow,arrow_x,arrow_y);
	bits_behind_arrow = NULL;
	}

/*
 * Thermometer display routines (for place name sort)
 */
#define ThermX0 30
#define ThermDX 100		/* This should not be changed */
#define ThermY0 80
#define ThermDY 9
static RectangleType thermometer = { {ThermX0, ThermY0 },			/* Outside of the thermometer */
					       {ThermDX, ThermDY } };

static RectangleType fillrect = { { ThermX0, ThermY0 }, { ThermDX, ThermDY } }; 
static int oldx;
static const CustomPatternType blackpatt = { -1, -1, -1, -1, -1, -1, -1, -1 };

#ifdef HANDERA
static void drawvline(int x, int y0, int yf) {	/* Draw vertical line using a rectangle */
	if (y0 > yf) {
		int t; t = y0; y0 = yf; yf = t; }
	{ RectangleType r = { { x, y0 }, { 1, yf - y0 } };
	WinSetPattern(&blackpatt);
	WinFillRectangle(&r, 0); }
	}
#endif

static Boolean DrawFrame;

void InitThermometer(void) {
	DrawFrame = true;
	if (!LocFormOnTop()) return;
	WinDrawRectangleFrame(simpleFrame, &thermometer);
	WinSetPattern(&blackpatt);
	fillrect.topLeft.x = ThermX0;
	fillrect.extent.x = oldx = 0;
	DrawFrame = false;
	}

void SetThermometer(int pct) {
	if (!LocFormOnTop()) {
		DrawFrame = true;	/* Window covered.  Set to reinit when it's back. */
		return;
		}
	if (DrawFrame) InitThermometer();
	pct /= 10;			/* Unscale it */
	if (oldx >= pct) return;
	fillrect.topLeft.x = ThermX0 + oldx;
	fillrect.extent.x = pct - oldx;
	setcolor(THERMFILL);
	WinFillRectangle(&fillrect,0);
	restorecolor();
	oldx = pct;
	}

void EraseThermometer(void) {
	if (!LocFormOnTop()) WinEraseRectangle( &thermometer, 0);
	DrawFrame = true;
	}

#ifdef PRINT_DEBUG
/* For debugging */
void printime(LocalTime_t t) {
	char time[longDateStrLength],date[longDateStrLength];
	do_timestamp(time,date,localtime(&t));
	printf("%s %s",date,time);
	}

void prinloctime(LocalTime_t t) {
	char time[longDateStrLength],date[longDateStrLength];
	do_timestamp(time,date,LocalTimeToDST(&t));
	printf("%s %s",date,time);
	}

void prindate(LocalTime_t t) {
	char b[longDateStrLength];
	do_timestamp(NULL, b, LocalTimeToDST(&t));
	printf("%s",b);
	}
#endif
