/*  tidelib.c - tide computation functions for Tide Tool for PalmPilot.

Tide Tool - A tide and current prediction program for the USR PalmPilot
Copyright (c) 1997-2003 Jeff Dairiki and Walt Bilofsky

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    This code was originally based on the tide computation library
	routines in XTide 1.5, Copyright (C) 1995, 1996  David Flater, and
	also starring:  Jef Poskanzer; Rob Miracle; Geoff Kuenning; Dale DePriest.
    It has been essentially rewritten by Jeff Dairiki to use integer math
      and the computation otherwise improved.

    The tide prediction algorithm originally used in this program was developed
    with United States Government funding, so no proprietary rights
    can be attached to it.  For more information, refer to the
    following publications:

    Manual of Harmonic Analysis and Prediction of Tides.  Special
    Publication No. 98, Revised (1940) Edition.  United States
    Government Printing Office, 1941.

    Computer Applications to Tides in the National Ocean Survey.
    Supplement to Manual of Harmonic Analysis and Prediction of Tides
    (Special Publication No. 98).  National Ocean Service, National
    Oceanic and Atmospheric Administration, U.S. Department of
    Commerce, January 1982.

*/

#include <PalmOS.h>
#include <PalmCompatibility.h> 

#include "sysinclude.h"
#include "tide.h"
#include "tidelib.h"


/*
 * Time management
 * 	NOTE - ALL TIMES IN THIS MODULE ARE GMT.
 *
 * FIXME: try to get rid of these globals.
 *
 */
#define TIDE_BAD_TIME		((time_t)-1)

static time_t epoch 		= TIDE_BAD_TIME; /* this years newyears */
static time_t next_epoch	= TIDE_BAD_TIME; /* next years newyears */
extern GMTime_t faketime;		/* GMT time to begin computation */

/* FIXME: make this a float */
extern double DATUM;
static float tide_alert;	/* Remember how much tide is off for alert */

Boolean offset_event_good = false;
Boolean need_zeros;		/* Whether 0 crossings are needed (current or alarm) */

void InitializeEpoch(void){
	epoch = TIDE_BAD_TIME; 
	next_epoch= TIDE_BAD_TIME; 
	year = 0;
	}

static time_t
epoch_for_year (int year)
{
  struct tm ht;

  ht.tm_year = year;
  ht.tm_sec = ht.tm_min = ht.tm_hour = 0;
  ht.tm_mday = ht.tm_mon = 1;	/* Months start with 1 on Pilot */
  ht.tm_isdst = 0;
  return mktime(&ht);
}

static void
happy_new_year (int new_year)
{

  if (GetYearData(new_year))	/* this sets year too */
      return;
  if (year < first_year || year >= first_year + num_epochs)
      barf(MISSINGYEAR);
      
  epoch = epoch_for_year(year);
  next_epoch = epoch_for_year(year + 1);

  figure_multipliers ();
}

static void
possible_happy_new_year (time_t t)
{
  if (epoch == TIDE_BAD_TIME || t < epoch || t >= next_epoch)
      happy_new_year(gmtime(&t)->tm_year);
}



/*************************************************************************
 *
 * Here is a much more efficient (by a factor of 15 in CPU time on my
 * computer), but more complicated scheme for finding the times
 * of the next high/low tides, or tide transitions.
 *
 * The old method was just a linear search, advancing in one minute
 * steps until a local max/minimum or transition was found.
 *
 * The new method is somewhat more detailed --- the key improvement
 * is the use of a modified Newton-Raphson to find level-crossings
 * or extrema.
 *
 * Geoffrey T. Dairiki Fri Jul 19 15:44:21 PDT 1996
 *
 ************************************************************************/

#define TIDE_LOG2_MAX_SCALED_TIDE (3)
#define TIDE_MAX_SCALED_TIDE	(1 << TIDE_LOG2_MAX_SCALED_TIDE)
#define TIDE_LOG2_TIMESCALE	(13)
#define TIDE_TIMESCALE		(1 << TIDE_LOG2_TIMESCALE)

/* TIDE_TIME_PREC
 *   Precision (in seconds) to which we will find roots
 */
#define TIDE_TIME_PREC (15)

/* TIDE_TIME_BLEND
 *   Half the number of seconds over which to blend the tides from
 *   one epoch to the next.
 */
#define TIDE_BLEND_TIME (3600)

/* TIDE_TIME_STEP
 *   We are guaranteed to find all high and low tides as long as their
 * spacing is greater than this value (in seconds).
 */
#define TIDE_TIME_STEP (TIDE_TIME_PREC)

#define TIDE_MAX_DERIV (2)      /* Maximum derivative supported by
                                 * time2dt_tide() and family. */

/* #define TIDE_BAD_TIME   ((time_t) -1) */

#define TIDE_EPOCH_MARGIN (48L * 3600L)



/*
 * Fixed precision mathematics:
 */
#ifndef TEST
/* #include <Common.h> */

typedef SWord int16;
typedef SDWord int32;
typedef Word uint16;
typedef DWord uint32;
#else
typedef short int16;
typedef long int32;
typedef short unsigned uint16;
typedef long unsigned uint32;
#endif

#define M_PI (3.14159265358979)
#define ABS(x) ((x) < 0 ? -(x) : (x))
/*
 * Fixed: 32 bit fixed precision real, 16 bits of fractional part.
 */
typedef int32 Fixed;
#define FIXED_FRAC_BITS	(16)
#define FIXED_ONE	((int32)1 << FIXED_FRAC_BITS)
#define FIXED_FRAC_MASK	(FIXED_ONE - 1)

#define _REAL2FIXED_P(x) ((Fixed)((x) * FIXED_ONE + 0.5f))
#define _REAL2FIXED(x)  ((x) > 0 ? _REAL2FIXED_P(x) : - _REAL2FIXED_P(-x))
#define FIXED2REAL(f)	((float)(f) / FIXED_ONE)

/*
 * UFixed: unsigned 16 bit fixed precision in [0,1.0)
 */
typedef uint16 UFixed;
#define REAL2UFIXED(x)  ((uint16)((x) * (float)(0x010000) + 0.5))

/*
 * UFixed1: unsigned 16 bit fixed precision in [0,2.0)
 */
typedef uint16 UFixed1;


static inline uint16
ufixed_mul (uint16 x, uint16 y)
{
#ifndef GTD_DO_ASM
  return (uint16)(((uint32)(x) * (y) + 0x8000u) >> 16);
#else
  
  uint16 result;
  asm ("mov.w %1,%0;"
       "mulu.w %2,%0;"
       "addi.l #0x08000,%0;"
       "swap %0"
       : "=&d" (result)
       : "g" ((uint16)x), "g" ((uint16)y) );
  return result;
#endif
}

/* Careful! this can overflow!!! */
static inline uint16
ufixed1_mul (uint16 x, uint16 y)
{
#ifndef GTD_DO_ASM
  return (uint16)(((uint32)(x) * (y) + 0x4000u) >> 15);
#else
  uint16 result;
  asm ("mov.w %1,%0;"
       "mulu.w %2,%0;"
       "addi.l #0x04000,%0;"
       "asl.l #1,%0;"
       "swap %0"
       : "=&d" (result)
       : "g" ((uint16)x), "g" ((uint16)y) );
  return result;
#endif
}

static /* inline */ int32
fixed_mul (int32 x, int32 y)
{
/* FIXME: overflow? */
#ifndef GTD_DO_ASM
  uint32 x_frac = (uint32)(x & FIXED_FRAC_MASK);
  uint32 y_frac = (uint32)(y & FIXED_FRAC_MASK);
  int32  x_int  = (int32)(int16)(x >> FIXED_FRAC_BITS);
  int32  y_int  = (int32)(int16)(y >> FIXED_FRAC_BITS);

  return ( ((x_frac * y_frac + (uint32)(FIXED_ONE / 2)) >> 16)
	   + x_int * y
	   + y_int * x_frac );
#else  
  int32 result;
  asm ("mov.w %1,%0;"  /* (x.low * y.low + 0x8000) >> 16 */
       "mulu.w %2,%0;"
       "addi.l #0x08000,%0;"
       "clr.w %0;"
       "swap %0;"

       "swap %1;"      /* + (x.high * y.low) */
       "mov.w %1,%%d1;"
       "blt 0f;"
       "mulu.w %2,%%d1;"
       "add.l %%d1,%0;"
       "bra 1f;"
       "0:"
       "neg.w %%d1;"
       "mulu.w %2,%%d1;"
       "sub.l %%d1,%0;"
       "1:"

       "swap %2;"      /* + ((x.high * y.high) << 16) */
       "mov.w %1,%%d1;"
       "muls.w %2,%%d1;"
       "swap %%d1;"
       "clr.w %%d1;"
       "add.l %%d1,%0;"

       "swap %1;"      /* + (x.low * y.high) */
       "mov.w %2,%%d1;"
       "blt 2f;"
       "mulu.w %1,%%d1;"
       "add.l %%d1,%0;"
       "bra 3f;"
       "2:"
       "neg.w %%d1;"
       "mulu.w %1,%%d1;"
       "sub.l %%d1,%0;"
       "3:"
       
       "swap %2;"     /* restore %2 */

       : "=&d" (result)
       : "d" (x), "d" (y)
       : "d1");
  return result;

#endif
}

static Fixed
fixed_div (Fixed num, Fixed denom)
{
  int neg = 0, shift = 0;
  uint32 x, y;
  uint16 yhi;
  
  if (num == 0)
      return 0;
  else if (num > 0)
      x = num;
  else
    {
      x = -num;
      neg = 1;
    }
  
  if (denom == 0)
      return neg ? -0x7fffffff : 0x7fffffff;
  else if (denom > 0)
      y = denom;
  else
    {
      y = -denom;
      neg = !neg;
    }

  for (yhi = (uint16)(y >> 16); yhi; yhi >>= 1)
      shift++;
  if (shift)
    {
      x = (x + (1 << (shift-1))) >> shift;
      y = (y + (1 << (shift-1))) >> shift;
    }
 if (y == 0x10000)
     /* Bug fix GTD v. 2.1a: y was all ones --- rounding added an extra bit on the left. */
     /* (y = 0x10000 => y = 1.0) */
     return neg ? -x : x;

#ifndef GTD_DO_ASM
  if (x > fixed_mul(0x7fffffff, y))
      x = 0x7fffffff;
  else
      x = ((x / y) << 16) + ((x % y) << 16) / y;
  assert(x <= 0x7fffffff);
#else  
  asm("divu.w %2, %0;"
      "bvs 1f;"
      "tst.w %0;"
      "bpl 2f;"
      "1:"
      "mov.l #0x7fffffff, %0;"
      "bra 3f;"
      
      "2:"
      "mov.l %0, %%d1;"
      "clr.w %%d1;"
      "divu.w %2, %%d1;"
      
      "swap %0;"
      "mov.w %%d1, %0;"
      "3:"
      
      : "=d" (x)
      : "0" (x), "g" ((uint16)y)
      : "d1");
#endif
  return neg ? -x : x;
}

static inline Fixed
_ufixed1_sqrt(UFixed1 x)
{
  /*
   * Compute sqrt(x) for x in [ 0.5, 2.0 ).
   *
   * Algorithm:
   *
   * Compute u = 1/sqrt(x) using Newton's iterative method:
   *
   *   u    = 0.5 * u  * ( 3 - x * u  * u  )
   *    i+1          i              i    i
   */

  /* This lookup table provides rough guess for the initial value of u. */
  static UFixed1 u0[] = {
      0xab33, /* 1/sqrt( sqrt(0.500 * 0.625) ) */
      0x9ab2, /* 1/sqrt( sqrt(0.625 * 0.750) ) */
      0x8e37, /* 1/sqrt( sqrt(0.750 * 0.875) ) */
      0x8458, /* 1/sqrt( sqrt(0.875 * 1.000) ) */
      0x7c49, /* 1/sqrt( sqrt(1.000 * 1.125) ) */ 
      0x758b, /* 1/sqrt( sqrt(1.125 * 1.250) ) */ 
      0x6fca, /* 1/sqrt( sqrt(1.250 * 1.375) ) */ 
      0x6acf, /* 1/sqrt( sqrt(1.375 * 1.500) ) */
      0x6671, /* 1/sqrt( sqrt(1.500 * 1.625) ) */ 
      0x6291, /* 1/sqrt( sqrt(1.625 * 1.750) ) */
      0x5f1b, /* 1/sqrt( sqrt(1.750 * 1.875) ) */ 
      0x5bfb, /* 1/sqrt( sqrt(1.875 * 2.000) ) */
  };
  UFixed1 delta, u;

  assert (x >= 0x04000);

  u = u0[(x >> 12) - 4];
  /* This loop usually finishes in two passes (rarely three, I don't think ever four) */
  do {
    delta = u;
    u = ufixed1_mul(u, 0x0c000 - ufixed1_mul(u, ufixed1_mul(x,u)) / 2);
    delta = delta > u ? delta - u : u - delta;
  } while (delta >= 200); /* delta < 200 seems to insure delta_{i+1} <= 1 */

  return ((uint32)u * x) >> 14;	/* Renormalize back to a Fixed */
}

static Fixed
fixed_sqrt (Fixed x)
{
  int   shift = 0;
  Fixed res;
  
  assert(x >= 0);

  if (x <= 0)
      return 0;

  /* Prescale to get x in [0.5, 2) */
  while (x >= 2 * FIXED_ONE)
    {
      shift++;
      x >>= 2;
    }
  while (x < FIXED_ONE / 2)
    {
      shift--;
      x <<= 2;
    }

  res = _ufixed1_sqrt(x / 2);

  return shift > 0 ? (res << shift) : (res >> (-shift));
}

/*
 * Angle: real angle in radians scaled by 8/(2*pi);
 *  One rotation = 8.
 */
typedef Fixed Angle;

/* FIXME: rounding? */
#define RADIANS2ANGLE(rad)	\
    (int32)((4 * FIXED_ONE / M_PI) * (rad) + 0.5)
#define DEGREES2ANGLE(rad)	\
    (int32)((4 * FIXED_ONE / 180.0) * (rad) + 0.5)

typedef struct
{
    uint32	h;
    uint16	l;
}			Speed;

/* Convert double-precision speed in cycles per second to Speed. */
/*  speed in [0.0, 1.0) cycles per second, EPS = 3.5e-15. */
#define SPEEDCONV(speed, cps) (						\
    ((speed).h = (uint32)((cps) * (65536.0f * 65536.0f))),		\
    ((speed).l = (uint16)( (((cps) * (65536.0f * 65536.0f))-(speed).h) * 65536.0f)))


static inline int32
speed_mul (Speed * s, uint32 t)
{
  uint32 product;
  
#ifndef GTD_DO_ASM
  product = (s->h * t
	     + s->l * (t >> 16)
	     + ((s->l * (t & 0x0ffffu) + 0x08000u) >> 16));
#else

  asm ("move.w (%1),%0;"	/* %0 = s->h.hi * t.lo */
       "mulu.w %2,%0;"

       "swap %2;"		/* %0 += s->h.lo * t.hi */
       "move.w 2(%1),%%d1;"
       "mulu.w %2,%%d1;"
       "add.w %%d1,%0;"

       "swap %0;"		/* %0 <<= 16 */
       "clr.w %0;"

       "move.w 4(%1),%%d1;"	/* %0 += s->l * t.hi */
       "mulu.w %2,%%d1;"
       "add.l %%d1,%0;"

       "swap %2;"		/* %0 += s->h.lo * t.lo */
       "move.w 2(%1),%%d1;"
       "mulu.w %2,%%d1;"
       "add.l %%d1,%0;"

       "move.w 4(%1),%%d1;"	/* %0 += (s->l * t.lo + 0x8000) >> 16 */
       "mulu.w %2,%%d1;"
       "addi.l #0x08000, %%d1;"
       "clr.w %%d1;"
       "swap %%d1;"
       "add.l %%d1,%0;"

       : "=&d" (product)
       : "a" (s), "d" (t)
       : "d1");

#endif

  return product >> 13;
}

struct annual_coefs
{
    int		year;
    time_t	epoch;
    
    Angle	phi0[MAX_CSTS];
    Fixed	amp[TIDE_MAX_DERIV + 1][MAX_CSTS];
};

struct fixed_coefs
{
    DBIndex	place;
    
    int		num_csts;
    short	order[MAX_CSTS];
    
    Speed	speed[MAX_CSTS];

    float	scale[TIDE_MAX_DERIV+2];
    Fixed	datum;

    Fixed	timescale[TIDE_MAX_DERIV + 1];
    Fixed	blend01, blend12, blend02;
#if TIDE_MAX_DERIV == 3
    Fixed	blend23, blend13, blend03;
#endif
    struct annual_coefs	annuals[2];
    struct annual_coefs * cur;
};


static struct fixed_coefs coefs;

#ifndef TEST

static float
max_node_factor ( void )
{
  static float max_fact = -1.0f;
  int wasyear, myyear, a;

  if (max_fact <= 0.0f)
    {
      if ((wasyear = year) <= 0)
	  wasyear = first_year;

      for (myyear = first_year; myyear < first_year + num_epochs; myyear++)
	{
	  GetYearData(myyear);
	  for (a = 0; a < num_csts; a++)
	      if (cst_nodes[a] > max_fact)
		  max_fact = cst_nodes[a];
	}
      
      if (wasyear >= first_year)
	  GetYearData(wasyear);

      max_fact *= 1.1f;		/* add a safety margin */
    }

  return max_fact;
}

static void figure_datum ( float offset ) {	/* Because we now dick with datum, got to reset it here */
	tide_alert = offset;
	coefs.datum     = _REAL2FIXED((float)(DATUM + offset) / coefs.scale[0]);
      }

void
figure_multipliers ( void )
{
  int a,d;
  float amp, max_node;
  
  if (!loc_amp || !cst_nodes)
      return;		/* If called before data loaded */
  offset_event_good = true;				/* Clear cache */

  assert(num_csts <= MAX_CSTS);
  
  if (year < first_year || year >= first_year + num_nodes)
      barf (MISSINGYEAR);

  if (coefs.num_csts == 0 || coefs.place != PlaceIndex)
    {
      /* We must do this stuff once every time the station changes. */
      coefs.place = PlaceIndex;

      coefs.num_csts = 0;
      for (d = 0; d <= TIDE_MAX_DERIV + 1; d++)
	  coefs.scale[d] = 0.0f;

      max_node = max_node_factor();

      for (a = 0; a < num_csts; a++)
	{
	  if (loc_amp[a] == 0.0f)
	      continue;		/* Skip the constituents with zero amplitude. */

	  /* Convert cst_speed to fixed precision */
	  /* FIXME: if this is slow, the cst_speeds never change, so we could just do this once... */
	  SPEEDCONV(coefs.speed[coefs.num_csts], cst_speeds[a] / 360.0f);

	  /* Keep a list of the non-zero constituents */
	  coefs.order[coefs.num_csts++] = a;

	  /* Compute scale for each of the derivatives. */
	  amp = loc_amp[a] * max_node / TIDE_MAX_SCALED_TIDE;
	  coefs.scale[0] += amp;
	  for (d = 1; d <= TIDE_MAX_DERIV + 1; d++)
	    {
	      amp *= cst_speeds[a] * ((float)M_PI / 180.0f);
	      coefs.scale[d] += amp;
	    }
	}
      coefs.order[coefs.num_csts] = -1;	/* flag end of list */

      /* Various other precomputed factors: */
	figure_datum(0.0);

      for (d = 0; d <= TIDE_MAX_DERIV; d++)
	  coefs.timescale[d] = _REAL2FIXED_P(coefs.scale[d]
					     / (coefs.scale[d+1] * TIDE_TIMESCALE));
      
      coefs.blend01 = _REAL2FIXED_P( coefs.scale[0]
				     / (coefs.scale[1]
					* (float)TIDE_BLEND_TIME) );
      coefs.blend12 = _REAL2FIXED_P( 2.0f * coefs.scale[1]
				     / (coefs.scale[2]
					* (float)TIDE_BLEND_TIME) );
      coefs.blend02 = _REAL2FIXED_P( coefs.scale[0]
				     / (coefs.scale[2]
					* (float)TIDE_BLEND_TIME
					* TIDE_BLEND_TIME) );
#if TIDE_MAX_DERIV == 3
      coefs.blend23 = _REAL2FIXED_P( 3.0f * coefs.scale[2]
				     / (coefs.scale[3]
					* (float)TIDE_BLEND_TIME) );
      coefs.blend13 = _REAL2FIXED_P( 3.0f * coefs.scale[1]
				     / (coefs.scale[3]
					* (float)TIDE_BLEND_TIME
					* TIDE_BLEND_TIME) );
      coefs.blend03 = _REAL2FIXED_P( coefs.scale[0]
				     / (coefs.scale[3]
					* (float)TIDE_BLEND_TIME
					* TIDE_BLEND_TIME
					* TIDE_BLEND_TIME) );
#endif      
      coefs.cur = coefs.annuals;
      coefs.annuals[0].year = -1;
      coefs.annuals[1].year = -1;
    }

  if (coefs.cur->year == year)
      return;

  /* Switch to the other set of coefficients */
  coefs.cur = coefs.annuals + (coefs.cur == coefs.annuals ? 1 : 0);

  if (coefs.cur->year == year)
      return;
  
  /* Use a epoch a little earlier, so that we don't need to deal with
   *  t < epoch */
  coefs.cur->epoch = epoch - TIDE_EPOCH_MARGIN - tmeridian;
  coefs.cur->year = year;
  
  for (a = 0; a < coefs.num_csts; a++)
    {
      int cst = coefs.order[a];
      amp = loc_amp[cst] * cst_nodes[cst];
      coefs.cur->amp[0][a] = _REAL2FIXED_P(amp / coefs.scale[0]);
      for (d = 1; d <= TIDE_MAX_DERIV; d++)
	{
	  amp *= cst_speeds[cst] * ((float)M_PI / 180.0f);
	  coefs.cur->amp[d][a] = _REAL2FIXED_P(amp / coefs.scale[d]);
	}

      coefs.cur->phi0[a] = DEGREES2ANGLE( 360.0f
					  + ( (90.0f/1024.0f)
					      * (cst_epochs[cst] - loc_epoch[cst])
					      ) );
      coefs.cur->phi0[a] -= speed_mul(coefs.speed + a, TIDE_EPOCH_MARGIN);
    }
}
#endif

static Fixed
_fixed_cos(Angle theta)
{
  /*
   * Input:
   *   theta is a 32 bit fixed-prec number (16 fractional bits) in
   *   units of octants.  ( theta = 0x00080000 == 2*pi radians )
   *
   * Output:
   *   The cosine of theta, as a fixed-precision 32-bit (16 fractional bits)
   *   number.  This routine is accurate to 2^-16.
   */
   
#define PI4 (M_PI / 4.0)
#define FIXED_SQRT_HALF _REAL2FIXED_P(0.7071067811865)
  
#define C_COS1  REAL2UFIXED(0.49973575582035 * PI4 * PI4)
#define C_COS2  REAL2UFIXED(0.04039092637465 * PI4 * PI4 * PI4 * PI4)
#define C_SIN0	REAL2UFIXED(PI4)	
#define C_SIN1  REAL2UFIXED(0.16662462953394 * PI4 * PI4 * PI4)
#define C_SIN2  REAL2UFIXED(0.00814383708387 * PI4 * PI4 * PI4 * PI4 * PI4)

  int octant;
  UFixed x, x2;
  Fixed cosval;

  octant = theta >> FIXED_FRAC_BITS;

  x = (uint16)theta;		/* fractional part. */
  
  if (x == 0)
    {
      /* Special cases */
      if ((octant & 1) == 0)
	  cosval = (octant & 2) ? 0 : FIXED_ONE;
      else
	  cosval = FIXED_SQRT_HALF;
    }
  else 
    {
      /* General case */
      if ((octant & 1) != 0)
	  x = ~x + 1;		/* pi/4 - x */
      x2 = ufixed_mul(x,x);

      if (((octant + 1) & 2) == 0)
	{
	  /* Compute cos(x) */
	  cosval = FIXED_ONE - ufixed_mul(x2, C_COS1 - ufixed_mul(x2, C_COS2));
	}
      else
	{
	  /* Compute sin(x) */
	  cosval = ufixed_mul(x,
			      C_SIN0
			      - ufixed_mul(x2,
					   C_SIN1
					   - ufixed_mul(x2, C_SIN2)));
	}
    }

  return  (((octant + 2) & 4) == 0) ? cosval : -cosval;
}

float
fixed_cos (double arg)
{
  return FIXED2REAL(_fixed_cos(RADIANS2ANGLE(arg)));
}

float
fixed_sin (double arg)
{
  return FIXED2REAL(_fixed_cos(RADIANS2ANGLE(arg) + 6 * FIXED_ONE));
}

#ifndef TEST
/*
 * We will need a function for tidal height as a function of time
 * which is continuous (and has continuous first and second derivatives)
 * for all times.
 *
 * Since the epochs & multipliers for the tidal constituents change
 * with the year, the regular time2tide(t) function has small
 * discontinuities at new years.  These discontinuities really
 * fry the fast root-finders.
 *
 * We will eliminate the new-years discontinuities by smoothly
 * interpolating (or "blending") between the tides calculated with one
 * year's coefficients, and the tides calculated with the next year's
 * coefficients.
 *
 * i.e. for times near a new years, we will "blend" a tide
 * as follows:
 *
 * tide(t) = tide(year-1, t)
 *                  + w((t - t0) / Tblend) * (tide(year,t) - tide(year-1,t))
 *
 * Here:  t0 is the time of the nearest new-year.
 *        tide(year-1, t) is the tide calculated using the coefficients
 *           for the year just preceding t0.
 *        tide(year, t) is the tide calculated using the coefficients
 *           for the year which starts at t0.
 *        Tblend is the "blending" time scale.  This is set by
 *           the macro TIDE_BLEND_TIME, currently one hour.
 *        w(x) is the "blending function", whice varies smoothly
 *           from 0, for x < -1 to 1 for x > 1.
 *
 * Derivatives of the blended tide can be evaluated in terms of derivatives
 * of w(x), tide(year-1, t), and tide(year, t).  The blended tide is
 * guaranteed to have as many continuous derivatives as w(x).  */

typedef enum which_derivs {
    DERIV0 = 1,
    DERIV1 = 2,
    DERIV2 = 4,
    DERIV3 = 8,
} derivs_t;
    
struct tide_pt_s 
{
    time_t	t;
    Fixed	f[TIDE_MAX_DERIV+1];
    derivs_t	valid;
};
typedef struct tide_pt_s	tide_pt_t;


/* time2dt_tide(time_t t, int n)
 *
 *   Calculate nth time derivative the normalized tide.
 *
 * Notes: This function does not check for changes in year.
 *  This is important to our algorithm, since for times near
 *  new years, we interpolate between the tides calculated
 *  using one years coefficients, and the next years coefficients.
 *
 *  Except for this detail, time2dt_tide(t,0) should return a value
 *  identical to time2tide(t).
 */

static const tide_pt_t *
_time2dt_tide(tide_pt_t * pt, derivs_t needed)
{
  int		i;
  uint32	tt	= pt->t - coefs.cur->epoch;
  Angle		arg;
  Fixed		cosval, sinval;
  
  assert (pt->t > coefs.cur->epoch);
  pt->f[0] = coefs.datum;
  pt->f[1] = pt->f[2] = 0;
#if TIDE_MAX_DERIV == 3	  
  pt->f[3] = 0;
#endif  

  for (i = 0; i < coefs.num_csts; i++)
    {
      arg = speed_mul(coefs.speed + i, tt) + coefs.cur->phi0[i];
      if ((needed & (DERIV0 | DERIV2)) != 0)
	{
	  cosval = _fixed_cos(arg);
	  if ((needed & DERIV0) != 0)
	      pt->f[0] += fixed_mul(coefs.cur->amp[0][i], cosval);
	  if ((needed & DERIV2) != 0)
	      pt->f[2] -= fixed_mul(coefs.cur->amp[2][i], cosval);
	}
      if ((needed & (DERIV1 | DERIV3)) != 0)
	{
	  sinval = _fixed_cos(arg + 6 * FIXED_ONE);
#if TIDE_MAX_DERIV == 3	  
	  if ((needed & DERIV3) != 0)
	      pt->f[3] += fixed_mul(coefs.cur->amp[3][i], sinval);
	  if ((needed & DERIV1) != 0)
#endif	      
	      pt->f[1] -= fixed_mul(coefs.cur->amp[1][i], sinval);
	}
    }
  pt->valid = needed;
  
  return pt;
}


/* blend_weight (double x, int deriv)
 *
 * Returns the value nth derivative of the "blending function" w(x):
 *
 *   w(x) =  0,     for x <= -1
 *
 *   w(x) =  1/2 + (15/16) x - (5/8) x^3 + (3/16) x^5,
 *                  for  -1 < x < 1
 *
 *   w(x) =  1,     for x >= 1
 *
 * This function has the following desirable properties:
 *
 *    w(x) is exactly either 0 or 1 for |x| > 1
 *
 *    w(x), as well as its first two derivatives are continuous for all x.
 */

static inline void
get_blend_weights (Fixed * w, Fixed x, derivs_t needed)
{
  UFixed xabs, x2, y, y2;
#if TIDE_MAX_DERIV == 3  
  UFixed y3;
#endif  
  
  if (x >= 0)
    {
      if (x >= FIXED_ONE)
	{
	  w[0] = FIXED_ONE;
	  w[1] = w[2] = 0;
	  return;
	}
      xabs = (UFixed) x;
    }
  else
    {
      if (x <= -FIXED_ONE)
	{
	  w[0] = w[1] = w[2] = 0;
	  return;
	}
      xabs = (UFixed)(-x);
    }
  
  x2 = ufixed_mul(xabs, xabs);
  if (x2 == 0)
    {
#if TIDE_MAX_DERIV < 3
      w[0] = FIXED_ONE / 2 + fixed_mul(x, _REAL2FIXED_P(15.0/16.0));
      w[1] = _REAL2FIXED_P(15.0/16.0);
      w[2] = fixed_mul(-_REAL2FIXED_P(15.0/4.0), x);
#else
      w[0] = FIXED_ONE / 2 + fixed_mul(x, _REAL2FIXED_P(35.0/32.0));
      w[1] = _REAL2FIXED_P(35.0/32.0);
      w[2] = fixed_mul(-_REAL2FIXED_P(105.0/16.0), x);
      w[3] = -_REAL2FIXED_P(105.0/16.0);
#endif
      return;
    }
  
  y = ~x2 + 1;
  y2 = ufixed_mul((UFixed)y, (UFixed)y);
#if TIDE_MAX_DERIV == 3
  y3 = ufixed_mul((UFixed)y, (UFixed)y2);
#endif      
  
  if ((needed & DERIV0) != 0)
    {
#if TIDE_MAX_DERIV < 3
      /* 1/2 + x * ( 1/2 + (1/4) * y + (3/16) * y^2 )
       *  where y = (1 - x2) */
      w[0] = FIXED_ONE / 2 + fixed_mul(x,
				       REAL2UFIXED(0.5)
				       + y / 4 + y2 / 8 + y2 / 16);
#else  
      /* 1/2 + x * ( 1/2 + (1/4) * y + (3/16) * y^2 + 5/32 y^3 )
       *  where y = (1 - x2) */
      w[0] = FIXED_ONE / 2 + fixed_mul(x,
				       _REAL2FIXED_P(0.5)
				       + y / 4
				       + y2 / 8 + y2 / 16
				       + y3 / 8 + y3 / 32);
#endif      
    }
  if ((needed & DERIV1) != 0)
    {
#if TIDE_MAX_DERIV < 3
      /* (15/16) * (1 - x^2)^2 */
      w[1] = ufixed_mul(REAL2UFIXED(15.0/16.0), y2);
#else
      /* (35/32) * y^3 */
      w[1] = fixed_mul(_REAL2FIXED_P(35.0/32.0), y3);
#endif      
    }
  if ((needed & DERIV2) != 0)
    {
#if TIDE_MAX_DERIV < 3
      /* -(15/4) * x * (1 - x^2) */
      w[2] = fixed_mul(-_REAL2FIXED_P(15.0/4.0), fixed_mul(x, y));
#else
      /* -(105/16) * x * y^2 */
      w[2] = fixed_mul(-_REAL2FIXED_P(105.0/16.0), fixed_mul(x, y2));
#endif  
    }
#if TIDE_MAX_DERIV == 3  
  if ((needed & DERIV3) != 0)
    {
      /* (105/16) * (4 * y - 5 * y^2) */
      w[3] = (fixed_mul(_REAL2FIXED_P(105.0/4.0), y)
	      - fixed_mul(_REAL2FIXED_P(525.0/16.0), y2));
    }
#endif
}

/*
 * This function does the actual "blending" of the tide
 * and its derivatives.
 */
static const tide_pt_t *
blend_tide (tide_pt_t * pt, derivs_t needed, int first_year, Fixed blend)
{
  tide_pt_t	l, r;
  derivs_t	myneeds = needed;
  Fixed		w[TIDE_MAX_DERIV + 1];

#if TIDE_MAX_DERIV == 3
  if ((needed & DERIV3) != 0)
      myneeds |= DERIV0 | DERIV1 | DERIV2;
  else
#endif  
  if ((needed & DERIV2) != 0)
      myneeds |= DERIV0 | DERIV1;
  else if ((needed & DERIV1) != 0)
      myneeds |= DERIV0;

  r.t = l.t = pt->t;
  
  /*
   * If we are already happy_new_year()ed into one of the two years
   * of interest, compute that years tide values first.
   */
  if (year == first_year + 1)
    {
      _time2dt_tide(&r, myneeds);
      happy_new_year(first_year);
      _time2dt_tide(&l, myneeds);
    }
  else
    {
      if (year != first_year)
	  happy_new_year(first_year);
      _time2dt_tide(&l, myneeds);
      happy_new_year(first_year + 1);
      _time2dt_tide(&r, myneeds);
    }

  get_blend_weights(w, blend, myneeds);

  /*
   * Do the blending.
   */
  
  if ((needed & DERIV0) != 0)
      pt->f[0] = l.f[0] + fixed_mul(w[0], r.f[0] - l.f[0]);
  if ((needed & DERIV1) != 0)
      pt->f[1] = ( l.f[1] + fixed_mul(w[0], r.f[1] - l.f[1])
		   + fixed_mul(coefs.blend01, fixed_mul(w[1], r.f[0] - l.f[0])) );
  if ((needed & DERIV2) != 0)
      pt->f[2] = ( l.f[2] + fixed_mul(w[0], r.f[2] - l.f[2])
		   + fixed_mul(coefs.blend12, fixed_mul(w[1], r.f[1] - l.f[1]))
		   + fixed_mul(coefs.blend02, fixed_mul(w[2], r.f[0] - l.f[0])) );
#if TIDE_MAX_DERIV == 3
  if ((needed & DERIV2) != 0)
      pt->f[2] = ( l.f[2] + fixed_mul(w[0], r.f[3] - l.f[3])
		   + fixed_mul(coefs.blend23, fixed_mul(w[1], r.f[2] - l.f[2]))
		   + fixed_mul(coefs.blend13, fixed_mul(w[2], r.f[1] - l.f[1]))
		   + fixed_mul(coefs.blend03, fixed_mul(w[3], r.f[0] - l.f[0])) );
#endif
  pt->valid = needed;
  return pt;
}


static const tide_pt_t *
time2dt_tide (tide_pt_t * pt, derivs_t needed)
{
  /* FIXME: maybe don't do this yet? */
  possible_happy_new_year(pt->t);	/* make sure we're in the right epoch */

//printf("%ld %ld\n%ld\n",pt->t,epoch,next_epoch);
  assert(pt->t >= epoch);
  assert(pt->t < next_epoch);

  /*
   * If we're close to either the previous or the next
   * new years we must blend the two years tides.
   */
  if (pt->t - epoch <= TIDE_BLEND_TIME && year > first_year)
      return blend_tide(pt, needed, year - 1,
			_REAL2FIXED_P((float)(pt->t - epoch)/TIDE_BLEND_TIME));
  else if (next_epoch - pt->t <= TIDE_BLEND_TIME
           && year + 1 < first_year + num_epochs)
      return blend_tide(pt, needed, year,
                        -_REAL2FIXED_P((float)(next_epoch - pt->t)
				       /TIDE_BLEND_TIME));
  /*
   * Else, we're far enough from newyears to ignore the blending.
   */
  return _time2dt_tide(pt, needed);
}

static inline const tide_pt_t *
check_tide_pt (const tide_pt_t * pt, derivs_t needed)
{
  tide_pt_t * volp = (tide_pt_t *) pt; /* const cast */
  return  ((needed & ~pt->valid) != 0) ? time2dt_tide(volp, pt->valid | needed) : pt;
}

/* find_zero (time_t t1, time_t t2, double (*f)(time_t t, int deriv))
 *   Find a zero of the function f, which is bracketed by t1 and t2.
 *   Returns a value which is either an exact zero of f, or slightly
 *   past the zero of f.
 */
static const tide_pt_t * 
find_zero (const tide_pt_t * pt1, const tide_pt_t * pt2, int deriv)
{
  static tide_pt_t	buf[3];
  tide_pt_t *	left	= buf;
  tide_pt_t *	right	= buf + 1;
  tide_pt_t *	next	= buf + 2;
  tide_pt_t *	new;

  derivs_t myneeds = DERIV1 | (deriv ? DERIV2 : DERIV0);
  
  int32  dt;
  Fixed f_thresh, abs_ft, tleft, tright, mft, f, fp;
  Boolean flip;
  
  assert(pt1->t < pt2->t);

  if (pt2->t - pt1->t <= TIDE_TIME_PREC)
      return pt2;

  *left = *pt1;
  *right = *pt2;
  
  check_tide_pt(left, myneeds);
  check_tide_pt(right, myneeds);
  
  flip = left->f[deriv] > 0;

  assert(left->f[deriv] != 0 && right->f[deriv] != 0);
  assert(flip ? right->f[deriv] < 0 : right->f[deriv] > 0);

  /* Pick best end, from which we will attempt first Newton step. */
  f_thresh = ABS(left->f[deriv]);
  abs_ft = ABS(right->f[deriv]);
  if (abs_ft > f_thresh)
    {
      f_thresh = abs_ft;
      new = left;
    }
  else
      new = right;

  while (new->f[deriv] != 0 && right->t - left->t > TIDE_TIME_PREC)
    {
      abs_ft = ABS(new->f[deriv]);
      dt = 0;

      /* Only attempt Newton step if value is decreasing fast enough: */
      if (abs_ft < f_thresh)
	{
	  f = new->f[deriv];
	  if ((fp = new->f[deriv+1]) < 0)
	    {
	      fp = -fp;
	      f = -f;
	    }

	  mft = - fixed_mul(f, coefs.timescale[deriv]);
	  tright = (right->t - new->t) * (FIXED_ONE / TIDE_TIMESCALE);
	  tleft = (new->t - left->t) * (FIXED_ONE / TIDE_TIMESCALE);

	  /* Only attempt Newton step if it looks like step will land in bounds: */
	  if (mft <= fixed_mul(tright, fp) && -mft <= fixed_mul(tleft, fp))
	    {
	      dt = fixed_div(mft, fp) / (FIXED_ONE / TIDE_TIMESCALE);

	      /* Since our goal specifically is to reduce our bracket size
		 as quickly as possible (rather than getting as close to
		 the zero as possible) we should ensure that we don't take
		 steps which are too small.  (We'd much rather step over
		 the root than take a series of steps which approach the
		 root rapidly but from only one side.) */
	      if (dt < TIDE_TIME_PREC && dt > -TIDE_TIME_PREC)
		dt = f < 0 ? TIDE_TIME_PREC : -TIDE_TIME_PREC;

	      next->t = new->t + dt;
	      f_thresh = abs_ft / 2;
	    }
	}
      
      if (dt == 0 /* bad step */
	  || next->t >= right->t || next->t <= left->t /* newton step too big */)
        {
          /* Newton step failed, do bisection */
          next->t = left->t + (right->t - left->t) / 2;

	  f_thresh = ABS(left->f[deriv]);
	  abs_ft = ABS(right->f[deriv]);
	  if (abs_ft > f_thresh)
	      f_thresh = abs_ft;
	  //printf ("MIDPOINT\n");
        }

      /* Compute next point */
      time2dt_tide(new = next, myneeds);

      /* Narrow the bracket */
      if (flip ? new->f[deriv] > 0 : new->f[deriv] < 0)
	  next = left, left = new;
      else
	  next = right, right = new;
    }

  return right;		/* found root */
}


/* next_zero(time_t t, double (*f)(), double max_fp, double max_fpp)
 *   Find the next zero of the function f which occurs after time t.
 *   The arguments max_fp and max_fpp give the maximum possible magnitudes
 *   that the first and second derivative of f can achieve.
 *
 *   Algorithm:  Our goal here is to bracket the next zero of f ---
 *     then we can use find_zero() to quickly refine the root.
 *     So, we will step forward in time until the sign of f changes,
 *     at which point we know we have bracketed a root.
 *     The trick is to use large steps in are search, which making
 *     sure the steps are not so large that we inadvertently
 *     step over more than one root.
 *
 *     The big trick, is that since the tides (and derivatives of
 *     the tides) are all just harmonic series', it is easy to place
 *     absolute bounds on their values.
 */

static const tide_pt_t *
next_zero (const tide_pt_t * pt, int deriv, int * risingflag, Boolean backwards)
{
  tide_pt_t buf[2];
  tide_pt_t *	left	= buf;
  tide_pt_t *	right	= buf + 1;
  tide_pt_t *	swap_tmp;
/* #define GTD_DEBUG_NEXT_ZERO */
#ifdef GTD_DEBUG_NEXT_ZERO
  time_t start = pt->t;
#endif
#if TIDE_MAX_DERIV < 3  
  derivs_t myneeds = DERIV1 | (deriv ? DERIV2 : DERIV0);
#else
  derivs_t myneeds = DERIV1 | DERIV2 | (deriv ? DERIV3 : DERIV0);
  Fixed ft2, t_tp, ft2sq;
#endif  
  int32 small_step = backwards ? -TIDE_TIME_PREC : TIDE_TIME_PREC;
  int32 step;
  Fixed ft, ft1, t_0, ft1sq, term1;
  Boolean flip;

  *left = *pt;
  check_tide_pt(left, myneeds);
  
  /* If we start at a zero, step forward until we're past it. */
  while (left->f[deriv] == 0)
    {
      left->t += small_step;
      time2dt_tide(left, myneeds);
    }
  
  flip = !(*risingflag = left->f[deriv] < 0);
  if (backwards)
      *risingflag = !*risingflag;

#ifdef GTD_DEBUG_NEXT_ZERO
  printf ("START: %lu %e\n", start, FIXED2REAL(left->f[deriv]));
#endif
  
  while (1)
    {
      /* These times are non-dimensionalized.
       * they need to be scaled by (TIDE_TIMESCALE / TIDE_MAX_SCALED_TIDE)
       * to get real seconds. */

      /* (We proceed, for the while, on the assumption at f' is
       *  of the "right" sign.  i.e. f is trending towards zero.)
       */
      
      /* Minimum bound on time to next turning point:
       *
       *    t   = |f'| / sup(f'');
       *     tp
       */
      ft1 = fixed_mul(ABS(left->f[deriv+1]), coefs.timescale[deriv+1]);


      /* A minimum bound on time to next zero:
       *
       *    t_0  = |f| / sup(f')
       */
      ft  = fixed_mul(ABS(left->f[deriv]), coefs.timescale[deriv]);

      /* Alternative minimum bound on time to next zero:
       *  NOTE: It may well pay off to do a computation like this one for
       *    for the time to turning point as well.   However this
       *    would require the computation of one more derivative
       *    than we currently are set up to handle.
       *
       *                           2                     1/2
       *    t_0 = ( (|f'|/sup(f''))  + 2 * |f|/sup(f'') )     - |f'|/sup(f'')
       *
       * For our step, we will take the max(tau_tp, tau_0, tau_1).
       * This guarantees that we will step past at most one zero.
       */
      ft1sq = fixed_mul(ft1,ft1);
      term1 = fixed_mul(ft, coefs.timescale[deriv+1]) * TIDE_MAX_SCALED_TIDE;
      if (ft1sq < term1)
	{
	  t_0 = ( fixed_sqrt(ft1sq + 2 * term1) - ft1 );
	  if (t_0 < ft)
	      t_0 = ft;
	}
      else
	{
	  /* in this case, t_0 is guaranteed to be less than ft1
	   * (time to turning point) so there's no point in computing
	   * it. */
	  t_0 = ft;
	}

      if (t_0 < ft1)
	{
	  /* Use the greater of (time-to-zero, time-to-turning-point)
	   * as our step-size */
	  t_0 = ft1;
	}

#if TIDE_MAX_DERIV == 3
      ft2 = fixed_mul(ABS(left->f[deriv+2]), coefs.timescale[deriv+2]);
      ft2sq = fixed_mul(ft2,ft2);
      term1 = fixed_mul(ft1, coefs.timescale[deriv+2]) * TIDE_MAX_SCALED_TIDE;
      t_tp = ( fixed_sqrt(ft2sq + 2 * term1) - ft2 );

      if (t_0 < t_tp)
	{
	  /* Use the greater of (time-to-zero, time-to-turning-point)
	   * as our step-size */
	  //printf ("t_tp wins: %ld -> %ld\n", t_0, t_tp);
	  t_0 = t_tp;
	}
      //else
	  //printf ("t_tp lose: %ld -> %ld\n", t_0, t_tp);
#endif

      /* tau_0 now contains the best bound */
      
      if (*risingflag ? left->f[deriv+1] < 0 : left->f[deriv+1] > 0)
        {
          /*
	   * Derivative is in the "wrong" direction.
	   * (f is trending away from zero.)
	   * The logic used up to now assumed that
	   * f' was in the "right" direction.
	   *
	   * Now before we can get to a zero we have to go through
	   * a turning point, and come back to where we are now,
	   * then proceed in the "right" direction towards zero.
	   *
	   * It will take at least 2 * t_tp to go through a turning
	   * point and get back to where we are now, at which time
	   * the maximum value |f'| can have is the current value
	   * of |f'|.   So at that time tau_0 will still be a good
	   * step size...
	   */
	  t_0 += 2 * ft1;
        }

      /* Scale from our scaled time to seconds */
      step = t_0  / (TIDE_MAX_SCALED_TIDE * FIXED_ONE / TIDE_TIMESCALE);
      
      if (step < TIDE_TIME_STEP)
          step = TIDE_TIME_STEP; /* No rediculously small steps... */
      if (backwards)
	  step = -step;

      right->t = left->t + step;
      time2dt_tide(right, myneeds);

#ifdef GTD_DEBUG_NEXT_ZERO
      printf ("STEP: %ld %e\n", step, FIXED2REAL(right->f[deriv]));
#endif

      /*
       * If we hit upon an exact zero, step right until we're off
       * the zero.  If the sign has changed, we are bracketing a desired
       * root, if the sign hasn't changed, then the zero was at
       * an inflection point (i.e. a double-zero to within TIDE_TIME_PREC)
       * and we want to ignore it.
       */
      while (right->f[deriv] == 0)
	{
	  right->t += small_step;
	  time2dt_tide(right, myneeds);
	}
      
      if (flip ? right->f[deriv] < 0 : right->f[deriv] > 0)
	{
          /* Found a bracket, call Newton finder to refine the root. */
	  return ( backwards
		   ? find_zero(right, left, deriv)
		   : find_zero(left, right, deriv) );
	}

      swap_tmp = left;
      left = right;
      right = swap_tmp;
    }
}


static const tide_pt_t * 
next_high_or_low_tide (const tide_pt_t * pt, int * hiflag, Boolean backwards)
{
  /* FIXME: if this is the only thing that calls next_zero, we should get rid of next_zero(). */
  int rising = 0;
  const tide_pt_t * next = next_zero(pt, 1, &rising, backwards);
  *hiflag = !rising;
  return next;
}

static const tide_pt_t *
find_mark_crossing (const tide_pt_t * pt1, const tide_pt_t * pt2, int * risingflag)
{
  check_tide_pt(pt1, DERIV0);
  check_tide_pt(pt2, DERIV0);
  
  assert(pt1->t < pt2->t);

  *risingflag = pt1->f[0] <= 0 && pt2->f[0] >= 0;
  if (!(*risingflag) && !(pt1->f[0] >= 0 && pt2->f[0] <= 0))
      return 0;			/* no crossing. */
  
  if (pt1->f[0] == 0)
      return pt1;
  else if (pt2->f[0] == 0)
      return pt2;

  return find_zero(pt1, pt2, 0);
}

/* Next high tide, low tide, transition of the mark level, or some
   combination.
       Bit      Meaning
        0       low tide
        1       high tide
        2       falling transition
        3       rising transition
*/
struct tide_event_s 
{
    time_t	t;
    int		type;
    tide_pt_t	cache;
};
typedef struct tide_event_s tide_event_t;

static inline void
init_event (tide_event_t * evnt, time_t start_t)
{
  evnt->cache.t = start_t;
  evnt->type = evnt->cache.valid = 0;
}

static time_t
next_big_event (tide_event_t * evnt)
{
  tide_pt_t	next_hilo;
  int		is_rising;
  int		is_high;
  const tide_pt_t * slack;

  if ((evnt->type & 0x0c) != 0 && evnt->cache.valid)
    {
      /* last event was a slack, and cache contains a valid event */
      evnt->type >>= 2;
      evnt->t = evnt->cache.t;
    }
  else
    {
      /* Find time of next high or low tide */
      next_hilo = *next_high_or_low_tide(&(evnt->cache), &is_high, false);
      
      if ( need_zeros
	   && ( (slack = find_mark_crossing(&(evnt->cache), &next_hilo, &is_rising)) != 0) )
	{
	  assert (is_rising ? is_high : !is_high);
	  evnt->type = is_rising ? 0x08 : 0x04;
	  evnt->t = slack->t;
	}
      else
	{
	  evnt->type = is_high ? 0x02 : 0x01;
	  evnt->t = next_hilo.t;
	}

      evnt->cache = next_hilo;
    }
  
  return evnt->t;
}

static time_t
previous_hilo_event (tide_event_t * evnt)
{
  int    is_high;
  /* Find previous high/low tide */
  evnt->cache.t -= TIDE_TIME_PREC;
  evnt->cache.valid = 0;
  
  evnt->cache = *next_high_or_low_tide(&(evnt->cache), &is_high, true);
  evnt->type = is_high ? 0x02 : 0x01;
  return evnt->t = evnt->cache.t;
}

float
time2atide (time_t t)
{
  tide_pt_t tide;
  tide.t = t;
  time2dt_tide(&tide, DERIV0);
  return FIXED2REAL(tide.f[0] * coefs.scale[0]);
}


struct corrected_event_s 
{
    tide_event_t event;

    enum TideEvent eventno;
    int		etype;
    
    int	offset;
    float	afactor;			/* Allow both mult. and add offsets - WB 8/00 */
    float	mfactor;

    time_t	t;
    float	currtide;
};
typedef struct corrected_event_s corrected_event_t;

time_t correct_event (corrected_event_t *);

time_t
correct_event (corrected_event_t * tide)
{
  /* now, next_big_event only returns one event at a time */
  if (tide->event.type & 1)
      tide->eventno = dis_low;
  else if (tide->event.type & 2)
      tide->eventno = dis_high;
  else if (tide->event.type & 4)
      tide->eventno = dis_fall;
  else if (tide->event.type & 8)
      tide->eventno = dis_rise;
  else
      return TIDE_BAD_TIME;

  tide->etype = tide->eventno & 3;
  tide->currtide = (!iscurrent || (tide->event.type & 3)) ? time2atide(tide->event.t) : SLACKER;
  if (PlaceOff)
    {
      tide->offset = PlaceOff->offsets[tide->etype];
      tide->t = tide->event.t + tide->offset * 60L;
      if (tide->etype <= 1)
	{
	  tide->mfactor = PlaceOff->mulfactors[tide->etype] * 0.01f;
	  tide->afactor = PlaceOff->addfactors[tide->etype] * 0.01f;
	  tide->currtide = ( tide->currtide * tide->mfactor ) + tide->afactor;
	}
    }
  else
    {
      tide->offset = 0;
      tide->t = tide->event.t;
      tide->mfactor = 1.0f;
      tide->afactor = 0.0f;
    }
  return tide->t;
}


/* Calculate tide at time t, possibly interpolating for offset.  
   If interpolated value not yet available, return SLACKER (funny value).
   Set *s to slope. 
*/

float
TideAnyplace(GMTime_t t, int *s)
{
  static corrected_event_t offset_event_prev, offset_event_next;
  static float mfactor, afactor;				/* Caches between calls */
  tide_pt_t tide;
  static int offset;
  float r0;

  figure_datum(0.0);						/* Make sure no alert applied */
  if (PlaceOff)
    {								
	if (!offset_event_good || 				/* If cache doesn't apply */
			offset_event_prev.t > t || offset_event_next.t < t)
	   {								/* then recompute the bracketing events */
	      init_event(&offset_event_next.event, t);
      
     	 	do
		  previous_hilo_event(&offset_event_next.event);
      	while (correct_event(&offset_event_next) >= t);
      
      	do
		{
		  offset_event_prev = offset_event_next;
		  next_big_event(&offset_event_next.event);
		}
      	while (correct_event(&offset_event_next) < t);

		offset_event_good = true;
	   }
     	r0 = (float)(t - offset_event_prev.t) / (offset_event_next.t - offset_event_prev.t);
	offset = (int)(r0 * offset_event_next.offset + (1.0f - r0) * offset_event_prev.offset);

	/* Following kludge contributed by Jeff Dairiki 12/00 to repair Morris Jesup graph horror */
	/* FIXME: inefficiencies */
	/* FIXME: check for instabilities (small delta tide) */
	{
	  /* Compute tide at guess */
	  float guess_h = time2atide(t - offset * 60L);
	  float left_h  = time2atide(offset_event_prev.event.t);
	  float right_h = time2atide(offset_event_next.event.t);
	  r0 = (guess_h - left_h) / (right_h - left_h);
	}
     	offset = (int)(r0 * offset_event_next.offset + (1.0f - r0) * offset_event_prev.offset);

     	/* If either event is slack  use only the non-slack factor */
     	if (offset_event_prev.eventno > 1) {
	  mfactor = offset_event_next.mfactor;
 	  afactor = offset_event_next.afactor; }
     	else if (offset_event_next.eventno > 1) {
	  mfactor = offset_event_prev.mfactor;
	  afactor = offset_event_prev.afactor; }
     	else {
	  mfactor = r0 * offset_event_next.mfactor + (1.0f - r0) * offset_event_prev.mfactor;
	  afactor = r0 * offset_event_next.afactor + (1.0f - r0) * offset_event_prev.afactor; }
    }
  else
    {
      offset = 0;
      mfactor = 1.0;
      afactor = 0;
    }
  
  tide.t = t - offset * 60L;
  time2dt_tide(&tide, DERIV0 | DERIV1);

  *s = tide.f[1] > 0;

  return FIXED2REAL(tide.f[0] * coefs.scale[0]) * mfactor + afactor;
}



/* FIXME: there are some problems with offsets.
 *
 * What if two events happen close enough together that their
 * offsets cause them to reverse order?
 */


/* Compute and display a day's worth of tide or current events. */
void
list_tides (void)
{
  int i, k;
  static int column;			/* For stepping the graph display */
  static time_t graph_time;
  static corrected_event_t	tide;
  static time_t midnight_tonite; /* Cut off display at this time */
  static time_t midnight_offset; /* Cut off computation at this time */
  time_t start_t;		/* Start computation at this time */
  time_t t;
  static time_t thedayislong;
  
  restart: switch (ComputeState) {
  case CSIdle:
      return;

  case CSStart:				/* Through here once to start computation. */
      coeff_displaying = coeff_computing = false;
	if (coeff_swapped)		/* If coeff data is swapped (may have interrupted */
		SwapCoefficientData();	/*  previous computation), put it back. */
	need_zeros = iscurrent;
	figure_datum(0.0);		/* Make sure no offset applied */
  case CSRestart:
      offset_event_good = false;	/* Clear cache if we are starting again */
      MoveFaketime(1);
      midnight_tonite = midnight_offset = faketime;
      MoveFaketime(-1);
      midnight_start = faketime;
      thedayislong = midnight_tonite - faketime;
      start_t = faketime - 120;	/* Start a bit early to catch midnight events */
      if (PlaceOff)
	{		/* If location is an offset, adjust limits */
	  start_t -= maxoff * 60L;
	  midnight_offset -= minoff * 60L;
	}
      possible_happy_new_year(faketime);
//printf("Real win %ld - %ld\n",start_t/3600,midnight_tonite/3600);
      if ((Place->flags & P_COEFF)  &&	/* Do we need to compute coefficients? */
			Preferences.Pusecoeffs && !disgraph && coeff_displaying == false ) {
assert (!coeff_computing && !coeff_swapped);
		ncoeffs = 0;			/* Compute a new set */
		/* coeff_time was for caching but it's so fast, why bother? */
		coeff_time = faketime;
		coeff_computing = true;
		SwapCoefficientData();
		midnight_start = start_t -= 11L * hoursInSeconds;	/* Widen the time window 8 hours */
		midnight_tonite = midnight_offset += 5L * hoursInSeconds;
//printf("Win %ld - %ld\n",start_t/3600,midnight_tonite/3600);
		ComputeState = CSCoeff;
		}		
/*
 *	This should now be the only place the tide display is actually cleared and redrawn.
 */
	RedoDisplay(coeff_computing);	/* Clear the display */
	init_event(&tide.event, start_t);
	if (ComputeState == CSCoeff) goto restart;

	need_zeros = 			/* Compute zero events for current, or if alert set */
		iscurrent || (!disgraph && !PlaceOff && Preferences.Puselevel);
	column = 0;			/* Init graph display */
	graph_time = midnight_start;
	if (disgraph) {
		InitializeGraph(thedayislong);
		DisplayGraphLine(-1,TideAnyplace(graph_time - (24L * 60L * 60L)/160L, &k));
		}
        ComputeState = disgraph ? CSGraphing : CSComputing; 
	goto restart;

  case CSGraphing:		/* Displaying a graph */

	for (i = PlaceOff ? 8 : 20; i--; ) {	/* Do a few at a time, then check an event */
		DisplayGraphLine(column++,TideAnyplace(graph_time, &k));
		graph_time += thedayislong / 160L;
		if (column == 160) {
#ifdef LABELGRAPH
			ComputeState = CSComputing;
#else
			ComputeState = CSIdle;
#endif
			return;
		}	}
	return;

  case CSComputing:		/* Event list display */
  case CSCoeff:			/* or computing the coefficients */

//printf("Mark 3\n"); SysTaskDelay(50);
	if (!iscurrent && need_zeros) 		/* If user has asked for an alert, jack tides */
		figure_datum( - Preferences.Plevel 	/* down by that amount, corrected to locn's units */
			* known_units[Preferences.Plevelunits].conv_factor
			/ known_units[unit].conv_factor);
//printf("Mark 4\n"); SysTaskDelay(50);
	if ((t = next_big_event (&tide.event)) > midnight_offset) {
		if (ComputeState == CSComputing) break;		/* imapumpkin */
		SwapCoefficientData();				/* If coefficient is computed */
		coeff_computing = false;			/* Put everything back */
		if (ncoeffs) coeff_displaying = true;
		coeff_time = faketime;				/* Note time of coeff computation */
		ComputeState = CSRestart;			/* Go back and start real computation */
//SysTaskDelay(1000);
		return;
	  	}
	if (AbortFlag)
	  break;

//printf("Mark 5\n"); SysTaskDelay(50);
	/* now, next_big_event only returns one event at a time */
	if ((t = correct_event (&tide)) < midnight_start || t > midnight_tonite)
	  return;

//printf("Mark 6\n"); SysTaskDelay(50);
	if (ComputeState == CSComputing) {			/* If computing, display result */
//	printf("Event %d (%d) at %ld\n", tide.eventno, tide.event.type, t); SysTaskDelay(50);
		if (!ShowResult(t, tide.eventno, tide.currtide - tide_alert)) break;
			return;
		}
 	if (tide.eventno == HighTide) {				/* If storing coefficients and it's */
		coeff_values[ncoeffs].time = 				/*  a high tide, store the value - */
			t + 3 * hoursInSeconds;				/*  but add 3 hrs to center tide on */
		coeff_values[ncoeffs++].val = tide.currtide * 10 + 0.5;	/* French coast */
//printf("Coeff %d = %d at %ld\n",ncoeffs,coeff_values[ncoeffs-1].val,coeff_values[ncoeffs-1].time/3600 ); SysTaskDelay(100);
		}
	return;
  
  default: assert (0);		/* Shouldn't happen :-) */

  }  /* end switch */

  ComputeState = CSIdle;	/* a "break" will terminate the computation */
}
#endif /* not TEST */

#ifdef TEST

#include <math.h>

int
main ()
{
  uint32 t;
  double speed = 6.558641e-5;
  Speed s;

  SPEEDCONV(s,speed);
  for (t = 0; t < 100; t++)
    {
      double theta = 2 * M_PI * t / 100.0;
      printf("%lf %lf %lf\n",
	     theta,
	     FIXED2REAL(_fixed_cos(RADIANS2ANGLE(theta))),
	     FIXED2REAL(_fixed_cos(RADIANS2ANGLE(theta))) - cos(theta));
    }

  printf ("===================\n");
  
  for (t = 0; t < 3141592700u; t +=5432100) 
    {
      printf("%8lu %lf %lf\n",
	     t,
	     FIXED2REAL(_fixed_cos(SPEEDMUL(s,t))),
	     FIXED2REAL(_fixed_cos(SPEEDMUL(s,t))) - cos(2.0 * M_PI * speed * t));
    }
  return 0;
}
#endif
